package easik.database;

import easik.database.Options;
import easik.sketch.Sketch;
import easik.sketch.constraint.*;
import easik.sketch.vertex.EntityNode;
import easik.database.export.SketchExporter;
import easik.database.types.*;
import easik.ui.datamanip.ColumnEntry;

import easik.ui.menu.popup.*;
import easik.ui.*;
import easik.Easik;

import java.util.Map;
import java.util.Set;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import java.sql.*;

/**
 * This is the base class for all Easik database drivers.  Individual drivers
 * must extend this class, and the resulting driver can be used to interact
 * with the database and perform database-specific operations (such as
 * quoting).  Drivers must have a constructor that takes a
 * <code>Map&lt;String, ?&gt;</code> of connection options, however the driver
 * does *not* have to connect when a new object is created; if it is a typical
 * SQL server it should wait until getConnection() is called.
 */
public abstract class DatabaseDriver extends Options {

	/** Used by SQL drivers that support getConnection() to cache their connection.
	 */
	protected Connection sqlConn;

	/** Set to true to print executed statements to System.err */
	private static final boolean STATEMENT_DEBUGGING = false;

	/**
	 * Loads the database driver sketch of the given type (such as <code>MySQL</code>).  Whatever is
	 * passed in here will be looked for at easik.database.driver.{database type}.  That class must be
	 * a subclass of this one; if it loads successfully and is a subclass, it will be returned.  If
	 * loading fails for any reason, a DatabaseDriver.LoadException occurs.
	 *
	 * The options Map contains many driver-specific options, plus the following general options:
	 * <ul>
	 *   <li><option>pkFormat</option> &mdash; @see ColumnNaming </li>
	 *   <li><option>fkFormat</option> &mdash; @see ColumnNaming </li>
	 * </ul>
	 *
	 * @param type the database driver class to load from easik.database.driver.  For example, <code>MySQL</code>.
	 * @param options a Map of options to be passed to the individual driver.  See individual drivers for details.
	 * @throws DatabaseDriver.LoadException if the class cannot be loaded or instantiated for some reason
	 * @return a DatabaseDriver instance
	 * @see easik.database.driver.PostgreSQL
	 * @see easik.database.driver.MySQL
	 */
	public static DatabaseDriver loadDriver(String type, Map<String, ?> options) throws LoadException {
		// Only allow SQL types consisting of word characters:
		if (!type.matches("^\\w+$"))
			throw new LoadException("Invalid export SQL type '" + type + "'");

		// Attempt to load the specified class:
		Class subclass;
		try {
			subclass = DatabaseDriver.class.getClassLoader().loadClass("easik.database.driver." + type);
		} catch (ClassNotFoundException cnfe) {
			throw new LoadException(cnfe);
		}

		// Make sure the new class is a subclass of this class:
		if (!DatabaseDriver.class.isAssignableFrom(subclass))
			throw new LoadException("Driver class '" + subclass.getName() + "' is not a valid DatabaseDriver");

		Constructor<? extends DatabaseDriver> c;
		// Get the constructor that we can call:
		try {
			c = subclass.getConstructor(Map.class);
		} catch (NoSuchMethodException nsme) {
			throw new LoadException("Method does not exist: " + nsme.getMessage(), nsme);
		}

		DatabaseDriver dbd;
		try {
			dbd = c.newInstance(options);
		} catch (IllegalAccessException iae) { // Attempting to access something that is, for example, private
			throw new LoadException(iae);
		} catch (InstantiationException ie) { // This happens if something fails (such as invalid parameter types)
			throw new LoadException(ie);
		} catch (InvocationTargetException ite) { // This happens if the constructor throws an exception
			Throwable cause = ite.getCause();
			if (cause instanceof LoadException)
				throw (LoadException) cause;
			else
				throw new LoadException(cause);
		}

		return dbd;
	}

	/**
	 * Determines if this driver currently has an active connection.
	 * @return True if there is an active connection, false otherwise
	 */
	public boolean hasConnection(){
		return sqlConn != null;
	}
		
	/**
	 * Returns a list of available SQL drivers.  Currently this is just a hard-coded array,
	 * but in the future this may look for any class in the easik.database.driver package.
	 */
	public static String[] availableDrivers() {
		return new String[] {"MySQL", "PostgreSQL"};
	}

	/**
	 * Returns a new connection object, using information from options supplied
	 * to loadDriver().  Only needs to be supported by drivers that actually
	 * support SQL server connections.  The connection established by the first
	 * getConnection() on a DatabaseDriver instance will be cached; subsequent
	 * getConnection() calls will return the existing object instead of
	 * establishing a new connection.
	 *
	 * @throws DatabaseDriver.LoadException if the driver does not support SQL connections
	 * @throws SQLException if some SQL error (typically a connection error) occurs
	 */
	public Connection getConnection() throws LoadException, SQLException {
		throw new LoadException("Driver does not support SQL Connections");
	}

	/**
	 * Returns true if the driver supports making an SQL connection and enough
	 * options have been provided to establish a connection.  Subclasses that
	 * override getConnection() must also override this method.
	 */
	public boolean isConnectable() {
		return false;
	}

	/**
	 * The statement separator for SQL statements.  This is generally only used
	 * when generating multiple SQL statements (e.g. into a file), not when performing
	 * ordinary single database statements.
	 */
	public String getStatementSeparator() {
		return ";";
	}

	/**
	 * Quotes an identifier (table name, column name, etc.).  By default, this just returns
	 * the name trimmed, with spaces changed to underscores, however many drivers recognize
	 * the "quoteIdentifiers" option, which, when enabled, allows various other characters
	 * including whitespace.  See exporter subclasses for details.  If a driver overrides
	 * this method, it *must* recognize the "quoteIdentifiers" option.
	 *
	 * @param identifier the name to quote.  identifier.toString() will be the value used.
	 * @return the (possibly) quoted, database-safe string
	 */
	public String quoteId(Object identifier) {
		// By default, we just pass it off to cleanId
		return cleanId(identifier);
	}

	/**
	 * Takes an object, and returns an appropriate database name.  By default, this means
	 * trimming the name, changing all sequences of one or more non-word characters (i.e.
	 * anything other than a-z, A-Z, 0-9, and _) to a single underscore.  This name is meant
	 * to be used directly when embedding in another name; if the object is to be used as
	 * an identifier in its entirety, quoteId() should be called instead (which internally
	 * calls this).  If a driver supports the quoteIdentifiers option, it should override
	 * this method to strip as little as is required when quoteIdentifiers is enabled.
	 *
	 * @param identifier the identifier to be cleaned up
	 * @return the database-safe identifier string which, if quoteIdentifiers is enabled,
	 * 		   might only be safe after going through quoteId()
	 * @see quoteId(Object)
	 */
	public String cleanId(Object identifier) {
		if(identifier == null)
			return null;
		return identifier.toString().trim().replaceAll("\\W+", "_");
	}

	/**
	 * This creates and returns a SketchExporter instance for this driver type.
	 * If this driver doesn't have an Exporter class, a LoadException will occur.
	 *
	 * @param sketch the Sketch being exported
	 * @param exportOptions the options for the export, such as database names, create database options,
	 * 		etc.  See SketchExporter and individual export drivers for supported options.
	 * @see easik.database.export.SketchExporter
	 * @see easik.database.export.MySQL
	 * @see easik.database.export.PostgreSQL
	 */
	public SketchExporter getSketchExporter(Sketch sketch, Map<String, ?> exportOptions) throws LoadException {
		throw new LoadException("Driver does not support sketch exporting");
	}

	/**
	 * Executes a given SQL query that produces a ResultSet. Typically a SELECT statement.
	 *
	 * @param s The statement to execute
	 * @return The ResultSet returned from the query's execution
	 * @throws SQLExeception if an SQL error occurs; if getConnection() throws a LoadException, it will be
	 * 						 rethrown wrapped in an SQLException
	 */
	public ResultSet executeQuery(String s) throws SQLException {
		try {
			return getConnection().createStatement().executeQuery(s);
		}
		catch (LoadException le) {
			throw new SQLException("Unable to attempt SQL query: " + le.getMessage());//, le);
		}
	}

	/**
	 * A simple alias for obj.getConnection().prepareStatement(...).  Takes a
	 * query, which may contain literal value placeholders (such as "SELECT
	 * col1, col2 FROM tablename WHERE col1 = ? AND col = ?"), and executes it
	 * safely, quoting values replacing placeholders as required.  When handling
	 * user input of any sort, using executeQuery with setString() etc. calls
	 * on the PreparedStatement object is highly recommended over using
	 * executeQuery(). The returned prepared statement is set to not return
	 * auto-generated keys.
	 */
	public PreparedStatement prepareStatement(String sql) throws SQLException {
		try {
			return getConnection().prepareStatement(sql);
		}
		catch (LoadException le) {
			throw new SQLException("Unable to attempt SQL query: " + le.getMessage());//, le);
		}
	}

	/**
	 * Executes a given SQL updating statement. Which may be an INSERT, UPDATE, or DELETE statement.
	 * @param s The SQL statement to execute
	 */
	public void executeUpdate(String s) throws SQLException {
		if (STATEMENT_DEBUGGING) System.err.println("Executing statement: " + s);
		try {
			getConnection().createStatement().executeUpdate(s);
		}
		catch (LoadException le) {
			throw new SQLException("Unable to execute SQL statement: " + le.getMessage());//, le);
		}
	}


	/**
	 * Executes a given SQL updating statement. Which may be an INSERT, UPDATE, or DELETE statement.
	 * The statement must be in the form expected by java.sql.PreparedStatement
	 * @param s The SQL statement to execute
	 * @param input The set of ColumnEntry objects from which our values are retrieved. Note: This works
	 *   but relys on 'input' being iterated over in the same order that it was when 'sql' was generated.
	 */
	public void executePreparedUpdate(String sql, Set<ColumnEntry> input) throws SQLException {
		PreparedStatement ps = prepareStatement(sql);
		int col = 0;
		for(ColumnEntry entry : input){
			Type type = entry.getType();
			String value = entry.getValue();

			col++;
			// if value is NULL, assign and continue
			if(entry.getValue().equals("")){
				ps.setNull(col, type.getSqlType());
				continue;
			}

			// bind appropriate type to the prepared statement 
			type.bindValue(ps, col, entry.getValue());
		}

		ps.execute();
	}

	/**
	 * Returns the part of statement for an insertion with no columns, such as:
	 * <code>() VALUES ()</code> (to be used after <code>INSERT INTO tablename </code>).
	 * By default, we return "() VALUES ()", but some databases (e.g. postgresql) need
	 * something else (e.g. "DEFAULT VALUES");
	 *
	 * @return String to add when inserting without specifying any columns/values.
	 */
	public String emptyInsertClause() {
		return "() VALUES ()";
	}

	/**
	 * Takes a database Type and returns the string representation of that type, or	
	 * something as close as possible for this particular database.
	 *
	 * @param type the Type object desired
	 * @return the string containing the type, e.g., for an SQL create query
	 */
	public String getTypeString(Type type) {
		return type.toString(); // By default, just do whatever the type implements
	}

	/**
	 * Called from the sketch frame for determining if the insert/delete options should be enabled
	 * in the popup menu.
	 */
	public boolean editable(EntityNode entity){
		for(Constraint c : entity.getConstraints()){
			// this is the sum in c iff this is the codomain of any path in c
            if(c instanceof SumConstraint && c.getPaths().get(0).getCoDomain() == entity){
				return false;
			}
			// this is the product in c iff this is the domain of any path in c
			if(c instanceof ProductConstraint && c.getPaths().get(0).getDomain() == entity){
				return false;
			}
			if(c instanceof PullbackConstraint && ((PullbackConstraint)c).getSource() == entity){
				return false;
			}
		}
		return true;
	}

	/**
	 * Creates a database correspoinding to the current set of options, with the ability to drop the database
	 * before crating.
	 * @param drop Indicates if the database should be droped before creation, should a naming conflict arise.
	 * @return The success of the database creation.
	 */
	public boolean createDatabase(boolean drop) throws LoadException, SQLException {
		return false;
	}

	/**
	 * Drops a recreates a schema corresponding to the current set of options, if the driver supports schemas.
	 * @return Success of operation
	 */
	public boolean dropAndRecreateSchema() throws LoadException, SQLException {
		return false;
	}

	public static class LoadException extends Exception {
		private static final String messagePrefix = "Unable to load database driver: ";
		public LoadException(String message) {
			super(messagePrefix + message);
		}
		public LoadException(String message, Throwable cause) {
			super(messagePrefix + message + " (" + cause.getClass().toString() + ")", cause);
		}
		public LoadException(Throwable cause) {
			this(cause == null ? "An unknown error occured" : cause.getMessage(), cause);
		}
	}
}	
