package easik.database.driver;

import easik.database.DatabaseDriver;
import easik.database.types.*;
import easik.sketch.Sketch;

import java.util.Map;
import java.util.Set;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Arrays;
import java.util.Properties;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;

/**
 * EASIK PostgreSQL database driver.  Options supported by this driver are as follows.  Note that almost
 * all of these options are optional as PostgreSQL has defaults that can establish a connection with almost
 * no parameters (i.e. connecting to the local system as the current user).
 * , however getting a Connection object (from getConnection()) will
 * fail if certain options aren't provided.  This is used, for example, when a PostgreSQL driver is needed
 * but a connection is not: for example, when producing an SQL dump file.  Note that all of the options
 * marked as required for getConnection() are also required when exporting to a database (as
 * the exporter calls getConnection()).
 *
 * <ul>
 * 	<li><code>database</code> &mdash; required; the name of the database to use <i>[required for getConnection() and getSketchExporter()]</i></li>
 * 	<li><code>hostname</code> &mdash; the hostname of the server to connect to <i>[optional; will fall back to localhost if specifying a port, or use postgresql's default otherwise]</i></li>
 * 	<li><code>port</code> &mdash; the port of the server to connect to <i>[optional]</i></li>
 * 	<li><code>username</code> &mdash; the username to use when connecting <i>[optional]</i></li>
 * 	<li><code>password</code> &mdash; the password to use when connecting <i>[optional]</i></li>
 * 	<li><code>ssl</code> &mdash; if set to the string <code>"true"</code>, the driver will
 * 		attempt to use SSL when connecting to the database</li>
 * 	<li><code>connectProperties</code> &mdash; can be set to a {@link java.util.Properties} object
 * 		containing postgresql driver-specific options to use when connecting.</li>
 * 	<li><code>schema</code> &mdash; the PostgreSQL SCHEMA name to use.  Optional: if not specified,
 * 		the default schema (usually "public") will be used by PostgreSQL.  If provided, this will
 * 		issue a "SET search_path = schema" query.  See PostgreSQL documentation for details on schemas.</li>
 * 	<li><code>quoteIdentifiers</code> &mdash; if set to the string <code>"true"</code>, all identifiers (tables,
 * 		columns, etc.) will be quoted.  This means column and table names may contain any characters (including
 * 		spaces).  Also note that this will make tables and columns case-sensitive: unquoted identifiers are always
 * 		treated by PostgreSQL as lower-case names.  Note that certain reserved names (e.g. "Table"&mdash;see the
 * 		list below) will always be quoted: if quoteIdentifiers is not enabled, the quoted, lower-cased name will be
 * 		used.  Also note that if quoteIdentifiers is not enabled, sequences of characters other than a-z, A-Z, 0-9,
 * 		and _ will be converted to underscores.</li>
 * </ul>
 *
 * @see easik.database.export.DatabaseDriver
 */
public class PostgreSQL extends DatabaseDriver {
	/**
	 * The set of reserved PostgreSQL keywords.  If not using the quoteIdentifiers option,
	 * any columns/tables using these words will be lower-cased and quoted anyway.  The
	 * keywords stored here are all upper-case.  This list was compiled from the PostgreSQL
	 * 8.3 documentation, Appendix C.
	 */
	public static final Set<String> RESERVED_KEYWORDS = new HashSet<String>(Arrays.asList((
			"ALL ANALYSE ANALYZE AND ANY ARRAY AS ASC ASYMMETRIC AUTHORIZATION "+
			"BETWEEN BINARY BOTH "+
			"CASE CAST CHECK COLLATE COLUMN CONSTRAINT CREATE CROSS CURRENT_DATE "+
			"CURRENT_ROLE CURRENT_TIME CURRENT_TIMESTAMP CURRENT_USER "+
			"DEFAULT DEFERRABLE DESC DISTINCT DO "+
			"ELSE END EXCEPT FALSE FOR FOREIGN FREEZE FROM FULL "+
			"GRANT GROUP HAVING ILIKE IN INITIALLY INNER INTERSECT INTO IS ISNULL "+
			"JOIN LEADING LEFT LIKE LIMIT LOCALTIME LOCALTIMESTAMP "+
			"NATURAL NEW NOT NOTNULL NULL "+
			"OFF OFFSET OLD ON ONLY OR ORDER OUTER OVERLAPS "+
			"PLACING PRIMARY REFERENCES RETURNING RIGHT "+
			"SELECT SESSION_USER SIMILAR SOME SYMMETRIC "+
			"TABLE THEN TO TRAILING TRUE UNION UNIQUE USER USING "+
			"VERBOSE WHEN WHERE WITH"
		).split("\\s+")
	));

	/**
	 * Creates a new PostgreSQL database driver object.  Should not be called directly.
	 * @see DatabaseDriver.loadDriver(String, Map)
	 */
	public PostgreSQL(Map <String, ?> opts) {
		options = new HashMap<String, Object>(opts);
	}

	public boolean isConnectable() {
		return hasOption("database");
	}

	/**
	 * Creates a database corresponding to the current set of options.
	 * @param drop Indicates if the database should be droped before creation, should a naming conflict arise.
	 * @return Success of creation.
	 */
	public boolean createDatabase(boolean drop) throws LoadException, SQLException {
		// Add createDatabase to connection options so we know not to append database name to dsn
		options.put("createDatabase", "true");

		// clear a preexisting connection and get a new one in case the database already exists (will result in an error
		// when we try to creat, if we do not have drop enabled
		// This will be reset when getConnection() gets called (inside executeUpdate)
		sqlConn = null;

		if(drop)
			executeUpdate("DROP DATABASE IF EXISTS " + options.get("database"));

		executeUpdate("CREATE DATABASE " + options.get("database"));

		// We have to establish a new connection to the database (we're probably
		// currently connected to "postgres"):
		options.remove("createDatabase");
		sqlConn = null;
		getConnection();

		return true;
	}

	/**
	 * Drops and recreates a schema corresponding to the current set of options.  Does nothing
	 * if the specified schema name is null or an empty string.
	 * @param drop Indicates if the database should be droped before creation, should a naming conflict arise.
	 * @return Success of creation.
	 */
	public boolean dropAndRecreateSchema() throws LoadException, SQLException {

		String schemaName = (String) options.get("schema");
		if (schemaName == null || schemaName.equals(""))
			return false;

		executeUpdate("DROP SCHEMA IF EXISTS " + quoteId(options.get("schema")) + " CASCADE");
		executeUpdate("CREATE SCHEMA " + quoteId(options.get("schema")));
		executeUpdate("SET search_path = " + quoteId(options.get("schema")));

		return true;
	}

	public Connection getConnection() throws DatabaseDriver.LoadException, SQLException {
		if (sqlConn != null) return sqlConn;

		if (!isConnectable())
			throw new LoadException("Insufficient driver parameters for establishing a database connection: database is required");

		String[] trydb;
		if (hasOption("createDatabase")) {
			// We need to connect to some database, so try "postgres", which usually exists;
			// if that fails, don't use anything to see if there is some default DB
			trydb = new String[] {"postgres", ""};
		}
		else {
			trydb = new String[] {"" + options.get("database")};
		}

		Properties props;
		if (options.containsKey("connectProperties") && options.get("connectProperties") instanceof Properties)
			props = new Properties((Properties) options.get("connectProperties"));
		else
			props = new Properties();

		if (options.containsKey("username"))
			props.setProperty("user", options.get("username")+"");
		if (options.containsKey("password"))
			props.setProperty("password", options.get("password")+"");
		if (optionEnabled("ssl"))
			props.setProperty("ssl", "true");

		try {
			// Registers the pgsql driver with the DriverManager
			Class.forName("org.postgresql.Driver");
		} catch (ClassNotFoundException cnfe) {
			throw new DatabaseDriver.LoadException("Unable to load PostgreSQL driver", cnfe);
		}

		for (int i = 0; i < trydb.length; i++) {
			String dbname = trydb[i];
			StringBuilder dsn = new StringBuilder("jdbc:postgresql:");
			boolean usingHostname = false;
			// Pg allows blah.hostname.com, 10.11.12.13, and IPv6 addresses enclosed in [], such as [::1]
			if (optionMatches("hostname", "^(?:\\[[[:xdigit:]:]+\\]|[A-Za-z0-9.-]+)$")) {
				usingHostname = true;
				dsn.append("//" + options.get("hostname"));
			}

			if (optionMatches("port", "^[0-9]+$")) {
				if (!usingHostname) {
					usingHostname = true;
					dsn.append("//localhost");
				}
				dsn.append(":" + options.get("port"));
			}
			if (usingHostname)
				dsn.append("/");

			dsn.append(dbname);

			try {
				sqlConn = DriverManager.getConnection(dsn.toString(), props);
			}
			catch (SQLException sqle) {
				if (i == trydb.length-1) {
					throw sqle;
				}
			}
		}

		if (hasOption("schema")) {
			try {
				executeUpdate("SET search_path = " + quoteId(options.get("schema")+""));
			}
			catch (SQLException e) {} // Ignore an SQL exception here: the schema is allowed to not exist yet (for instance, when we're exporting).
		}

		return sqlConn;
	}

	/**
	 * Creates a new SketchExporter object for this PostgreSQL driver.  The database parameter
	 * must have been specified when this database driver was created.
	 *
	 * @throws DatabaseDriver.LoadException if the <code>database</code> option was not
	 * 		specified when this PostgreSQL database driver was created.
	 */
	public easik.database.export.PostgreSQL getSketchExporter(Sketch sketch, Map<String, ?> exportOpts)
				throws DatabaseDriver.LoadException {

		if (!hasOption("database"))
			throw new DatabaseDriver.LoadException("PostgreSQL exporter requires \"database\" parameter!");

		return new easik.database.export.PostgreSQL(sketch, this, exportOpts);
	}

	/**
	 * PostgreSQL-specific identifier cleaning.  Identifer cleaning is significantly different depending on whether
	 * quoteIdentifiers is enabled.  If enabled, we allow anything in an identifier (but double-up " characters).
	 * Otherwise we fall back to the conservative SketchExporter default.
	 */
	public String cleanId(Object id) {
		if (optionEnabled("quoteIdentifiers"))
			return id.toString().replaceAll("\"", "\"\"");
		else
			return super.cleanId(id);
	}
	/**
	 * PostgreSQL-specific identifier quoting.  If the "quoteIdentifiers" option is
	 * enabled, we quote identifiers using the standard " quote character.  Otherwise,
	 * if the cleaned identifier is a reserved keyword, we force it to lower-case (since
	 * unquoted identifiers are treated as lower-case in PostgreSQL) and quote it;
	 * otherwise we use the straight cleaned identifier.
	 */
	public String quoteId(Object id) {
		String cleaned = cleanId(id);
		if (optionEnabled("quoteIdentifiers"))
			return "\"" + cleaned + "\"";
		else if (RESERVED_KEYWORDS.contains(cleaned.toUpperCase()))
			// If the name is reserved, we need to force it to lower-case and quote it anyway
			return "\"" + cleaned.toLowerCase() + "\"";
		else
			return cleaned;
	}

	public String emptyInsertClause() {
		return "DEFAULT VALUES";
	}

	public String getTypeString(Type type) {
		if (type instanceof Blob)
			return "BYTEA";
		else
			return type.toString();
	}
}

