package easik.database.export;

import easik.EasikTools;
import easik.overview.vertex.*;
import easik.sketch.Sketch;
import easik.database.DatabaseDriver;
import easik.sketch.vertex.EntityNode;
import easik.sketch.attribute.EntityAttribute;
import easik.sketch.edge.*;
import easik.sketch.path.SketchPath;
import easik.sketch.keys.*;
import easik.sketch.constraint.*;
import easik.view.vertex.*;

import java.io.*;

import java.util.Map;
import java.util.List;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Collection;
import java.util.Collections;


/**
 * EASIK MySQL sketch export driver.  This driver is loaded by calling getSketchExporter()
 * on a MySQL database driver object.  The following options passed in to the
 * getSketchExporter options parameter, in addition to those mentioned in SketchExporter,
 * are supported by this export driver:
 *
 * <ul>
 * 	<li><code>createDatabase</code> &mdash; if set to <code>"true"</code>, <code>CREATE DATABASE</code>
 * 		statements will be added to the SQL dump <i>[{@link exportToString()} and
 * 		{@link exportToFile(File)} only]</i></li>
 * 	<li><code>dropDatabase</code> &mdash; if set to <code>"true"</code>, and using the above
 * 		<code>createDatabase</code> option, <code>DROP DATABASE</code> statements will be added to
 * 		the SQL dump <i>[exportToString() and exportToFile() only]</i></li>
 * </ul>
 *
 * @see easik.database.export.SketchExporter
 * @see easik.database.driver.MySQL
 */
public class MySQL extends SketchExporter {

	/**
	 * The table options we want to use--usually just the storage engine we
	 * want.  InnoDB is by far the most reliable, and typically the fastest
	 * (especially under concurrent, mixed read/write access), and most
	 * importantly, it is the only engine that supports foreign keys.
	 */
	public static final String TABLE_OPTIONS = " ENGINE=InnoDB";

	// Will be set to true if we need to add a fail procedure
	private boolean addFail = false;

	// MySQL doesn't support multiple triggers for the same action,
	// so we store them up here to group them together
	private LinkedHashMap<EntityNode, LinkedHashMap<String, LinkedList<String>>> triggers =
		new LinkedHashMap<EntityNode, LinkedHashMap<String, LinkedList<String>>>();

	/** The DatabaseDriver object used in this export.*/
	private DatabaseDriver dbd;

	public MySQL(Sketch sketch, DatabaseDriver db, Map<String, ?> options) throws DatabaseDriver.LoadException {
		super(sketch, db, options);
		if (!db.hasOption("database")){
			throw new DatabaseDriver.LoadException("MySQL export driver requires \"database\"  parameter!");
		}
		dbd = db;
	}

	public List<String> initialize() {
		// super.initialize() gives us the current date/time and Easik blurb in a comment
		LinkedList<String> sql = new LinkedList<String>(super.initialize());

		String dbNameQ = quoteId(dbd.getOption("database"));

		sql.addAll(comment("All sorts of things won't work properly without InnoDB tables:"));
		sql.add("SET SESSION sql_mode='NO_ENGINE_SUBSTITUTION'"+$);

		// If we're working on a live connection, don't try this drop/create stuff: on
		// almost any secured MySQL setup, it won't work anyway.  This means that to
		// work on a live database, we need the user to first create the database
		// and grant permissions to use it.
		if (mode != Mode.DATABASE) {
		   	if (optionEnabled("createDatabase")) {
				if (optionEnabled("dropDatabase"))
					sql.add("DROP DATABASE IF EXISTS " + dbNameQ+$);

				sql.add("CREATE DATABASE IF NOT EXISTS " + dbNameQ+$);
			}

			sql.add("USE " + dbNameQ+$);
		}

		return sql;
	}

	public List<String> createTable(EntityNode table, boolean includeRefs) {
		StringBuilder tableDef = new StringBuilder();
		tableDef.append("CREATE TABLE " + quoteId(table) + " (\n");
		tableDef.append("	" + quoteId(tablePK(table)) + " "+pkType()+" PRIMARY KEY AUTO_INCREMENT");
		for (EntityAttribute attr: table.getEntityAttributes()) {
			// FIXME -- what about NOT NULL, etc. on the column?  Defaults?
			tableDef.append(",\n	" + quoteId(attr.getName()) + " " + dbd.getTypeString(attr.getType()));
		}

		// Include the foreign key columns, and the actual foreign keys (if includeRefs set)
		for (SketchEdge edge: table.getOutgoingEdges()) {
			tableDef.append(",\n	" + quoteId(tableFK(edge)) + " "+pkType());
			if (!edge.isPartial())
				tableDef.append(" NOT NULL");
			if (edge.isInjective())
				tableDef.append(" UNIQUE");
			if (includeRefs) {
				tableDef.append(",\n	FOREIGN KEY (" + quoteId(tableFK(edge)) +
						") REFERENCES " + quoteId(edge.getTargetEntity()) + " (" +
						quoteId(tablePK(edge.getTargetEntity())) + ")");
				String refMode =
					edge.getCascading() == SketchEdge.Cascade.SET_NULL ? "SET NULL" :
					edge.getCascading() == SketchEdge.Cascade.CASCADE ? "CASCADE" :
					"NO ACTION"; // Effectively the same as RESTRICT, but RESTRICT won't even work if cleaned up in the same transaction
				tableDef.append(" ON DELETE " + refMode + " ON UPDATE " + refMode);
			}
		}

		// Add the unique keys
		for (UniqueKey key: table.getUniqueKeys()) {
			tableDef.append(",\n	UNIQUE " + quoteId(table.getName() + "_" + key.getKeyName()) + " (");
			LinkedList<String> keyCols = new LinkedList<String>();
			for (UniqueIndexable elem: key.getElements()) {
				if (elem instanceof SketchEdge) {
					keyCols.add(quoteId(tableFK((SketchEdge) elem)));
				}
				else { // Probably an attribute
					keyCols.add(quoteId(elem.getName()));
				}
			}
			tableDef.append(EasikTools.join(", ", keyCols) + ")");
		}

		if (includeRefs) {
			// We need a copy of any sum tables' not-null columns (which are simply the
			// outgoing, non-partial edges) if this table is a summand; we shadow the
			// columns here, then handle propagating them in the SumConstraint triggers.
			for (SketchEdge edge: table.getShadowEdges()) {
				tableDef.append(",\n	" + quoteId(tableFK(edge)) + " "+pkType());
				tableDef.append(",\n	FOREIGN KEY (" + quoteId(tableFK(edge)) +
						") REFERENCES " + quoteId(edge.getTargetEntity()) + " (" +
						quoteId(tablePK(edge.getTargetEntity())) + ") ON DELETE SET NULL ON UPDATE SET NULL");
			}
		}

		tableDef.append("\n)" + TABLE_OPTIONS + $);

		return Collections.singletonList(tableDef.toString());
	}

	public List<String> createReferences(EntityNode table) {
		LinkedList<String> refs = new LinkedList<String>();

		String tableQ = quoteId(table);
		for (SketchEdge edge: table.getOutgoingEdges()) {
			EntityNode target = edge.getTargetEntity();

			String refMode =
				edge.getCascading() == SketchEdge.Cascade.SET_NULL ? "SET NULL" :
				edge.getCascading() == SketchEdge.Cascade.CASCADE ? "CASCADE" :
				"NO ACTION"; // No different from RESTRICT under MySQL (in theory, RESTRICT shouldn't work even if cleaned up in the same transaction, while NO ACTION should)
			refs.add(
				"ALTER TABLE " + tableQ + " ADD FOREIGN KEY (" + quoteId(tableFK(edge)) + ") " +
				"REFERENCES " + quoteId(target) + " (" + quoteId(tablePK(target)) + ") " +
				"ON DELETE " + refMode + " ON UPDATE " + refMode + $);
		}

		// We need a copy of any sum tables' not-null columns (which are simply the
		// outgoing, non-partial edges) if this table is a summand; we shadow the
		// columns here, then handle propagating them in the SumConstraint triggers.
		for (SketchEdge edge: table.getShadowEdges()) {
			refs.add("ALTER TABLE " + tableQ + " ADD COLUMN " + quoteId(tableFK(edge)) + " "+pkType());
			refs.add("ALTER TABLE " + tableQ + " ADD FOREIGN KEY (" + quoteId(tableFK(edge)) +
					") REFERENCES " + quoteId(edge.getTargetEntity()) + " (" + quoteId(tablePK(edge.getTargetEntity())) + ") " +
					"ON DELETE SET NULL ON UPDATE SET NULL");
		}

		return refs;
	}

	public List<String> createConstraint(CommutativeDiagram cd, String id) {
		LinkedList<String> sql = new LinkedList<String>();
		List<SketchPath> paths = cd.getPaths();
		EntityNode dom = paths.get(0).getDomain();

		StringBuilder proc = new StringBuilder();
		LinkedList<String> declarations = new LinkedList<String>();
		LinkedList<String> values = new LinkedList<String>();
		LinkedList<String> args = new LinkedList<String>();
		LinkedList<String> params = new LinkedList<String>();
		int targetNum = 0;
		for (SketchPath path: paths) {
			++targetNum;
			LinkedList<SketchEdge> tmpPath = new LinkedList<SketchEdge>(path.getEdges());
			tmpPath.removeFirst();

			String fk = "_path" + targetNum + "fk";
			args.add(fk + " " + pkType());
			params.add("NEW." + quoteId(tableFK(path.getFirstEdge())));
			declarations.add("_cdTarget" + targetNum);
			if (tmpPath.size() == 0)
				values.add("	SELECT " + fk + " INTO _cdTarget"+targetNum + ";\n");
			else
				values.add("	SELECT " + qualifiedFK(path.getLastEdge()) + " INTO _cdTarget"+targetNum +
						" FROM " + joinPath(tmpPath, false) +
						" WHERE " + qualifiedPK(tmpPath.getFirst().getSourceEntity()) + " = " + fk + ";\n");
		}

		proc.append(
				"CREATE PROCEDURE " + quoteId("commutativeDiagram" + id) + "(" + EasikTools.join(", ", args) + ") BEGIN\n" +
				"	DECLARE " + EasikTools.join(", ", declarations) + " " + pkType() + ";\n" +
				EasikTools.join("", values) +
				"	IF\n");
		for (int i = 2; i <= targetNum; i++) {
			proc.append("		NOT (_cdTarget1 <=> _cdTarget"+i+")");
			if (i < targetNum)
				proc.append(" OR");
			proc.append("\n");
		}

		proc.append(
				"	THEN CALL constraint_failure('Commutative diagram constraint failure.');\n" +
				"	END IF;\n" +
				"END");
		addFail = true;

		sql.addAll(delimit("$$", proc));

		addTrigger(dom, "BEFORE INSERT", "CALL " + quoteId("commutativeDiagram" + id) + "(" +
				EasikTools.join(", ", params) + ")");

		return sql;
	}

	public List<String> createConstraint(ProductConstraint constraint, String id) {
		LinkedList<String> sql = new LinkedList<String>();

		String delConName = "productConstraint" + id + "Delete";

		StringBuilder proc = new StringBuilder();
		proc.append("CREATE PROCEDURE " + quoteId(delConName) + "(id "+pkType()+") BEGIN\n");

		EntityNode begin = null;
		for (SketchPath p: constraint.getPaths()) {
			if (begin == null) begin = p.getDomain();
			proc.append(
				"	DELETE " + quoteId(p.getCoDomain()) + " FROM " + joinPath(p) + "\n" +
				"		WHERE " + qualifiedPK(begin) + " = id;\n");
		}
		proc.append("END");

		sql.add(proc.toString());

		// Now create the trigger to call the new procedure
		addTrigger(begin, "BEFORE DELETE",
				"CALL " + quoteId("productConstraint" + delConName) + "(OLD." + quoteId(tablePK(begin)) + ")");

		for (SketchPath p: constraint.getPaths()) {
			EntityNode dest = p.getCoDomain();
			String conName = "productConstraint" + id + "Insert" + cleanId(dest);

			LinkedList<String> args = new LinkedList<String>(), params = new LinkedList<String>();
			args.add("id " + pkType());
			params.add("NEW." + quoteId(tablePK(dest)));

			for (SketchEdge shadow: dest.getShadowEdges()) {
				args.add(quoteId("NEW_shadow_" + tableFK(shadow)) + " " + pkType());
				params.add("NEW." + quoteId(tableFK(shadow)));
			}

			proc = new StringBuilder();
			proc.append(
				"CREATE PROCEDURE " + quoteId(conName) + "("+EasikTools.join(", ", args)+") BEGIN\n" +
				"	DECLARE _lastId "+pkType()+";\n");

			StringBuilder createIntermediates = new StringBuilder();

			LinkedList<String> clauses = new LinkedList<String>();
			for (SketchPath q: constraint.getPaths()) {
				if (p == q) continue;
				// Older versions were doing "SELECT COUNT(*) FROM TABLE", but that has a big
				// performance hit for very large, busy InnoDB tables, when all we really care
				// about is non-empty:
				clauses.add("(SELECT 1 FROM " + quoteId(q.getCoDomain()) + " LIMIT 1) = 1");

				// If we end up putting anything in createIntermediate, we'll use these
				// insertions when we insert the first row into one of the path targets
				SketchEdge[] edges = q.getEdges().toArray(new SketchEdge[0]);
				for (int i = edges.length - 1; i > 0; i--) {
					SketchEdge e = edges[i];
					EntityNode source = e.getSourceEntity(), target = e.getTargetEntity();
					createIntermediates.append("			");
					insertInto(
							createIntermediates, true, source, dest.getShadowEdges(), qualifiedPK(target), quoteId(target),
							Collections.singletonList(quoteId(tableFK(e))), null);
				}
			}

			// In words: If the tables forming the domains of the other paths contain items.
			proc.append("	IF " + EasikTools.join(" AND ", clauses) + " THEN\n");

			if (createIntermediates.length() > 0) {
				// If we just inserted the first row, we're going to have to build a path of intermediate tables
				// for the other paths, which we built in createIntermediate.
				proc.append("		IF (SELECT COUNT(*) FROM (SELECT 1 FROM " + quoteId(dest) + " LIMIT 2) a) = 1 THEN\n");
				proc.append(createIntermediates);
				proc.append("		END IF;\n\n");
			}

			// Produce the intermediate path insertion strings
			SketchEdge[] edges = p.getEdges().toArray(new SketchEdge[0]);
			if (edges.length > 1) {
				for (int i = edges.length - 1; i > 0; i--) {
					SketchEdge e = edges[i];
					EntityNode source = e.getSourceEntity(), target = e.getTargetEntity();
					proc.append("		");
					insertInto(proc, true, source, dest.getShadowEdges(), null, null,
							Collections.singletonList(quoteId(tableFK(e))),
							Collections.singletonList(i == edges.length-1 ? "id" : "LAST_INSERT_ID()"));
				}
				proc.append("		SET _lastId = LAST_INSERT_ID();\n");
			}
			else {
				proc.append("		SET _lastId = id;\n");
			}


			// Now the proper insertion
			LinkedList<String> columns = new LinkedList<String>();
			LinkedList<String> values = new LinkedList<String>();
			LinkedList<String> from = new LinkedList<String>();
			EntityNode thisTarget = null;
			for (SketchPath q: constraint.getPaths()) {
				EntityNode target = q.getFirstEdge().getTargetEntity();
				columns.add(quoteId(tableFK(q.getFirstEdge())));
				values.add(qualifiedPK(target));
				from.add(quoteId(target));
				if (q == p) thisTarget = target;
			}
			proc.append("		");
			insertInto(proc, true, begin, dest.getShadowEdges(),
					EasikTools.join(", ", values),
					EasikTools.join(" CROSS JOIN ", from) + " WHERE " + qualifiedPK(thisTarget) + " = _lastId",
					columns, null);

			proc.append("	END IF;\n");
			proc.append("END");

			sql.add(proc.toString());

			// Now create the trigger to call the new procedure
			addTrigger(dest, "AFTER INSERT",
					"CALL " + quoteId(conName) + "("+EasikTools.join(", ", params)+")");
		}

		return delimit("$$", sql);
	}

	public List<String> createConstraint(PullbackConstraint constraint, String id) {
		LinkedList<String> sql = new LinkedList<String>();

		// Create the trigger for inserting into the pullback table:
		EntityNode pb = constraint.getSource(), target = constraint.getTarget();
		SketchPath projA = constraint.getProjectionA(), projB = constraint.getProjectionB();
		SketchPath targetA = constraint.getTargetPathA(), targetB = constraint.getTargetPathB();
		SketchPath pathA = constraint.getFullPathA(), pathB = constraint.getFullPathB();

		String insConName = "pullbackConstraint" + id + "Insert" + cleanId(pb);
		StringBuilder proc = new StringBuilder();
		proc.append("CREATE PROCEDURE " + quoteId(insConName) + "(_newPBid "+pkType()+") BEGIN\n");
		proc.append("	DECLARE _pathAId, _pathBId "+pkType()+";\n");
		String basePK = qualifiedPK(pb), targetPK = qualifiedPK(target);
		proc.append(
			"	SELECT " + targetPK + " INTO _pathAId\n" +
			"		FROM " + joinPath(pathA) + "\n" +
			"		WHERE " + basePK + " = _newPBid;\n");
		proc.append(
			"	SELECT " + targetPK + " INTO _pathBId\n" +
			"		FROM " + joinPath(pathB) + "\n" +
			"		WHERE " + basePK + " = _newPBid;\n");
		proc.append( // Make sure the two paths match
			"	IF NOT (_pathAId <=> _pathBId) THEN\n" +
			"		CALL constraint_failure('Invalid entry in pullback constraint.');\n" +
			"	END IF;\n");
		addFail = true;

		String pbQ = quoteId(pb);
		SketchEdge edgeA = pathA.getFirstEdge(), edgeB = pathB.getFirstEdge();
		String colA = qualifiedFK(edgeA), colB = qualifiedFK(edgeB);

		// Make sure there are no duplicate entries; we do this by adding a unique key on
		// colA and colB of the pullback table (unless one already exists)
		if (!addedUnique.isUnique(pb, edgeA, edgeB)) {
			sql.add("ALTER TABLE "+pbQ+" ADD UNIQUE ("+quoteId(tableFK(edgeA))+", "+quoteId(tableFK(edgeB))+")"+$);
			addedUnique.add(pb, edgeA, edgeB);
		}

		proc.append("END");
		sql.add(proc.toString());

		addTrigger(pb, "AFTER INSERT",
				"CALL " + quoteId(insConName) + "(NEW." + quoteId(tablePK(pb)) + ")");

		pbAfterInsertProcs(sql, id, constraint, targetA, projA, targetB, projB);
		pbAfterInsertProcs(sql, id, constraint, targetB, projB, targetA, projA);

		// Finally, the before-delete on the pullback:
		String delConName = "pullbackConstraint" + id + "Delete" + cleanId(pb);
		proc = new StringBuilder();
		proc.append("CREATE PROCEDURE " + quoteId(delConName) + "(toDeleteId "+pkType()+") BEGIN\n");
		LinkedList<SketchEdge>
			tmpPathA = new LinkedList<SketchEdge>(pathA.getEdges()),
			tmpPathB = new LinkedList<SketchEdge>(pathB.getEdges());
		SketchEdge
			firstEdgeA = tmpPathA.removeFirst(),
			firstEdgeB = tmpPathB.removeFirst();
		SketchEdge
			lastEdgeA = tmpPathA.getLast(),
			lastEdgeB = tmpPathB.getLast();
		SketchEdge
			secondEdgeA = tmpPathA.removeFirst(),
			secondEdgeB = tmpPathB.removeFirst();

		EntityNode firstA = firstEdgeA.getTargetEntity(), firstB = firstEdgeB.getTargetEntity();

		proc.append(
			"	DELETE " + quoteId(target) + " FROM " + pbQ + " JOIN " +
			"(" + quoteId(firstA) + ", " + quoteId(firstB) + ") ON " +
			qualifiedFK(firstEdgeA) + " = " + qualifiedPK(firstA) + " AND " + qualifiedFK(firstEdgeB) + " = " + qualifiedPK(firstB));

		if (tmpPathA.size() > 0)
			proc.append(" JOIN " + joinPath(tmpPathA, false, qualifiedFK(secondEdgeA) + " = " + qualifiedPK(secondEdgeA.getTargetEntity())));
		if (tmpPathB.size() > 0)
			proc.append(" JOIN " + joinPath(tmpPathB, false, qualifiedFK(secondEdgeB) + " = " + qualifiedPK(secondEdgeB.getTargetEntity())));

		proc.append(" JOIN " + quoteId(target) + " ON " +
			qualifiedFK(lastEdgeA) + " = " + qualifiedPK(target) + " AND " +
			qualifiedFK(lastEdgeB) + " = " + qualifiedPK(target) + "\n");
		proc.append("		WHERE " + qualifiedPK(pb) + " = toDeleteId;\n");

		proc.append("END");
		sql.add(proc.toString());

		addTrigger(pb, "BEFORE DELETE",
				"CALL " + quoteId(delConName) + "(OLD." + quoteId(tablePK(pb)) + ")");

		return delimit("$$", sql);
	}

	private void pbAfterInsertProcs(List<String> sql, String id, PullbackConstraint constraint, SketchPath thisPath, SketchPath thisFirstPath, SketchPath otherPath, SketchPath otherFirstPath) {
		EntityNode source = thisPath.getDomain(), otherSource = otherPath.getDomain(),
			pb = thisFirstPath.getDomain(), target = thisPath.getCoDomain();

		String insConName = "pullbackConstraint" + id + "Insert" + cleanId(source);
		String thisJoinPath = joinPath(thisPath, false);

		LinkedList<String> args = new LinkedList<String>(), params = new LinkedList<String>();
		args.add("_newId " + pkType());
		params.add("NEW." + quoteId(tablePK(source)));

		for (SketchEdge shadow: source.getShadowEdges()) {
			args.add(quoteId("NEW_shadow_" + tableFK(shadow)) + " " + pkType());
			params.add("NEW." + quoteId(tableFK(shadow)));
		}

		StringBuilder proc = new StringBuilder();
		proc.append("CREATE PROCEDURE " + quoteId(insConName) + "(" + EasikTools.join(", ", args) + ") BEGIN\n");
		String lastJoinOn = "JOIN " + quoteId(target) + " ON " +
			qualifiedFK(thisPath.getLastEdge()) + " = " + qualifiedPK(target) + " AND " +
			qualifiedFK(otherPath.getLastEdge()) + " = " + qualifiedPK(target);
		proc.append("	IF (SELECT COUNT(*) FROM (" + thisJoinPath + ", " + joinPath(otherPath, false) +
				") " + lastJoinOn + " WHERE " + qualifiedPK(source) + " = _newId) > 0 THEN\n");

		SketchEdge[] edges = otherFirstPath.getEdges().toArray(new SketchEdge[0]);
		if (edges.length >= 2) {
			proc.append("		IF (SELECT COUNT(*) FROM (" + joinPath(otherFirstPath.getEdges(), true) + ", " + thisJoinPath +
					") " + lastJoinOn + " WHERE " + qualifiedPK(source) + " = _newId) = 0 THEN\n");

			LinkedList<SketchEdge> partialPath = new LinkedList<SketchEdge>(otherPath.getEdges());
			for (int i = edges.length-1; i > 0; i--) {
				EntityNode s = edges[i].getSourceEntity(), t = edges[i].getTargetEntity();
				proc.append("			");
				insertInto(proc, true, s, source.getShadowEdges(),
						qualifiedPK(t), // SELECT this FROM:
						"(" + joinPath(partialPath, false) + ", " + thisJoinPath + ") " + lastJoinOn +
								" WHERE " + qualifiedPK(source) + " = _newId",
						Collections.singletonList(quoteId(tableFK(edges[i]))), null);
				partialPath.addFirst(edges[i]);
			}
			proc.append("		END IF;\n");
		}

		edges = thisFirstPath.getEdges().toArray(new SketchEdge[0]);
		LinkedList<SketchEdge> partialPath = new LinkedList<SketchEdge>(thisPath.getEdges());
		for (int i = edges.length - 1; i > 0; i--) {
			SketchEdge e = edges[i];
			EntityNode s = e.getSourceEntity(), t = e.getTargetEntity();
			proc.append("		");
			insertInto(proc, true, s, source.getShadowEdges(),
					qualifiedPK(t),
					joinPath(partialPath) + " WHERE " + qualifiedPK(source) + " = _newId",
					Collections.singletonList(quoteId(tableFK(e))), null);
		}

		EntityNode firstStop = thisFirstPath.getFirstCoDomain(),
				   otherFirstStop = otherFirstPath.getFirstCoDomain();
		// We need the two paths, but we need to remove the first element, because we don't want to include the pullback table
		LinkedList<SketchEdge> tmpPathA = new LinkedList<SketchEdge>(constraint.getFullPathA().getEdges());
		LinkedList<SketchEdge> tmpPathB = new LinkedList<SketchEdge>(constraint.getFullPathB().getEdges());
		tmpPathA.removeFirst(); tmpPathB.removeFirst();
		proc.append("		");
		LinkedList<String> cols = new LinkedList<String>();
		cols.add(quoteId(tableFK(thisFirstPath.getFirstEdge())));
		cols.add(quoteId(tableFK(otherFirstPath.getFirstEdge())));
		insertInto(proc, true, pb, source.getShadowEdges(),
				qualifiedPK(firstStop) + ", " + qualifiedPK(otherFirstStop),
				"(" + joinPath(tmpPathA, false) + ", " + joinPath(tmpPathB, false) + ") " + lastJoinOn +
						" WHERE " + qualifiedPK(source) + " = _newId",
				cols, null);

		proc.append("	END IF;\n");
		proc.append("END");
		sql.add(proc.toString());

		// Now create the trigger to call the new procedure
		addTrigger(source, "AFTER INSERT",
				"CALL " + quoteId(insConName) + "("+EasikTools.join(", ", params)+")");
	}

	public List<String> createConstraint(EqualizerConstraint constraint, String id) {
		LinkedList<String> sql = new LinkedList<String>();

		EntityNode eq = constraint.getEqualizerEntity(),
				   source = constraint.getSourceEntity(),
				   target = constraint.getTargetEntity();
		InjectiveEdge projEdge = (InjectiveEdge) constraint.getProjection().getFirstEdge();
		List<SketchPath> paths = constraint.getEqualizerPaths();

		String insConName = "eqCon" + id + "InsUpd" + cleanId(source);

		LinkedList<String> args = new LinkedList<String>(), params = new LinkedList<String>();
		args.add("_newId " + pkType());
		params.add("NEW." + quoteId(tablePK(source)));

		for (SketchEdge shadow: source.getShadowEdges()) {
			args.add(quoteId("NEW_shadow_" + tableFK(shadow)) + " " + pkType());
			params.add("NEW." + quoteId(tableFK(shadow)));
		}

		StringBuilder proc = new StringBuilder();
		proc.append("CREATE PROCEDURE " + quoteId(insConName) + "(" + EasikTools.join(", ", args) + ") BEGIN\n");
		proc.append("	DECLARE _path0Id");
		// Add a variable for each path, _path0Id, _path1Id, etc.:
		for (int i = 1; i < paths.size(); i++) {
			proc.append(", _path" + i + "Id");
		}
		proc.append(" " + pkType() + ";\n");
		String sourcePK = qualifiedPK(source),
			   targetPK = qualifiedPK(target),
			   newPK = "_newId";
		int i = 0;
		for (SketchPath p: paths) {
			proc.append(
				"	SELECT " + targetPK + " INTO _path"+i+"Id\n" +
				"		FROM " + joinPath(p) + "\n" +
				"		WHERE " + sourcePK + " = " + newPK + ";\n");
			++i;
		}
		// Build equality clause, such as: _path0Id IS DISTINCT FROM _path1Id OR _path0Id IS DISTINCT FROM _path2Id OR ...
		proc.append("	IF NOT (_path0Id <=> _path1Id)");
		for (int j = 2; j < paths.size(); j++)
			proc.append(" OR NOT (_path0Id <=> _path"+j+"Id)");

		proc.append(
			" THEN\n"+
			// If one of the paths doesn't match, we need to need to delete it from the equalizer (if it's there)
			"		DELETE FROM "+quoteId(eq)+" WHERE "+qualifiedFK(projEdge)+" = "+newPK+";\n"+
			"	ELSEIF (SELECT COUNT(*) FROM "+quoteId(eq)+" WHERE "+qualifiedFK(projEdge)+" = "+newPK+") = 0 THEN\n"+
			// Otherwise, if it isn't already in the equalizer, we need to add it:
			"		");
		insertInto(proc, true, eq, source.getShadowEdges(), null, null, Collections.singletonList(quoteId(tableFK(projEdge))), Collections.singletonList(newPK));
		proc.append(
			"	END IF;\n"+
			"END\n");

		sql.add(proc.toString());

		// Create the trigger for inserting into the source table:
		String call = "CALL " + quoteId(insConName) + "(" + EasikTools.join(", ", params) + ")";
		addTrigger(source, "AFTER INSERT", call);
		addTrigger(source, "AFTER UPDATE", call);

		// If the projection isn't a cascading edge, cascade:
		if (projEdge.getCascading() != SketchEdge.Cascade.CASCADE)
			addTrigger(source, "BEFORE DELETE", "DELETE FROM "+quoteId(eq)+" WHERE "+qualifiedFK(projEdge)+" = OLD."+quoteId(tablePK(source)));

		return delimit("$$", sql);
	}

	public List<String> createConstraint(SumConstraint constraint, String id) {

		LinkedList<String> sql = new LinkedList<String>();

		for (SketchPath path: constraint.getPaths()) {
			EntityNode dom = path.getDomain(), codo = path.getCoDomain();
			String baseConName = "sumConstraint" + id + cleanId(dom);
			StringBuilder proc = new StringBuilder();
			proc.append(
				"CREATE PROCEDURE " + quoteId(baseConName + "Delete") + "(_deleteFK "+pkType()+") BEGIN\n" +
				"	DELETE FROM " + quoteId(codo) + " WHERE " + qualifiedPK(codo) + " = ");
			if (path.getEdges().size() == 1) {
				proc.append("_deleteFK");
			}
			else {
				LinkedList<SketchEdge> trailing = new LinkedList<SketchEdge>(path.getEdges());
				trailing.removeFirst();
				proc.append("(SELECT " + qualifiedFK(trailing.getLast()) + " FROM " + joinPath(trailing, false) + " WHERE " + qualifiedPK(path.getFirstCoDomain()) + " = _deleteFK)");
			}
			proc.append(";\nEND");
			sql.add(proc.toString());
			addTrigger(dom, "AFTER DELETE",
					"CALL " + quoteId(baseConName + "Delete") + "(OLD." + quoteId(tableFK(path.getFirstEdge())) + ")");

			LinkedList<String> insertions = new LinkedList<String>();
			proc = new StringBuilder();
			insertInto(proc, false, codo, dom.getShadowEdges(), null, null, null, null);
			insertions.add(proc.toString());

			SketchEdge[] edges = path.getEdges().toArray(new SketchEdge[0]);
			for (int i = edges.length - 1; i > 0; i--) {
				proc = new StringBuilder();
				EntityNode source = edges[i].getSourceEntity();
				insertInto(proc, false, source, dom.getShadowEdges(),
						null, null,
						Collections.singletonList(quoteId(tableFK(edges[i]))),
						Collections.singletonList("LAST_INSERT_ID()"));
				insertions.add(proc.toString());
			}
			insertions.add("SET NEW." + quoteId(tableFK(edges[0])) + " = LAST_INSERT_ID()");
			addTrigger(dom, "BEFORE INSERT", insertions);
		}

		return delimit("$$", sql);
	}

	/**
	 * Creates various extra statements that we need.
	 *
	 * One of these is to create a constraint_failure() procedure, since MySQL
	 * has no way to abort a transaction other than performing an invalid query
	 * (constraint_failure() attempts an "UPDATE" of "fail = 1" on a table
	 * named whatever failure message is passed in; that way at least the error
	 * appears in the MySQL error message.  This sucks, I know.  Blame MySQL.)
	 *
	 * FIXME: Right now, MySQL <b>does not support</b> updating the same table
	 * a trigger applies to in an AFTER trigger, so we have no way to clear
	 * shadow columns.  This was presumably a result of MySQL incompetence, or
	 * MySQL's philosophy of limiting itself in order to prevent idiot PHP
	 * programmers from shooting themselves in the foot with--*GASP*--a
	 * recursive trigger.  While it would be very nice to clear the shadow
	 * columns, we simply can't do so: the values are often needed in AFTER
	 * INSERT or AFTER UPDATE triggers, so we can't clear them before then,
	 * however the only entry point for clearing these with MySQL is doing a
	 * 'SET NEW.col = NULL' inside the BEFORE trigger.
	 *
	 * If MySQL some day supports an update on the same table a trigger is
	 * applied to, the following will work.  Until then, MySQL tables will end
	 * up retaining values in the shadow columns.  N.B. that these values are
	 * actual references to the appropriate tables and columns, but are 'ON
	 * DELETE SET NULL ON UPDATE SET NULL', so should not pose data
	 * manipulation problems.
	 */
	public List<String> createExtras() {
		LinkedList<String> extras = new LinkedList<String>();
		if (addFail) {
			// This hack is needed for MySQL to abort an insert if a trigger fails.
			// This is disgusting, but is what MySQL documentation recommends doing.
			extras.addAll(delimit("$$",
				"CREATE PROCEDURE constraint_failure(_message VARCHAR(255)) BEGIN\n" +
				"	-- This update is going to fail: this hack is needed because MySQL\n" +
				"	-- lacks the ability to do an (SQL-standard) SIGNAL from a procedure.\n" +
				"	SET @sql = CONCAT('UPDATE `', _message, '` SET fail=1');\n" +
				"	PREPARE constraint_fail_statement_handle FROM @sql;\n" +
				"	EXECUTE contraint_fail_statement_handle;\n" +
				"	DEALLOCATE PREPARE contraint_fail_statement_handle;\n" +
				"END"));
		}

		LinkedList<StringBuilder> trigs = new LinkedList<StringBuilder>();

		/*
		for (EntityNode table: sketch.getEntities()) {
			// Add the triggers for clearing the shadow edges
			HashSet<SketchEdge> shadowEdges = table.getShadowEdges();
			if (!shadowEdges.isEmpty()) {
				LinkedList<String> sets = new LinkedList<String>();
				LinkedList<String> checks = new LinkedList<String>();
				for (SketchEdge edge: shadowEdges) {
					sets.add(quoteId(tableFK(edge)) + " = NULL");
					checks.add("NEW."+quoteId(tableFK(edge)) + " IS NOT NULL");
				}

				String clear = "IF " + EasikTools.join(" OR ", checks) + " THEN " +
					"UPDATE " + quoteId(table) + " SET " + EasikTools.join(", ", sets) + " WHERE " + quoteId(tablePK(table)) + " = NEW." + quoteId(tablePK(table)) + "; " +
					"END IF";
				addTrigger(table, "AFTER INSERT", clear);
				addTrigger(table, "AFTER UPDATE", clear);
			}
		}
		*/

		for (EntityNode table: triggers.keySet()) {
			LinkedHashMap<String, LinkedList<String>> tableTriggers = triggers.get(table);
			for (String when: tableTriggers.keySet()) {
				LinkedList<String> actions = new LinkedList<String>(tableTriggers.get(when));
				StringBuilder commands = new StringBuilder();
				commands.append(
						"CREATE TRIGGER " + quoteId(table + "_" + when.replaceAll("(\\w)\\w*\\s*", "$1").toLowerCase() + "Trig") +
						" " + when + " ON " + quoteId(table) + " FOR EACH ROW\n");
				if (actions.size() == 1) {
					commands.append(actions.getFirst());
				}
				else {
					commands.append("BEGIN\n");
					for (String action: actions)
						commands.append("\t" + action.replaceFirst(";\\s*$", "") + ";\n");

					commands.append("END");
				}

				trigs.add(commands);
			}
		}

		if (!trigs.isEmpty())
			extras.addAll(delimit("$$", trigs));

		return extras;
	}

	/**
	 * Adds a trigger to triggers, which is used in createExtras to add the triggers.
	 *
	 * @param table the table the trigger applies to
	 * @param when the trigger time, such as "BEFORE UPDATE", "AFTER DELETE", "BEFORE INSERT", etc.
	 * @param action one or more SQL statements (or array or list of SQL statements) to execute in the
	 * 		trigger (statements should not end with ; or $).
	 */
	// triggers is a: LinkedHashMap<EntityNode, LinkedHashMap<String, LinkedList<String>>>
	//                              table                     when               action
	private void addTrigger(EntityNode table, String when, String... action) {
		addTrigger(table, when, Arrays.asList(action));
	}
	private void addTrigger(EntityNode table, String when, List<String> action) {
		LinkedHashMap<String, LinkedList<String>> tableTriggers;
		if (triggers.containsKey(table))
			tableTriggers = triggers.get(table);
		else
			triggers.put(table, tableTriggers = new LinkedHashMap<String, LinkedList<String>>());

		when = when.toUpperCase();
		LinkedList<String> actions;
		if (tableTriggers.containsKey(when))
			actions = tableTriggers.get(when);
		else
			tableTriggers.put(when, actions = new LinkedList<String>());

		actions.addAll(action);
	}

	/**
	 * Takes statement(s) (typically of a procedure) and, if for a non-database
	 * export, adds the delimiter change statement and adds the specified
	 * delimiter to the end of each provided statement.
	 */
	private List<String> delimit(String delimiter, List<? extends CharSequence> statements) {
		LinkedList<String> sql = new LinkedList<String>();
		if (mode == Mode.DATABASE) {
			for (CharSequence st: statements)
				sql.add(st.toString());
		}
		else {
			sql.add("DELIMITER " + delimiter);
			for (CharSequence st: statements)
				sql.add(st + delimiter);
			sql.add("DELIMITER ;");
		}
		return sql;
	}
	private List<String> delimit(String delimiter, CharSequence... statements) {
		return delimit(delimiter, Arrays.asList(statements));
	}

	private String pkType() {
		return optionEnabled("bigKeys") ? "BIGINT" : "INTEGER";
	}

	/**
	 * Appends an INSERT INTO statement to the provided StringBuilder, and stores
	 * the shadow edges to be cleared at the end of the trigger.  Note that MySQL
	 * currently has limitations preventing us from clearing shadow edges; see the
	 * createExtra() documentation for why.
	 *
	 * @param proc the StringBuilder into which the INSERT statement will be
	 *     appended
	 * @param inProcedure true if the INSERT is in a procedure called by a trigger, false if the INSERT is
	 *     to be used directly inside the trigger.  If in a procedure, NEW_shadow_shadowCol will be used (and
	 *     assumed passed in to the procedure); when used within a trigger the row value will be accessed
	 *     using NEW.shadowCol.  This shouldn't be necessary, but MySQL doesn't support passing a row to a
	 *     procedure.
	 * @param node the EntityNode of the table the trigger applies to
	 * @param inShadowEdges the shadowEdges available from the original insert, which may not have been into node
	 * @param selectWhat if non-null, this generates a "INSERT INTO ... SELECT <i>selectWhat</i> FROM <i>selectFrom</i>"
	 *     statement, with appropriate shadow columns added.
	 * @param selectFrom see the selectWhat parameter.
	 * @param cols if selectWhat is non-null, this is a list of column names that are being inserted (not including
	 *     the shadow columns).  This will be treated as an empty list if null.
	 * @param vals if selectWhat is non-null, this is a list of values associated with the names provided by cols
	 *     that are to be inserted.  This will be treated as an empty list if null.
	 */
	private void insertInto(
			StringBuilder proc,
			boolean inProcedure, // If we're producing code for direct inclusion in a trigger
			EntityNode node,
			Collection<SketchEdge> inShadowEdges,
			String selectWhat,
			String selectFrom,
			List<String> cols,
			List<String> vals) {
		HashSet<SketchEdge> shadowCols = new HashSet<SketchEdge>();
		if (cols == null) cols = new LinkedList<String>();
		if (vals == null) vals = new LinkedList<String>();
		LinkedList<SketchEdge> shadowEdges = node.getNonPartialEdges();
		shadowEdges.addAll(node.getShadowEdges());
		shadowEdges.retainAll(inShadowEdges);
		proc.append("INSERT INTO " + quoteId(node) + " ");
		proc.append("(");
		StringBuilder values = new StringBuilder();
		for (String col: cols)
			proc.append(col + ", ");
		for (String val: vals)
			values.append(val + ", ");

		for (SketchEdge e: shadowEdges) {
			proc.append(quoteId(tableFK(e)) + ", ");
			if (inProcedure)
				values.append(quoteId("NEW_shadow_" + tableFK(e)) + ", ");
			else
				values.append("NEW." + quoteId(tableFK(e)) + ", ");

			shadowCols.add(e);
		}
		// Remove the last ", " (unless both are empty, since then we didn't add any):
		if (!(cols.isEmpty() && shadowEdges.isEmpty()))
			proc.delete(proc.length()-2, proc.length());

		// And remove it from values, if it has anything:
		if (values.length() > 0)
			values.delete(values.length()-2, values.length());

		if (selectWhat != null) {
			proc.append(") SELECT " + selectWhat + (values.length() > 0 ? ", " + values : "") + " FROM " + selectFrom);
		}
		else {
			proc.append(") VALUES (" + values + ")");
		}
		proc.append(";\n");
	}


	public List<String> createView(ViewNode node){
		LinkedList<String> viewSQL = new LinkedList<String>();
		
		for( QueryNode qn : node.getFrame().getView().getQueryNodes() ){
			viewSQL.add(createView(qn));
		}

		return viewSQL;
	}

	public String createView(QueryNode node){
		return "CREATE VIEW " + node.getName() + " AS " + node.getQuery() + $;
	}

	/**
	 * Simple wrapper around the driver's cleanId method.
	 * @see easik.database.driver.MySQL.cleanId(Object)
	 */
	public String cleanId(Object id) { return dbd.cleanId(id); }

	/**
	 * Simple wrapper around the driver's quoteId method.
	 * @see easik.database.driver.MySQL.quoteId(Object)
	 */
	public String quoteId(Object id) { return dbd.quoteId(id); }
}
