package easik.database.export;

import easik.EasikTools;
import easik.sketch.Sketch;
import easik.database.DatabaseDriver;
import easik.database.export.SketchExporter;
import easik.database.types.*;
import easik.sketch.vertex.EntityNode;
import easik.sketch.attribute.EntityAttribute;
import easik.sketch.edge.*;
import easik.sketch.path.SketchPath;
import easik.sketch.keys.*;
import easik.sketch.constraint.*;
import easik.view.vertex.*;
import easik.overview.vertex.*;

import java.io.*;

import java.util.Map;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.LinkedList;
import java.util.LinkedHashMap;
import java.util.Collection;
import java.util.Collections;
import java.util.Random;

import java.sql.SQLException;


/**
 * EASIK PostgreSQL sketch export driver.  This driver is loaded by calling getSketchExporter()
 * on a PostgreSQL database driver object.  The following options passed in to the
 * getSketchExporter() options parameter, in addition to those mentioned in SketchExporter, are
 * supported by this export driver:
 *
 * <ul>
 * 	<li><code>createDatabase</code> &mdash; if set to <code>"true"</code>, <code>CREATE DATABASE</code>
 * 		statements will be added to the SQL dump <i>[{@link exportToString()} and {@link exportToFile(File)} only]</i></li>
 * 	<li><code>dropDatabase</code> &mdash; if set to <code>"true"</code>, and using the above
 * 		<code>createDatabase</code> option, <code>DROP DATABASE</code> statements will be added to
 * 		the SQL dump <i>[exportToString() and exportToFile() only]</i></li>
 * 	<li><code>dropSchema</code> &mdash; if set to the string <code>"true"</code>, and using a <code>schema</code>
 * 		(@see easik.database.driver.PostgreSQL) the schema of the specified name will be dropped and recreated
 * 		before exporting tables.</li>
 * 	<li><code>bigKeys</code> &mdash; if set to the string <code>"true"</code>, the generated tables will use
 * 		BIGSERIAL instead of SERIAL for primary keys, and BIGINT instead of INT for foreign key columns.  Use this
 * 		if your data could conceivably ever go through more than 2 billion rows.</li>
 * </ul>
 *
 * @see easik.database.export.SketchExporter
 * @see easik.database.driver.PostgreSQL
 */
public class PostgreSQL extends SketchExporter {

	public PostgreSQL(Sketch sketch, DatabaseDriver db, Map<String, ?> options) throws DatabaseDriver.LoadException {
		super(sketch, db, options);
		if (!db.hasOption("database"))
			throw new DatabaseDriver.LoadException("PostgreSQL export driver requires \"database\" parameter!");
	}

	public List<String> initialize() {
		// super.initialize() gives us the current date/time and Easik blurb in a comment
		LinkedList<String> sql = new LinkedList<String>(super.initialize());

		String dbNameQ = quoteId(dbd.getOption("database"));

		// If we're working on a live connection, don't try this drop/create stuff: on
		// almost any secured pgsql setup, it won't work anyway.
		if (mode != Mode.DATABASE && optionEnabled("createDatabase")) {
			if (optionEnabled("dropDatabase"))
				sql.add("DROP DATABASE " + dbNameQ+$);

			sql.add("CREATE DATABASE " + dbNameQ+$);

			sql.add("\\connect " + dbNameQ);
		}

		if (dbd.hasOption("schema")) {
			String schemaQ = quoteId(dbd.getOption("schema"));

			if (optionEnabled("dropSchema"))
				sql.add("DROP SCHEMA IF EXISTS " + schemaQ + " CASCADE"+$);

			sql.add("CREATE SCHEMA " + schemaQ+$);

			sql.add("SET search_path = " + schemaQ+$);
		}

		// This might fail (for instance, if the language already exists), but that isn't necessarily
		// a fatal error (unfortunately, postgresql doesn't, as of 8.3, support CREATE OR REPLACE LANGUAGE)
		sql.add("CREATE LANGUAGE plpgsql"+$);

		return sql;
	}

	/**
	 * Some things in initialize are allowed to fail; if we get an exception
	 * during the initialize queries, it gets passed here, and if this returns
	 * true, we ignore that exception and keep going.
	 *
	 * @param e the SQLException that occurred 
	 * @return true, if the SQL exception should be ignored, false if it is a proper error.
	 */
	protected boolean ignoreError(SQLException e) {
		if (e.getMessage().toLowerCase().contains("language \"plpgsql\" already exists"))
			return true;
		else if (e.getMessage().toLowerCase().matches(".*schema .* already exists.*"))
			return true;
		return false;
	}

	public List<String> createTable(EntityNode table, boolean includeRefs) {
		StringBuilder tableDef = new StringBuilder();
		tableDef.append("CREATE TABLE " + quoteId(table) + " (\n");

		// The primary key: either a SERIAL or a BIGSERIAL.  These are pg pseudo-types that combine
		// an INT/BIGINT with an automatically-created sequence and make the default column value
		// the next sequence value (this accomplishes the same thing as MySQL's AUTO_INCREMENT
		// property, but in a more general way).
		tableDef.append("	" + quoteId(tablePK(table)) + " " + (optionEnabled("bigKeys") ? "BIGSERIAL" : "SERIAL") + " PRIMARY KEY");

		for (EntityAttribute attr: table.getEntityAttributes()) {
			// FIXME -- what about NOT NULL, etc. on the column?  Defaults?
			tableDef.append(",\n	" + quoteId(attr.getName()) + " " + dbd.getTypeString(attr.getType()));
		}

		// Include the foreign key columns, and the actual foreign keys (if includeRefs set)
		for (SketchEdge edge: table.getOutgoingEdges()) {
			tableDef.append(",\n	" + quoteId(tableFK(edge)) + " "+pkType());
			if (!edge.isPartial())
				tableDef.append(" NOT NULL");
			if (edge.isInjective())
				tableDef.append(" UNIQUE");
			if (includeRefs) {
				tableDef.append(" REFERENCES " + quoteId(edge.getTargetEntity()) + " (" +
						quoteId(tablePK(edge.getTargetEntity())) + ")");
				String refMode =
					edge.getCascading() == SketchEdge.Cascade.SET_NULL ? "SET NULL" :
					edge.getCascading() == SketchEdge.Cascade.CASCADE ? "CASCADE" :
					"NO ACTION"; // Effectively the same as RESTRICT, but RESTRICT won't even work if cleaned up in the same transaction
				tableDef.append(" ON DELETE " + refMode + " ON UPDATE " + refMode);
			}
		}

		if (includeRefs) {
			// We need a copy of any sum tables' not-null columns (which are simply the
			// outgoing, non-partial edges) if this table is a summand; we shadow the
			// columns here, then handle propagating them in the SumConstraint triggers.
			for (SketchEdge edge: table.getShadowEdges()) {
				tableDef.append(",\n	" + quoteId(tableFK(edge)) + " "+pkType());
				tableDef.append(",\n	FOREIGN KEY (" + quoteId(tableFK(edge)) +
						") REFERENCES " + quoteId(edge.getTargetEntity()) + " (" +
						quoteId(tablePK(edge.getTargetEntity())) + ") ON DELETE SET NULL ON UPDATE SET NULL");
			}
		}

		tableDef.append("\n)"+$);

		LinkedList<String> statements = new LinkedList<String>();
		statements.add(tableDef.toString());

		// Add unique keys; we have to include this after because we can't specify a name
		// when adding a unique index as part of the CREATE TABLE statement.
		for (UniqueKey key: table.getUniqueKeys()) {
			LinkedList<String> keyCols = new LinkedList<String>();
			for (UniqueIndexable elem: key.getElements()) {
				if (elem instanceof SketchEdge) {
					keyCols.add(quoteId(tableFK((SketchEdge) elem)));
				}
				else { // Probably an attribute
					keyCols.add(quoteId(elem.getName()));
				}
			}
			statements.add(
					"CREATE UNIQUE INDEX " + quoteId(table.getName() + "_" + key.getKeyName()) +
					" ON " + quoteId(table) + " (" + EasikTools.join(", ", keyCols) + ")"+$);
		}

		return statements;
	}

	public List<String> createReferences(EntityNode table) {
		LinkedList<String> refs = new LinkedList<String>();

		String tableQ = quoteId(table);
		for (SketchEdge edge: table.getOutgoingEdges()) {
			EntityNode target = edge.getTargetEntity();
			String refMode =
				edge.getCascading() == SketchEdge.Cascade.SET_NULL ? "SET NULL" :
				edge.getCascading() == SketchEdge.Cascade.CASCADE ? "CASCADE" :
				"NO ACTION"; // Effectively the same as RESTRICT, but RESTRICT won't even work if cleaned up in the same transaction
			refs.add(
				"ALTER TABLE " + tableQ + " ADD FOREIGN KEY (" + quoteId(tableFK(edge)) + ") " +
				"REFERENCES " + quoteId(target) + " (" + quoteId(tablePK(target)) + ") " +
				"ON DELETE " + refMode + " ON UPDATE " + refMode + $);
		}

		// We need a copy of any sum tables' not-null columns (which are simply the
		// outgoing, non-partial edges) if this table is a summand; we shadow the
		// columns here, then handle propagating them in the SumConstraint triggers.
		for (SketchEdge edge: table.getShadowEdges()) {
			refs.add("ALTER TABLE " + tableQ + " ADD COLUMN " + quoteId(tableFK(edge)) + " "+pkType());
			refs.add("ALTER TABLE " + tableQ + " ADD FOREIGN KEY (" + quoteId(tableFK(edge)) +
					") REFERENCES " + quoteId(edge.getTargetEntity()) + " (" + quoteId(tablePK(edge.getTargetEntity())) + ") " +
					"ON DELETE SET NULL ON UPDATE SET NULL");
		}

		return refs;
	}

	public List<String> createConstraint(CommutativeDiagram cd, String id) {
		LinkedList<String> sql = new LinkedList<String>();
		List<SketchPath> paths = cd.getPaths();
		EntityNode dom = paths.get(0).getDomain();

		StringBuilder proc = new StringBuilder("\n	DECLARE\n");
		int targetNum = 0;
		for (SketchPath path: paths) {
			++targetNum;
			LinkedList<SketchEdge> tmpPath = new LinkedList<SketchEdge>(path.getEdges());
			tmpPath.removeFirst();

			String fk = quoteId(tableFK(path.getFirstEdge()));

			proc.append("		_cdTarget" + targetNum + " CONSTANT " + pkType() + " := " +
					(tmpPath.size() == 0 ? "NEW."+fk :
						"(SELECT " + qualifiedFK(path.getLastEdge()) + " FROM " + joinPath(tmpPath, false) +
						" WHERE " + qualifiedPK(tmpPath.getFirst().getSourceEntity()) + " = NEW." + fk + ")") +
					";\n");
		}

		proc.append(
				"	BEGIN\n" +
				"		IF\n");
		for (int i = 2; i <= targetNum; i++) {
			proc.append("			_cdTarget1 IS DISTINCT FROM _cdTarget"+i);
			if (i < targetNum)
				proc.append(" OR");
			proc.append("\n");
		}

		proc.append(
				"		THEN RAISE EXCEPTION 'Commutative diagram constraint failure';\n" +
				"		END IF;\n" +
				"		RETURN NEW;\n" +
				"	END;\n");

		sql.add(
			"CREATE FUNCTION " + quoteId("commutativeDiagram" + id) + "() RETURNS trigger AS " +
			quoteValue(proc.toString(), "commutativeDiagram"+id) +
			" LANGUAGE plpgsql"+$);


		sql.add("CREATE TRIGGER " + quoteId("commutativeDiagram" + id) + " BEFORE INSERT ON " + quoteId(dom) + "\n" +
			   "	FOR EACH ROW EXECUTE PROCEDURE " + quoteId("commutativeDiagram" + id) + "()"+$);

		return sql;
	}

	public List<String> createConstraint(ProductConstraint constraint, String id) {
		LinkedList<String> sql = new LinkedList<String>();

		String delConName = "prodConstraint" + id + "Del";

		StringBuilder proc = new StringBuilder("\nBEGIN\n");
		EntityNode begin = null;
		for (SketchPath p: constraint.getPaths()) {
			if (begin == null) begin = p.getDomain();
			EntityNode codo = p.getCoDomain();
			String pk = quoteId(tablePK(codo));
			proc.append(
				"	DELETE FROM " + quoteId(codo) + " WHERE " + pk + " IN (\n" +
				"		SELECT " + qualifiedFK(p.getLastEdge()) + "\n" +
				"			FROM " + joinPath(p, false) + "\n" +
				"			WHERE " + qualifiedPK(begin) + " = OLD." + quoteId(tablePK(begin)) + "\n" +
				"	);\n");
		}
		proc.append("END\n");

		sql.add("CREATE FUNCTION " + quoteId(delConName) + "() RETURNS trigger AS " +
				quoteValue(proc.toString(), delConName) + " LANGUAGE plpgsql"+$);

		// Now create the trigger to call the new procedure
		sql.add(
			"CREATE TRIGGER " + quoteId(delConName) + " BEFORE DELETE ON " + quoteId(begin) + "\n" +
			"	FOR EACH ROW EXECUTE PROCEDURE " + quoteId(delConName) + "()"+$);

		for (SketchPath p: constraint.getPaths()) {
			EntityNode dest = p.getCoDomain();
			String conName = "prodConstraint" + id + "Ins" + cleanId(dest);

			proc = new StringBuilder("\nDECLARE\n	_lastId " + pkType() + ";\nBEGIN\n");

			StringBuilder createIntermediates = new StringBuilder();

			// Clauses stores out IF clauses that let us figure out if the tables are non-empty
			ArrayList<String> clauses = new ArrayList<String>();
			for (SketchPath q: constraint.getPaths()) {
				if (p == q) continue;
				// Older versions were doing "SELECT COUNT(*) FROM TABLE", but that has a big
				// performance hit for very large, busy tables, when all we really care
				// about is non-empty, so we just select 1 with a LIMIT instead:
				clauses.add("(SELECT 1 FROM " + quoteId(q.getCoDomain()) + " LIMIT 1) = 1");

				// If we end up inserting anything in createIntermediates, we'll use these
				// insertions when we insert the first row into one of the path targets
				SketchEdge[] edges = q.getEdges().toArray(new SketchEdge[0]);
				for (int i = edges.length - 1; i > 0; i--) {
					SketchEdge e = edges[i];
					EntityNode source = e.getSourceEntity(), target = e.getTargetEntity();
					createIntermediates.append("			");
					insertInto(
							createIntermediates, source, dest.getShadowEdges(), qualifiedPK(target), quoteId(target),
							Collections.singletonList(quoteId(tableFK(e))), null);
				}
			}

			// In words: If the tables forming the domains of the other paths are non-empty.
			proc.append("	IF " + EasikTools.join(" AND ", clauses) + " THEN\n");

			if (createIntermediates.length() > 0) {
				// If we just inserted the first row, we're going to have to build a path of intermediate tables
				// for the other paths, which we built in createIntermediate.
				proc.append("		IF (SELECT COUNT(*) FROM (SELECT 1 FROM " + quoteId(dest) + " LIMIT 2) a) = 1 THEN\n");
				proc.append(createIntermediates);
				proc.append("		END IF;\n\n");
			}

			// Produce the intermediate path insertion strings
			SketchEdge[] edges = p.getEdges().toArray(new SketchEdge[0]);
			if (edges.length > 1) {
				for (int i = edges.length - 1; i > 0; i--) {
					SketchEdge e = edges[i];
					EntityNode source = e.getSourceEntity(), target = e.getTargetEntity();
					proc.append("		");
					LinkedList<String> cols = new LinkedList<String>(), vals = new LinkedList<String>();
					cols.add(quoteId(tableFK(e)));
					vals.add(i == edges.length-1 ? "NEW."+quoteId(tablePK(target)) : currval(target));
					insertInto(proc, source, dest.getShadowEdges(), null, null, cols, vals);
				}
				proc.append("		_lastId := " + currval(edges[0].getTargetEntity()) + ";\n");
			}
			else {
				proc.append("		_lastId := NEW." + tablePK(dest) + ";\n");
			}


			// Now the proper insertion
			ArrayList<String> columns = new ArrayList<String>();
			ArrayList<String> values = new ArrayList<String>();
			ArrayList<String> from = new ArrayList<String>();
			EntityNode thisTarget = null;
			for (SketchPath q: constraint.getPaths()) {
				EntityNode target = q.getFirstEdge().getTargetEntity();
				columns.add(quoteId(tableFK(q.getFirstEdge())));
				values.add(qualifiedPK(target));
				from.add(quoteId(target));
				if (q == p) thisTarget = target;
			}
			proc.append("		");
			insertInto(proc, begin, dest.getShadowEdges(),
					EasikTools.join(", ", values),
					EasikTools.join(" CROSS JOIN ", from) + " WHERE " + qualifiedPK(thisTarget) + " = _lastId",
					columns, null);

			proc.append("	END IF;\n");
			proc.append("	RETURN NULL;\nEND\n");

			sql.add("CREATE FUNCTION " + quoteId(conName) + "() RETURNS trigger AS " +
				quoteValue(proc.toString(), conName) + " LANGUAGE plpgsql"+$);

			// Now create the trigger to call the new procedure
			sql.add(
				"CREATE TRIGGER " + quoteId(conName) + " AFTER INSERT ON " + quoteId(dest) + "\n" +
			   	"	FOR EACH ROW EXECUTE PROCEDURE " + quoteId(conName) + "()"+$);
		}

		return sql;
	}

	public List<String> createConstraint(PullbackConstraint constraint, String id) {
		LinkedList<String> sql = new LinkedList<String>();

		EntityNode pb = constraint.getSource(), target = constraint.getTarget();

		SketchPath projA = constraint.getProjectionA(), projB = constraint.getProjectionB();
		SketchPath targetA = constraint.getTargetPathA(), targetB = constraint.getTargetPathB();
		SketchPath pathA = constraint.getFullPathA(), pathB = constraint.getFullPathB();

		String insConName = "pullbackCon" + id + "Ins" + cleanId(pb);

		StringBuilder proc = new StringBuilder("\n" +
				"DECLARE\n" +
				"	_pathAId " + pkType() + ";\n" +
				"	_pathBId " + pkType() + ";\n" +
				"BEGIN\n");
		String basePK = qualifiedPK(pb),
			   targetPK = qualifiedPK(target),
			   newPK = "NEW." + quoteId(tablePK(pb)),
			   oldPK = "OLD." + quoteId(tablePK(pb));
		proc.append(
			"	SELECT " + targetPK + " INTO _pathAId\n" +
			"		FROM " + joinPath(pathA) + "\n" +
			"		WHERE " + basePK + " = " + newPK + ";\n");
		proc.append(
			"	SELECT " + targetPK + " INTO _pathBId\n" +
			"		FROM " + joinPath(pathB) + "\n" +
			"		WHERE " + basePK + " = " + newPK + ";\n");
		proc.append( // Make sure the two paths match
			"	IF _pathAId IS DISTINCT FROM _pathBId THEN\n" +
			"		RAISE EXCEPTION 'Invalid entry in pullback table.';\n" +
			"	END IF;\n");

		String pbQ = quoteId(pb);
		SketchEdge edgeA = pathA.getFirstEdge(), edgeB = pathB.getFirstEdge();
		String colA = qualifiedFK(edgeA), colB = qualifiedFK(edgeB);
		// Make sure there are no duplicate entries; we do this by adding a unique key on
		// colA and colB of the pullback table (unless one already exists)
		if (!addedUnique.isUnique(pb, edgeA, edgeB)) {
			sql.add("ALTER TABLE "+pbQ+" ADD UNIQUE ("+quoteId(tableFK(edgeA))+", "+quoteId(tableFK(edgeB))+")"+$);
			addedUnique.add(pb, edgeA, edgeB);
		}

		proc.append("	RETURN NULL;\nEND\n");

		sql.add("CREATE FUNCTION " + quoteId(insConName) + "() RETURNS trigger AS " + quoteValue(proc.toString(), insConName) +
				" LANGUAGE plpgsql"+$);

		// Create the trigger for inserting into the pullback table:
		sql.add("CREATE TRIGGER " + quoteId(insConName) + " AFTER INSERT ON " + pbQ + "\n" +
			   	"	FOR EACH ROW EXECUTE PROCEDURE " + quoteId(insConName) + "()"+$);

		List<SketchPath> paths = constraint.getPaths();
		pbAfterInsertProcs(sql, id, constraint, targetA, projA, targetB, projB);
		pbAfterInsertProcs(sql, id, constraint, targetB, projB, targetA, projA);

		// Finally, the before-delete on the pullback:
		String delConName = "pullbackCon" + id + "Del" + cleanId(pb);
		proc = new StringBuilder("\nBEGIN\n");

		proc.append(
			"	DELETE FROM " + quoteId(target) + " WHERE\n" +
			"		" + targetPK + " = (SELECT " + qualifiedFK(pathA.getLastEdge()) + " FROM " + joinPath(pathA, false) + " WHERE " + basePK + " = " + oldPK + ")\n" +
			"			AND\n" +
			"		" + targetPK + " = (SELECT " + qualifiedFK(pathB.getLastEdge()) + " FROM " + joinPath(pathB, false) + " WHERE " + basePK + " = " + oldPK + ");");

		proc.append("\n	RETURN NULL;\nEND\n");

		sql.add("CREATE FUNCTION " + quoteId(delConName) + "() RETURNS trigger AS " +
			quoteValue(proc.toString(), delConName) + " LANGUAGE plpgsql"+$);

		sql.add("CREATE TRIGGER " + quoteId(delConName) + " BEFORE DELETE ON " + pbQ + " FOR EACH ROW EXECUTE PROCEDURE " + quoteId(delConName) + "()"+$);

		return sql;
	}

	private void pbAfterInsertProcs(List<String> sql, String id, PullbackConstraint constraint, SketchPath thisPath, SketchPath thisFirstPath, SketchPath otherPath, SketchPath otherFirstPath) {
		EntityNode source = thisPath.getDomain(), otherSource = otherPath.getDomain(),
			pb = thisFirstPath.getDomain(), target = thisPath.getCoDomain();

		String insConName = "pullbackCon" + id + "Ins" + cleanId(source);
		String thisJoinPath = joinPath(thisPath, false);

		StringBuilder proc = new StringBuilder("\nBEGIN\n");
		String newId = "NEW." + quoteId(tablePK(source));
		String lastJoinWhere = 
				// This part of the WHERE clause is to join the paths to the target table
				qualifiedFK(thisPath.getLastEdge()) + " = " + qualifiedPK(target) + " AND " +
				qualifiedFK(otherPath.getLastEdge()) + " = " + qualifiedPK(target);

		proc.append("	IF (SELECT COUNT(*) FROM " + thisJoinPath + ", " + joinPath(otherPath, false) +
				", " + quoteId(target) + " WHERE " + lastJoinWhere + " AND " +
				// And this connects the two paths to the source value:
				qualifiedPK(source) + " = " + newId + ") > 0 THEN\n");

		SketchEdge[] edges = otherFirstPath.getEdges().toArray(new SketchEdge[0]);
		if (edges.length >= 2) {
			ArrayList<SketchEdge> fullEdges = new ArrayList<SketchEdge>(otherFirstPath.getEdges());
			fullEdges.addAll(otherPath.getEdges());
			proc.append("		IF (SELECT COUNT(*) FROM " + joinPath(otherFirstPath.getEdges(), true) + ", " + thisJoinPath +
					", " + quoteId(target) + " WHERE " + lastJoinWhere + " AND " + qualifiedPK(source) + " = " + newId + ") = 0 THEN\n");

			LinkedList<SketchEdge> partialPath = new LinkedList<SketchEdge>(otherPath.getEdges());
			for (int i = edges.length-1; i > 0; i--) {
				EntityNode s = edges[i].getSourceEntity(), t = edges[i].getTargetEntity();
				proc.append("			");
				insertInto(proc, s, source.getShadowEdges(),
						qualifiedPK(t), // SELECT this FROM:
						joinPath(partialPath, false) + ", " + thisJoinPath + ", " + quoteId(target) +
								" WHERE " + lastJoinWhere + " AND " + qualifiedPK(source) + " = " + newId,
						Collections.singletonList(quoteId(tableFK(edges[i]))), null);
				partialPath.addFirst(edges[i]);
			}
			proc.append("		END IF;\n");
		}

		edges = thisFirstPath.getEdges().toArray(new SketchEdge[0]);
		LinkedList<SketchEdge> partialPath = new LinkedList<SketchEdge>(thisPath.getEdges());
		for (int i = edges.length - 1; i > 0; i--) {
			SketchEdge e = edges[i];
			EntityNode s = e.getSourceEntity(), t = e.getTargetEntity();
			proc.append("		");
			insertInto(proc, s, source.getShadowEdges(),
					qualifiedPK(t),
					joinPath(partialPath) + " WHERE " + qualifiedPK(source) + " = " + newId,
					Collections.singletonList(quoteId(tableFK(e))), null);
		}

		EntityNode firstStop = thisFirstPath.getFirstEdge().getTargetEntity(),
			otherFirstStop = otherFirstPath.getFirstEdge().getTargetEntity();
		// We need the two paths, but we need to remove the first element, because we don't want to include the pullback table
		LinkedList<SketchEdge> tmpPathA = new LinkedList<SketchEdge>(constraint.getFullPathA().getEdges());
		LinkedList<SketchEdge> tmpPathB = new LinkedList<SketchEdge>(constraint.getFullPathB().getEdges());
		tmpPathA.removeFirst(); tmpPathB.removeFirst();
		proc.append("		");
		LinkedList<String> cols = new LinkedList<String>();
		cols.add(quoteId(tableFK(thisFirstPath.getFirstEdge())));
		cols.add(quoteId(tableFK(otherFirstPath.getFirstEdge())));
		insertInto(proc, pb, source.getShadowEdges(),
				qualifiedPK(firstStop) + ", " + qualifiedPK(otherFirstStop),
				joinPath(tmpPathA, false) + ", " + joinPath(tmpPathB, false) + ", " + quoteId(target) +
						" WHERE " + lastJoinWhere + " AND " + qualifiedPK(source) + " = " + newId,
				cols, null);

		proc.append("	END IF;\n");

		proc.append("	RETURN NULL;\n" +
				"END\n");
		sql.add("CREATE FUNCTION " + quoteId(insConName) + "() RETURNS trigger AS " +
				quoteValue(proc.toString(), insConName) + " LANGUAGE plpgsql"+$);

		sql.add("CREATE TRIGGER " + quoteId(insConName) + " AFTER INSERT ON " + quoteId(source) + " FOR EACH ROW EXECUTE PROCEDURE " + quoteId(insConName) + "()"+$);
	}

	public List<String> createConstraint(EqualizerConstraint constraint, String id) {
		LinkedList<String> sql = new LinkedList<String>();

		EntityNode eq = constraint.getEqualizerEntity(),
				   source = constraint.getSourceEntity(),
				   target = constraint.getTargetEntity();
		InjectiveEdge projEdge = (InjectiveEdge) constraint.getProjection().getFirstEdge();
		List<SketchPath> paths = constraint.getEqualizerPaths();

		String insConName = "eqCon" + id + "InsUpd" + cleanId(source);

		StringBuilder proc = new StringBuilder("\nDECLARE\n");
		// Add a variable for each path, _path0Id, _path1Id, etc.:
		for (int i = 0; i < paths.size(); i++) {
			proc.append("	_path" + i + "Id " + pkType() + ";\n");
		}
		proc.append("BEGIN\n");
		String sourcePK = qualifiedPK(source),
			   targetPK = qualifiedPK(target),
			   newPK = "NEW." + quoteId(tablePK(source));
		int i = 0;
		for (SketchPath p: paths) {
			proc.append(
				"	SELECT " + targetPK + " INTO _path"+i+"Id\n" +
				"		FROM " + joinPath(p) + "\n" +
				"		WHERE " + sourcePK + " = " + newPK + ";\n");
			++i;
		}
		// Build equality clause, such as: _path0Id IS DISTINCT FROM _path1Id OR _path0Id IS DISTINCT FROM _path2Id OR ...
		proc.append("	IF _path0Id IS DISTINCT FROM _path1Id");
		for (int j = 2; j < paths.size(); j++)
			proc.append(" OR _path0Id IS DISTINCT FROM _path"+j+"Id");

		proc.append(
			" THEN\n"+
			// If one of the paths doesn't match, we need to need to delete it from the equalizer (if it's there)
			"		DELETE FROM "+quoteId(eq)+" WHERE "+qualifiedFK(projEdge)+" = "+newPK+";\n"+
			"	ELSIF (SELECT COUNT(*) FROM "+quoteId(eq)+" WHERE "+qualifiedFK(projEdge)+" = "+newPK+") = 0 THEN\n"+
			// Otherwise, if it isn't already in the equalizer, we need to add it:
			"		");
		insertInto(proc, eq, source.getShadowEdges(), null, null, Collections.singletonList(quoteId(tableFK(projEdge))), Collections.singletonList(newPK));
		proc.append(
			"	END IF;\n"+
			"	RETURN NULL;\n"+
			"END\n");

		sql.add("CREATE FUNCTION " + quoteId(insConName) + "() RETURNS trigger AS " + quoteValue(proc.toString(), insConName) +
				" LANGUAGE plpgsql"+$);

		// Create the trigger for inserting into the source table:
		sql.add("CREATE TRIGGER " + quoteId(insConName) + " AFTER INSERT OR UPDATE ON " + quoteId(source) + "\n" +
			   	"	FOR EACH ROW EXECUTE PROCEDURE " + quoteId(insConName) + "()"+$);

		// If the projection isn't a cascading edge, cascade anyway:
		if (projEdge.getCascading() != SketchEdge.Cascade.CASCADE) {
			String delConName = "eqCon" + id + "Del" + cleanId(source);
			sql.add("CREATE FUNCTION " + quoteId(delConName) + "() RETURNS trigger AS " +
					quoteValue("\nBEGIN\n	DELETE FROM "+quoteId(eq)+" WHERE "+qualifiedFK(projEdge)+" = OLD."+quoteId(tablePK(source)) + ";\n	RETURN NULL;\nEND\n", delConName) +
					" LANGUAGE plpgsql"+$);
			sql.add("CREATE TRIGGER " + quoteId("eqCon"+id+"Del"+cleanId(source)) + " BEFORE DELETE ON " + quoteId(source) + "\n" +
					"	FOR EACH ROW EXECUTE PROCEDURE " + quoteId(delConName) + "()"+$);
		}

		return sql;
	}

	public List<String> createConstraint(SumConstraint constraint, String id) {

		LinkedList<String> sql = new LinkedList<String>();

		for (SketchPath path: constraint.getPaths()) {
			EntityNode dom = path.getDomain(), codo = path.getCoDomain();
			String baseConName = "sumCon" + id + cleanId(dom);
			StringBuilder deleteClause = new StringBuilder();
			deleteClause.append("	DELETE FROM " + quoteId(codo) + " WHERE " + qualifiedPK(codo) + " = ");
			if (path.getEdges().size() == 1) {
				deleteClause.append("OLD." + quoteId(tableFK(path.getFirstEdge())));
			}
			else {
				LinkedList<SketchEdge> trailing = new LinkedList<SketchEdge>(path.getEdges());
				trailing.removeFirst();
				deleteClause.append(
					"(SELECT " + qualifiedFK(trailing.getLast()) + " FROM " + joinPath(trailing, false) +
					" WHERE " + qualifiedPK(path.getFirstCoDomain()) + " = " + quoteId(tableFK(path.getFirstEdge())) + ")");
			}

			sql.add("CREATE FUNCTION " + quoteId(baseConName + "Del") + "() RETURNS trigger AS " +
					quoteValue("\nBEGIN\n"+
						deleteClause + ";\n" +
						"	RETURN NULL;\n" +
						"END\n",
						baseConName + "Del") +
					" LANGUAGE plpgsql"+$);
			sql.add(
				"CREATE TRIGGER " + quoteId(baseConName + "Del") + " AFTER DELETE ON " + quoteId(dom) + " FOR EACH ROW EXECUTE PROCEDURE " + quoteId(baseConName + "Del") + "()"+$);

			StringBuilder proc = new StringBuilder("\nBEGIN\n");
			proc.append("	");
			insertInto(proc, codo, dom.getShadowEdges(), null, null, null, null);

			SketchEdge[] edges = path.getEdges().toArray(new SketchEdge[0]);
			for (int i = edges.length - 1; i > 0; i--) {
				EntityNode source = edges[i].getSourceEntity();
				proc.append("	");
				insertInto(proc, source, dom.getShadowEdges(), null, null,
						Collections.singletonList(quoteId(tableFK(edges[i]))),
						Collections.singletonList(currval(edges[i].getTargetEntity())));
			}
			proc.append("	NEW." + quoteId(tableFK(edges[0])) + " := " + currval(edges[0].getTargetEntity()) + ";\n");
			proc.append("	RETURN NEW;\n");
			proc.append("END");

			sql.add("CREATE FUNCTION " + quoteId(baseConName + "Ins") + "() RETURNS trigger AS " +
					quoteValue(proc.toString(), baseConName + "Ins") + " LANGUAGE plpgsql"+$);

			sql.add("CREATE TRIGGER " + quoteId(baseConName + "Ins") + " BEFORE INSERT ON " + quoteId(dom) + " FOR EACH ROW " +
					"EXECUTE PROCEDURE " + quoteId(baseConName + "Ins") + "()"+$);
		}

		return sql;
	}

	/**
	 * Returns the current value of a table's primary key sequence.
	 */
	private String currval(EntityNode node) {
		return "currval('" + quoteId(node + "_" + tablePK(node) + "_seq") + "')";
	}

	/**
	 * Appends an INSERT INTO statement to the provided StringBuilder, and stores
	 * the shadow edges to be cleared at the end of the trigger.
	 */
	private void insertInto(
			StringBuilder proc,
			EntityNode node,
			Collection<SketchEdge> inShadowEdges,
			String selectWhat,
			String selectFrom,
			List<String> cols,
			List<String> vals) {
		HashSet<SketchEdge> shadowCols = new HashSet<SketchEdge>();
		if (cols == null) cols = new LinkedList<String>();
		if (vals == null) vals = new LinkedList<String>();
		LinkedList<SketchEdge> shadowEdges = node.getNonPartialEdges();
		shadowEdges.addAll(node.getShadowEdges());
		shadowEdges.retainAll(inShadowEdges);
		proc.append("INSERT INTO " + quoteId(node) + " ");
		if (cols.isEmpty() && shadowEdges.isEmpty()) {
			proc.append("DEFAULT VALUES");
		}
		else {
			proc.append("(");
			StringBuilder values = new StringBuilder();
			for (String col: cols)
				proc.append(col + ", ");
			for (String val: vals)
				values.append(val + ", ");

			for (SketchEdge e: shadowEdges) {
				proc.append(quoteId(tableFK(e)) + ", ");
				values.append("NEW." + quoteId(tableFK(e)) + ", ");
				shadowCols.add(e);
			}
			proc.delete(proc.length()-2, proc.length());
			if (values.length() > 0)
				values.delete(values.length()-2, values.length());

			if (selectWhat != null) {
				proc.append(") SELECT " + selectWhat + (values.length() > 0 ? ", " + values : "") + " FROM " + selectFrom);
			}
			else {
				proc.append(") VALUES (" + values + ")");
			}
		}
		proc.append(";\n");
	}

	public List<String> createView(ViewNode node){
		LinkedList<String> viewSQL = new LinkedList<String>();
		
		for( QueryNode qn : node.getFrame().getView().getQueryNodes() ){
			viewSQL.add(createView(qn));
		}

		return viewSQL;
	}

	public String createView(QueryNode node){
		return "CREATE VIEW " + node.getName() + " AS " + node.getQuery() + $;
	}

	/**
	 * Takes a value, quotes it using PostgreSQL's $-quoting.  This means adding ${identifier}$ to the beginning
	 * and end.  If the identifier is already contained in the value, random characters will added to the
	 * identifier until it is not.
	 */
	private String quoteValue(String value, String identifier) {
		identifier = identifier.replaceAll("\\$", "");
		if (value.contains("$"+identifier+"$")) {
			identifier += "-";
			Random rand = new Random();
			do {
				int r = rand.nextInt(62);
				char rc = (char)(
					r < 26 ? ('A' + r) :
					r < 52 ? ('a' + (r-26)) :
							 ('0' + (r-52)));
				identifier += rc;
			} while (value.contains("$"+identifier+"$"));
		}
		return "$"+identifier+"$" + value + "$"+identifier+"$";
	}

	/**
	 * The type of primary keys.  Note that this is *not* applicable to table creation--that uses
	 * BIGSERIAL/SERIAL instead (which itself is a convenience wrapper around BIGINT/INT + a
	 * sequence).
	 */
	private String pkType() {
		return optionEnabled("bigKeys") ? "BIGINT" : "INTEGER";
	}


	/**
	 * Append triggers to clear shadow columns.  This needs to be done *after* any other triggers run,
	 * as the earlier triggers may require the shadow column values.
	 */
	public List<String> createExtras() {
		LinkedList<String> extras = new LinkedList<String>();
		LinkedList<StringBuilder> trigs = new LinkedList<StringBuilder>();

		// new LinkedHashMap<EntityNode, LinkedHashMap<String, HashSet<SketchEdge>>>();
		for (EntityNode table: sketch.getEntities()) {
			HashSet<SketchEdge> shadowEdges = table.getShadowEdges();
			if (!shadowEdges.isEmpty()) {
				StringBuilder proc = new StringBuilder("\n	BEGIN\n");
				LinkedList<String> sets = new LinkedList();
				LinkedList<String> checks = new LinkedList();
				for (SketchEdge edge: shadowEdges) {
					sets.add(quoteId(tableFK(edge)) + " = NULL");
					checks.add("NEW."+quoteId(tableFK(edge)) + " IS NOT NULL");
				}
				proc.append("		IF " + EasikTools.join(" OR ", checks) + " THEN\n");
				proc.append("			UPDATE " + quoteId(table) + " SET " + EasikTools.join(", ", sets) + " WHERE " + quoteId(tablePK(table)) + " = NEW." + quoteId(tablePK(table)) + ";\n");
				proc.append("		END IF;\n");
				proc.append("		RETURN NULL;\n");
				proc.append("	END\n");

				extras.add(
						"CREATE FUNCTION " + quoteId("clearShadow_" + table) + "() RETURNS trigger AS " +
						quoteValue(proc.toString(), "clearShadow_" + table) +
						" LANGUAGE plpgsql"+$);

				extras.add( // Pg executes triggers in alphabetical order, so prepend zzzzz:
						"CREATE TRIGGER " + quoteId("zzzzzClearShadow_" + table) + " AFTER INSERT OR UPDATE ON " + quoteId(table) + "\n" +
						"	FOR EACH ROW EXECUTE PROCEDURE " + quoteId("clearShadow_" + table) + "()"+$);
			}
		}

		return extras;
	}
}

