package easik.database.types;

import java.util.Map;
import java.util.HashMap;
import java.sql.Types;
import java.sql.PreparedStatement;
import java.sql.SQLException;

/**
 * A class representing a DECIMAL (also known as NUMERIC) SQL data type.  A
 * DECIMAL requires two arguments: the precision and scale.  The precision
 * indicates the maximum total number of significant digits of the number, and the
 * scale indicates the maximum total number of significant digits after the decimal point.
 * For example, the value 123.45678 has precision 8, and scale 5.  This type is
 * often used for monetary values, where floating point numbers are not appropriate.
 * Note, however, that DECIMAL types are significantly slower and occupy significantly more
 * space than FLOAT or DOUBLE PRECISION types.
 */
public class Decimal extends Type {
	private int precision, scale;
	public Decimal(int p, int s) {
		precision = p;
		scale = s;
	}
	public int getPrecision() {
		return precision;
	}
	public int getScale() {
		return scale;
	}
	public String toString() { return "DECIMAL(" + precision + "," + scale + ")"; }
	
	public boolean verifyInput(String input){
		int decimal = input.indexOf('.');
		
		if(decimal >= 0){ 
			if(decimal > (precision-scale))
				return input.matches("^[-+]?\\d{0," + (precision-1) + "}.\\d{1," + (precision-decimal) + "}$");
			else
				return input.matches("^[-+]?\\d{0," + (precision-1) + "}.\\d{1," + scale + "}$");
		}
		else
			return input.matches("^[-+]?\\d{1," + precision + "}$");
	}

	/**
	 * Returns the attributes of this object
	 */
	public Map<String, String> attributes() {
		HashMap<String, String> attr = new HashMap<String, String>();
		attr.put("precision", precision+"");
		attr.put("scale", scale+"");
		return attr;
	}
	/**
	 * Recreates the object from the attributes returned by attributes().
	 */
	public Decimal(Map<String, String> attr) {
		this(Integer.parseInt(attr.get("precision")), Integer.parseInt(attr.get("scale")));
	}

	public int getSqlType(){ return Types.DECIMAL; }

	public void bindValue(PreparedStatement ps, int col, String value) throws SQLException{
		ps.setDouble(col, Double.parseDouble(value));
	}
}
