package easik.os;

import com.apple.eawt.*;
import easik.Easik;
import easik.ui.ProgramSettingsUI;
import easik.ui.menu.AboutAction;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JSeparator;
import javax.swing.KeyStroke;
import java.awt.event.KeyEvent;
import java.awt.event.InputEvent;
import java.awt.Component;

/**
 * This class contains various OS X-specific features.  It will be loaded
 * dynamically and one object created (but not stored) by Easik *after* setting
 * up the main window when running under an OS X environment, and is
 * responsible for OS X-specific tasks (like setting up the About menu, the
 * Preferences menu, etc.).  We remove these menus from the usual Easik menu
 * list, as long as the system property apple.laf.useScreenMenuBar is set to
 * true (this happens at the beginning of Easik.java).
 */

public class OSX extends ApplicationAdapter implements EasikMenuAdjuster {
	public OSX() {
		Application app = Application.getApplication();
		app.addApplicationListener(this);
		app.addAboutMenuItem();
		app.addPreferencesMenuItem();
		app.setEnabledPreferencesMenu(true);
		app.setEnabledAboutMenu(true);

		Easik.getInstance().getFrame().setMenuAdjuster(this);
	}

	/**
	 * Go through the Easik menu bars and *remove* the settings and about
	 * menus, since those are located under the main Program menu in OS X.
	 */
	public void adjustMenu(JMenuBar menu) {
		if ("true".equals(System.getProperty("apple.laf.useScreenMenuBar"))) {
			int numMenus = menu.getMenuCount();
			boolean changedRedo = false, removedQuit = false, removedSettings = false, removedAbout = false;
			MENU: for (int i = 0; i < numMenus && !(changedRedo && removedSettings && removedAbout && removedQuit); i++) {
				JMenu m = menu.getMenu(i);
				boolean editMenu = m.getText().equals("Edit"), fileMenu = m.getText().equals("File"), helpMenu = m.getText().equals("Help");
				if (editMenu || fileMenu || helpMenu) {
					int numItems = m.getItemCount();
					for (int j = 0; j < m.getMenuComponentCount(); j++) {
						Component comp = m.getMenuComponent(j);
						if (
								comp instanceof JMenuItem && (
									(editMenu && ((JMenuItem) comp).getText().matches("(?i)Preferences"))
									||
									(fileMenu && ((JMenuItem) comp).getText().matches("(?i)Quit"))
									||
									(helpMenu && ((JMenuItem) comp).getText().matches("^(?i)\\s*about\\s*(?:easik)?\\s*$"))
								)
						) {
							// If the component is:
							// a) the first item, and followed by a separator, remove the separator, too
							if (j == 0 && m.getMenuComponentCount() > 1 && m.getMenuComponent(1) instanceof JSeparator) {
								m.remove(1);
								m.remove(0);
							}
							// b) the last item, preceeded by a separator, remove the separator, too
							else if (j > 0 && j == m.getMenuComponentCount() - 1 && m.getMenuComponent(j-1) instanceof JSeparator) {
								m.remove(j);
								m.remove(j-1);
							}
							// c) has a separator before and after; remove one of the separators
							else if (j > 0 && j < m.getMenuComponentCount() - 1 &&
									m.getMenuComponent(j-1) instanceof JSeparator && m.getMenuComponent(j+1) instanceof JSeparator) {
								m.remove(j);
								m.remove(j-1);
							}
							// otherwise, just remove the item
							else {
								m.remove(j);
							}

							if (editMenu)
								removedSettings = true;
							else if (fileMenu)
								removedQuit = true;
							else if (helpMenu)
								removedAbout = true;
						}
						else if (comp instanceof JMenuItem && editMenu && ((JMenuItem) comp).getText().matches("(?i)Redo")) {
							((JMenuItem) comp).setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Z, java.awt.Toolkit.getDefaultToolkit().getMenuShortcutKeyMask() | InputEvent.SHIFT_DOWN_MASK));
							changedRedo = true;
						}
					}
					if (m.getItemCount() == 0) {
						menu.remove(m);
						i--;
					}
				}

			}
		}
	}

	public void handleQuit(ApplicationEvent e) {
		e.setHandled(true);
		Easik.getInstance().getFrame().closeWindow();
	}

	public void handlePreferences(ApplicationEvent e) {
		e.setHandled(true);
		ProgramSettingsUI myUI = new ProgramSettingsUI(Easik.getInstance().getFrame());
		myUI.showDialog();
	}

	public void handleAbout(ApplicationEvent e) {
		e.setHandled(true);
		new AboutAction().showAboutDialog();
	}
}
