package easik.sketch.constraint;


import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.List;
import java.util.Collections;
import java.util.LinkedHashSet;

import javax.swing.tree.DefaultMutableTreeNode;

import org.jgraph.graph.*;

import easik.sketch.Sketch;
import easik.sketch.edge.GuideEdge;
import easik.sketch.edge.SketchEdge;
import easik.sketch.path.SketchPath;
import easik.sketch.vertex.EntityNode;
import easik.sketch.vertex.SketchVertex;
import easik.sketch.states.LoadingState;

/** 
 * This is the superclass for all constraints. Constraints get a non-unique label
 * that shows what kind of constraint it is (symbolically) and have a set of paths
 * which it deals with.
 *
 * Note: subclasses must call assignToEntites(..) after edge set has been populated.
 *
 * @author Rob Fletcher 2005
 * @author Kevin Green 2006
 * @author Vera Ranieri 2006
 * @version 2006-07-26 Kevin Green
 */
public abstract class Constraint extends SketchVertex {

	/** The paths involved in the constraint */
	protected ArrayList<SketchPath> _paths;

	/** The edge involved in the constraint */
	protected ArrayList<SketchEdge> _edges;

	/** Stores whether the constraint is visible or not */
	protected boolean _isVisible = true;

	/** Stores whether or not the visuals are part of the graph */
	protected boolean _edgesVisualized = false, _nodeVisualized = false;

	/** The Guide Edges of the constraint */
	protected ArrayList<GuideEdge> _visuals;

	/** The node representing this constraint in the info tree */
	protected DefaultMutableTreeNode _treeNode;
	
	/** Default constructor needed for subclasses */
	public Constraint(Sketch inSketch) {
		super(inSketch);

		_visuals = new ArrayList<GuideEdge>();
	}

	/**
	 * Creates a very simple constraint node.
	 * 
	 * @param name The label on the node
	 * @param x x coordinate
	 * @param y y coordinate
	 * @param isVisible If the constraint is visible in the graph or not
	 * @param inSketch the sketch this constraint belongs to
	 */
	public Constraint(String name, int x, int y, boolean isVisible, Sketch inSketch){
		super(name, x, y, inSketch);
		_isVisible = isVisible;
		_visuals = new ArrayList<GuideEdge>();
	}
	
	/**
	 * Returns if the constraint is visible or not (True = visible)
	 * 
	 * @return If the constraint is visible or not
	 */
	public boolean isVisible(){
		return _isVisible;
	}
	
	/**
	 * Sets if the constraint should be visible or not
	 * 
	 * @param inIsVisible If the constraint should be visible or not.
	 */
	public void setVisible(boolean inIsVisible){
		if(inIsVisible != _isVisible) {
			_isVisible = inIsVisible;
			_theSketch.setDirty();
		}

		GuideEdge[] visuals = _visuals.toArray(new GuideEdge[0]);
		GraphLayoutCache glc = _theSketch.getGraphLayoutCache();
		if (_isVisible && !_edgesVisualized) {
			addVisualsToSketch();
		}
		else if (!_isVisible && _edgesVisualized) {
			removeVisualsFromSketch();
		}

		_theSketch.clearSelection();
	}

	/**
	 * Adds the visual aids to the sketch
	 */
	private void addVisualsToSketch() {
		//Push loading state
		_theSketch.getStateManager().pushState(new LoadingState(_theSketch));

		_theSketch.getSketchModel().beginInsignificantUpdate();

		_visuals.clear();

		// Get Nodes involved in constraint
		LinkedHashSet<EntityNode> entitiesInvolved = new LinkedHashSet<EntityNode>();
		for (SketchEdge edge: _edges) {
			entitiesInvolved.add(edge.getSourceEntity());
			entitiesInvolved.add(edge.getTargetEntity());
		}

		// Place this object at the average position of all the involved entities, and draw
		// edges going to each entity involved
		int avgX = 0, avgY = 0;
		for (EntityNode node: entitiesInvolved) {
			GuideEdge myEdge = new GuideEdge(this, node);
			_visuals.add(myEdge);
			avgX += node.getX();
			avgY += node.getY();
		}

		//If only one entity is involved add some vertical space
		if(entitiesInvolved.size() == 1)
			avgY += 20;

		int posX = getX();
		int posY = getY();
		//If visual does not already have a position then position it
		if( posX == 0 && posY == 0 ){
        	posX = avgX / entitiesInvolved.size();
       		posY = avgY / entitiesInvolved.size();
		}

		// Set the on-screen position of our entity to the attributes of the entity
		AttributeMap attribs = getAttributes();
		GraphConstants.setAutoSize(attribs, true);
		GraphConstants.setBounds(attribs, new Rectangle2D.Double(posX, posY, 0, 0));
		GraphConstants.setSelectable(attribs, false);

		// Actually add them to the sketch display:
		GraphLayoutCache glc = _theSketch.getGraphLayoutCache();
		if (!_nodeVisualized) {
			glc.insert(this);
			_nodeVisualized = true;
		}
		glc.insert(_visuals.toArray(new GuideEdge[0]));

		_edgesVisualized = true;

		_theSketch.getSketchModel().endInsignificantUpdate();

		_theSketch.refresh();

		//Pop state
		_theSketch.getStateManager().popState();
	}

	/**
	 * Removes the visuals from the sketch (when hiding)
	 */
	private void removeVisualsFromSketch() {
		//Push loading state
		_theSketch.getStateManager().pushState(new LoadingState(_theSketch));

		_theSketch.getSketchModel().beginInsignificantUpdate();

		GraphLayoutCache glc = _theSketch.getGraphLayoutCache();
		glc.remove(_visuals.toArray(new GuideEdge[0]));
		_visuals.clear();

		_edgesVisualized = false;

		GraphConstants.setSelectable(getAttributes(), false);

		_theSketch.refresh();

		_theSketch.getSketchModel().cancelInsignificantUpdate();

		//Pop state
		_theSketch.getStateManager().popState();
	}

	/**
	 * Tests to see if an entity is either the domain or co-domain of
	 * any of the paths involved in the constraint
	 * 
	 * @param inEntity The entity to be tested
	 * @return True if it is either domain or co-domain of a path involved in the constraint, false otherwise.
	 */
	public boolean isDomainOrCoDomain(EntityNode inEntity){
		int size = _paths.size();
		for (SketchPath path: _paths) {
			if (path.getDomain() == inEntity || path.getCoDomain() == inEntity)
				return true;
		}
		return false;
	}
	
	/**
	 * Adds all edges participating in this constraint to an arraylist of edges.
	 */
	protected void addEdges(){
		_edges = new ArrayList<SketchEdge>();
		//Fill in _edges array
		for(SketchPath p : _paths){
			List<SketchEdge> curPathEdges = p.getEdges();
			for(SketchEdge e: curPathEdges)
			{
				if(!_edges.contains(e))
					_edges.add(e);
			}
		}
	}

	/**
	 * Finds all of the entities involved in this constraint and makes a call to addConstraint(this) on each one.
	 * Duplicates will be found but Java's Set.add(..) method takes care of that.
	 */
	protected void assignToEntities(){
		for ( SketchEdge edge : _edges ) {
			edge.getSourceEntity().addConstraint(this);
			edge.getTargetEntity().addConstraint(this);
		}
	}

	/**
	 * Accessor for the constraint label
	 * 
	 * @return The label
	 */
	public String toString() {
		return getName();
	}

	/**
	 * Checks to see if the constraint contains a path that uses an edge
	 * 
	 * @param inEdge The edge to check against.
	 * @return True if constraint contains inEdge, false otherwise.
	 */
	public boolean hasEdge(SketchEdge inEdge) {
		return _edges.contains(inEdge);
	}

	/**
	 *	Returns the edges dealt with in the constraint.
	 *
	 * @return The edges
	 */
	public List<SketchEdge> getEdges() {
		return Collections.unmodifiableList(_edges);
	}

	/**
	 * Returns the guide edges used by the constaint.
	 *
	 * @return the GuideEdges
	 */
	public List<GuideEdge> getGuideEdges() {
		return Collections.unmodifiableList(_visuals);
	}

	/**
	 * Returns the set of the paths involved in the constraint
	 * 
	 * @return The array of sketch paths
	 */
	public List<SketchPath> getPaths() {
		return Collections.unmodifiableList(_paths);
	}
	
	/**
	 * Returns a string corresponding to the constraint type.
	 * (used in XML generation).
	 * 
	 * @return A string of the type of constraint
	 */
	public String getType() {
		if (this instanceof SumConstraint) {
			return "sumconstraint";
		} else if (this instanceof PullbackConstraint) {
			return "pullbackconstraint";
		} else if (this instanceof EqualizerConstraint) {
			return "equalizerconstraint";
		} else if (this instanceof ProductConstraint) {
			return "productconstraint";
		} else if (this instanceof CommutativeDiagram) {
			return "commutativediagram";
		} else {		
			return "unknown";
		}
	}

	public DefaultMutableTreeNode getTreeNode() {
		if (_treeNode == null)
			_treeNode = new DefaultMutableTreeNode(this);

		return _treeNode;
	}
	
	/**
	 * Sets all constraints to visible/invisible in the sketch.
	 * 
	 * @param constraints The list of constraints.
	 * @param show true if constraints are to be set to visible, false otherwise
	 * @since 2006-05-29 Vera Ranieri
	 */
	public static void setAllConstraintsVisible(List<Constraint> constraints, boolean show){
		for(Constraint c : constraints){
			c.setVisible(show);
		}
	}
	/**
	 * Determines the entities involved in a constraint and lists them for an error message.
	 * @param paths The paths forming the constraint
	 * @return A string of the domains and codomains of the constraint, formatted for an error message
	 * @since 2006-08-04 Vera Ranieri
	 */
	public static String getTablesInvolvedForError(ArrayList<SketchPath> paths){
		ArrayList<String> domains = new ArrayList<String>();
		ArrayList<String> codomains = new ArrayList<String>();
		String domain, codomain;
		for(SketchPath p: paths){
			domain = p.getDomain().getName();
			codomain = p.getCoDomain().getName();
			if(!domains.contains(domain))
				domains.add(domain);
			if(!codomains.contains(codomain))
				codomains.add(codomain);
		}
		String result = "With domains: ";
		for(String d : domains){
			result += d + ", ";
		}
		result = result.substring(0, result.length()-2) + "\n and \n Codomains: ";
		for(String c : codomains){
			result += c +", ";
		}
		result = result.substring(0, result.length()-2);
		return result;
	}

	/**
	 * Overrides superclass getAttributes to override selectable status.
	 */
	public AttributeMap getAttributes() {
		AttributeMap attrs = super.getAttributes();

		// Selectable as long as it is visible:
		GraphConstants.setSelectable(attrs, isVisible());

		return attrs;
	}
}
