package easik.sketch.constraint;

import java.util.ArrayList;
import java.util.List;
import java.util.LinkedList;
import java.util.Collections;

import easik.sketch.Sketch;
import easik.sketch.path.SketchPath;
import easik.sketch.edge.InjectiveEdge;
import easik.sketch.vertex.EntityNode;


/** 
 * Class representing an equalizer constraint.  This requires 3 paths: one from
 * the equalizer node to the start node of the equalizer paths, which must be a
 * single, injective edge, and two from the start node to the same target node,
 * along different paths.
 *
 * @author Jason Rhinelander 2008
 */
/* To aid in visualizing the procedure, we describe paths as equalizer and left/right; 
 * those are obviously arbitrary designations, but are based on this sort of visual
 * representation:
 *     E
 *     |
 *     A
 *    / \
 *   B   C
 *    \ /
 *     D
 *
 * where E->A is the equalizer path, A->B->D is the "left" path, and A->C->D is the "right" path.
 */
public class EqualizerConstraint extends Constraint {

	/**
	 * Takes a List of paths, makes them an equalizer.  The paths can be specified in
	 * any order, as long as they can be arranged into an equalizer.  If the paths
	 * are invalid (that is, they cannot form an equalizer), a
	 * ConstraintException is thrown.
	 *
	 * @param inPaths A List of SketchPaths 
	 * @param inSketch the Sketch this constraint belongs to
	 * @throws ConstraintException if the paths do not form a valid equalizer constraint
	 */
	public EqualizerConstraint(List<SketchPath> inPaths, Sketch inSketch) throws ConstraintException {
		super(inSketch);
		setName("EQ");
		_isVisible = true;
		initPaths(inPaths);
	}

	/**
	 * Takes an ArrayList of paths, with coordinates and a visible boolean, makes them an equalizer.
	 *
	 * @param paths An ArrayList of SketchPaths 
	 * @param x X coordinate of visual aid
	 * @param y Y coordinate of visual aid
	 * @param isVisible If the constraint is visible in the graph or not
	 * @param inSketch the sketch
	 */
	public EqualizerConstraint(List<SketchPath> paths, int x, int y, boolean isVisible, Sketch inSketch) throws ConstraintException {
		super("EQ", x, y, isVisible, inSketch);

		initPaths(paths);
	}

	/**
	 * Common path verification shared by constructors
	 */
	private void initPaths(List<SketchPath> inPaths) throws ConstraintException {
		_paths = new ArrayList<SketchPath>(inPaths);
		if (!isEqualizerConstraint(_paths)) // This method rearranges _paths
			throw new ConstraintException("Unable to create equalizer constraint: specified paths do not form a valid equalizer constraint");

		addEdges();
		assignToEntities();
	}

	/**
	 * Returns the equalizer node.
	 *
	 * @return EntityNode the equalizer node
	 */
	public EntityNode getEqualizerEntity() {
		return _paths.get(0).getDomain();
	}

	/**
	 * Returns the "source" node of the equalizer--that is, the domain of both
	 * equalizer paths.
	 *
	 * @return EntityNode the source node
	 */
	public EntityNode getSourceEntity() {
		return _paths.get(1).getDomain();
	}

	/**
	 * Returns the "target" node of the equalizer (that is, the codomain of both
	 * equalizer paths).
	 *
	 * @return EntityNode the equalizer target node
	 */
	public EntityNode getTargetEntity() {
		return _paths.get(1).getCoDomain();
	}

	/**
	 * Returns the path containing the injective edge from the equalizer node
	 * to the source node.
	 */
	public SketchPath getProjection() {
		return _paths.get(0);
	}

	/**
	 * Returns the paths from the source node to target node.
	 */
	public List<SketchPath> getEqualizerPaths() {
		LinkedList<SketchPath> eqPaths = new LinkedList<SketchPath>(_paths);
		eqPaths.removeFirst();
		return Collections.unmodifiableList(eqPaths);
	}

	/**
	 * Method to determine whether a set of paths potentially forming a equalizer constraint could legally
	 * form a equalizer constraint.  The first path must be the equalizer projection, and must consist of
	 * a single, injective edge.  The remaining paths (of which there must be at least 2) must start at the
	 * project path's codomain, must all have the same codomain, and must be distinct paths from one another.
	 *
	 * @param paths The set of potential paths
	 * @return True if a legal equalizer configuration, false otherwise.
	 */
	public static boolean isEqualizerConstraint(ArrayList<SketchPath> paths) {

		// There has to be at least 3 paths: the projection, and 2 or more equalizer paths.
		if (paths.size() < 3)
			return false;

		// Null paths are not permitted, nor are empty paths:
		for (SketchPath p: paths) {
			if (p == null) return false;
			if (p.getEdges().isEmpty()) return false;
		}

		// Make sure the the first path is a valid projection path: it must start with an injective edge,
		// and must be of length 1.
		if (!paths.get(0).isInjective() || paths.get(0).getEdges().size() != 1)
			return false;

		EntityNode domain = paths.get(0).getCoDomain();
		EntityNode codomain = paths.get(1).getCoDomain();
		// and its domain must be the codomain of all the other edges.  Also make sure that all
		// the other paths agree on codomain:
		for (int i = 1; i < paths.size(); i++) {
			SketchPath p = paths.get(i);
			if (p.getDomain() != domain || p.getCoDomain() != codomain)
				return false;

			// We also need to make sure that there are no equal paths:
			for (int j = i+1; j < paths.size(); j++)
				if (SketchPath.pathsAreEqual(p, paths.get(j)))
					return false;
		}

		return true;
	}
}
