package easik.sketch.constraint;


import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import easik.sketch.edge.SketchEdge;
import easik.sketch.path.SketchPath;
import easik.sketch.vertex.EntityNode;
import easik.sketch.Sketch;

/**
 * This class represents the sum colimit constraint. It contains a list of
 * paths sharing a target. 
 *
 * @author Kevin Green 2006
 * @author Vera Ranieri 2006
 * @version 2006-07-13 Vera Ranieri
 */
public class SumConstraint extends Constraint {
	

	/**
	 * Default Constructor takes a List of SketchPaths
	 * 
	 * @param inPaths A List of SketchPaths
	 */
	public SumConstraint(List<SketchPath> inPaths, Sketch inSketch) {
		super(inSketch);
		_paths = new ArrayList<SketchPath>(inPaths);
		_edges = new ArrayList<SketchEdge>();
		//Fill in _edges array
		for(SketchPath p: inPaths){
			LinkedList<SketchEdge> curPathEdges = p.getEdges();
			for(SketchEdge e: curPathEdges) {
				if (!_edges.contains(e))
					_edges.add(e);
			}
		}
		_isVisible = true;
		setName("+");
		assignToEntities();
	}

	/**
	 * Constructor for specifying position on graph
	 * 
	 * @param inPaths A List of SketchPaths
	 * @param x X coordinate of visual on graph
	 * @param y Y coordinate of visual on graph
	 * @param isVisible If the constraint is visible in the graph or not
	 */
	public SumConstraint(List<SketchPath> inPaths, int x, int y, boolean isVisible, Sketch inSketch) {
		super("+", x, y,isVisible,inSketch);
		_paths = new ArrayList<SketchPath>(inPaths);
		_edges = new ArrayList<SketchEdge>();
			   
		//Fill in _edges array
		for(SketchPath p : inPaths){
			LinkedList<SketchEdge> curPathEdges = p.getEdges();
			for(SketchEdge e: curPathEdges)
			{
				if(!_edges.contains(e))
				_edges.add(e);
			}
		}
		assignToEntities();
	}
	
	/**
	 * Sets the path array, updates edge list, and updates display
	 * 
	 * @param inPaths The new array of paths
	 */
	public void setPaths(List<SketchPath> inPaths){
		_paths = new ArrayList<SketchPath>(inPaths);
		 
		addEdges();
		setVisible(!isVisible());
		setVisible(!isVisible());
	}
	
	/**
	 * Static method to determine whether a set of paths could theoretically 
	 * designate a sum constraint.
	 * @param paths The paths to check for a sum constraint formation.
	 * @return True if forms a sum constraint, false otherwise.
	 * @since 2006-05-29 Vera Ranieri
	 */
	public static boolean isSumConstraint(List<SketchPath> paths){
//		We must have a bunch of injective arrows aimed at the same spot
		EntityNode sharedTarget = null;
		int edgesInvolved = 0;
		
		SketchPath currentPath;
		SketchEdge edge;

		int size = paths.size();
		// Search through the selection...
		for (int i = 0; i < size; i++) {
				currentPath = (SketchPath) paths.get(i);

				// First target we come across becomes our 'sharedTarget'
				if (sharedTarget == null) {
					sharedTarget = (EntityNode) currentPath.getCoDomain();
				} 
				else if (sharedTarget != currentPath.getCoDomain()) {
					// If we encounter a source which is not the sharedSource, fail			
					return false;
				}
				
				//Make sure the first edge in the path is injective
				if(currentPath.getEdges().size() != 0){
					edge = (SketchEdge)currentPath.getEdges().get(0);
					if( !edge.isInjective() ){
						return false;
					}
					
					for(int j = i+1; j< size; j++){
						if(SketchPath.pathsAreEqual(currentPath, (SketchPath)paths.get(j))){
							return false;
						}
					}
				}
				// Otherwise, we're fine, add the edge
				edgesInvolved++;
		}

		return edgesInvolved >= 2;
	}
}

