package easik.sketch.path;

import java.util.*;

import easik.EasikTools;
import easik.sketch.Sketch;
import easik.sketch.edge.SketchEdge;
import easik.sketch.vertex.EntityNode;

import javax.swing.tree.DefaultMutableTreeNode;

/**
 * This class is used to track path data used for constraints
 * 
 * @author Kevin Green 2006
 * @author Vera Ranieri 2006
 * @since 2006-05-16 Kevin Green
 * @version 2006-07-04 Kevin Green
 */
public class SketchPath extends DefaultMutableTreeNode {
	
	/**
	 * The unique id of this path
	 */
	private String _id = null;
	/**
	 * The list of edges comprising this path
	 */
	private LinkedList<SketchEdge> _edges;
	/**
	 * The domain of this path
	 */
	private EntityNode _domain;
	/**
	 * The codomain of this path
	 */
	private EntityNode _coDomain;
	
	/**
	 * Constructor takes one entity, creates a 0-length path.
	 * 
	 * @param inEntity An entity
	 */
	public SketchPath(EntityNode inEntity){
		_domain = inEntity;
		_coDomain = inEntity;
		_edges = new LinkedList<SketchEdge>();
	}
	
	/**
	 * Constructor takes a List of one or more edges that form a path.
	 * 
	 * @param inPath A List of edges
	 */
	public SketchPath(List<SketchEdge> inPath){
		_edges = new LinkedList<SketchEdge>(inPath);
		_domain = _edges.getFirst().getSourceEntity();
		_coDomain = _edges.getLast().getTargetEntity();
	}
	
	/**
	 * Generates a new ID based on the edges in the path
	 *
	 */
	private void generateID(){
		if (_id == null) {
			if (_edges.size() == 0)
				_id = _domain.getName() + "_identity";
			else
				_id = EasikTools.join(";", _edges);
		}
	}

	/**
	 * Overloaded method returns the _id
	 * @return The id of this path
	 */
	public String toString(){
		return getId();
	}
	
	/**
	 * Returns the ID of the path
	 * 
	 * @return The ID of the path
	 */
	public String getId(){
		generateID();
		return _id;
	}
	
	/**
	 * Returns the list of edges
	 * 
	 * @return The list of edges
	 */
	public LinkedList<SketchEdge> getEdges(){
		return _edges;
	}

	/**
	 * Returns the first edge of this path
	 */
	public SketchEdge getFirstEdge() {
		return _edges.getFirst();
	}

	/**
	 * Returns the last edge of this path
	 */
	public SketchEdge getLastEdge() {
		return _edges.getLast();
	}

	/**
	 * Returns the domain of the path
	 * 
	 * @return The domain of the path
	 */
	public EntityNode getDomain(){
		return _domain;
	}
	
	/**
	 * Returns the codomain of the path
	 * 
	 * @return The codomain of the path
	 */
	public EntityNode getCoDomain(){
		return _coDomain;
	}

	/**
	 * Returns the second-last EntityNode in the path.  In other words,
	 * this gives the source (domain) of the last edge that makes up
	 * the path.
	 *
	 * @return the second-last EntityNode along the path
	 */
	public EntityNode getLastDomain() {
		return _edges.getLast().getSourceEntity();
	}

	/**
	 * Returns the second EntityNode in the path.  In other words,
	 * this gives the target (codomain) of the first edge that makes
	 * up the path.
	 */
	public EntityNode getFirstCoDomain() {
		return _edges.getFirst().getTargetEntity();
	}

	/**
	 * Returns the entities along this path, in order from domain to codomain.
	 *
	 * @return List of EntityNodes
	 */
	public List<EntityNode> getEntities() {
		LinkedList<EntityNode> nodes = new LinkedList<EntityNode>();
		if (!_edges.isEmpty()) {
			nodes.add(_domain);
			for (SketchEdge edge: _edges)
				nodes.add(edge.getTargetEntity());
		}

		return Collections.unmodifiableList(nodes);
	}

	/**
	 * Returns true if this path is injective, that is, if the first edge of
	 * this path is injective.
	 */
	public boolean isInjective() {
		return _edges.getFirst().isInjective();
	}

	/**
	 * Method to determine whether two paths are equal, based on the edges that comprise the path.
	 * 
	 * @param a The first path
	 * @param b The second path
	 * @return True if the paths are equal, false otherwise
	 */
	public static boolean pathsAreEqual(SketchPath a, SketchPath b) {
		if (a.getEdges().size() != b.getEdges().size())
			return false; // Unequal length paths are obviously not equal

		Iterator pathA = a.getEdges().iterator(), pathB = b.getEdges().iterator();

		while (pathA.hasNext()) {
			if (pathA.next() != pathB.next()) {
				return false; // We found a different edge
			}
		}

		// We got through all the edges with no differences: the paths are equal.
		return true;
	}
}
