package easik.sketch.states;

import java.util.LinkedList;

import javax.swing.JOptionPane;

import easik.sketch.Sketch;
import easik.sketch.constraint.CommutativeDiagram;
import easik.sketch.constraint.Constraint;
import easik.sketch.path.SketchPath;
import easik.sketch.vertex.EntityNode;


/** 
 * This state makes a commutative diagram. When pushed, it will then push another
 * state to get a path. Once the first path is received, it will push another request for
 * a path. After adding the first path the options to either add another path or finish
 * the constraint are available.
 */
public class AddCommutativeDiagramState	extends SketchState implements PathAcceptingState {
	/**
	 * The paths involved in this constraint
	 */
	private LinkedList<SketchPath> _paths;

	/**
	 * Boolean to ascertain whether the selection of paths for this constraint has been finished
	 */
	private boolean _finished;

	/**
	 * If set, we delete this CD immediately before adding the new one.  Used when adding
	 * paths to an existing CD.
	 */
	private CommutativeDiagram _replace;
	
	/**
	 * Constructor for creating a new state to collect paths and make a diagram.
	 * 
	 * @param inSketch The sketch being worked with.
	 */
	public AddCommutativeDiagramState(Sketch inSketch) {
		super(inSketch);
		_finished = false;
		_paths = new LinkedList<SketchPath>();
	}

	/**
	 * Constructor for modifying an existing commutative diagram.
	 *
	 * @param inSketch The sketch being worked with
	 * @param existing The existing constraint to have paths added to
	 */
	public AddCommutativeDiagramState(Sketch inSketch, CommutativeDiagram existing) {
		super(inSketch);
		_finished = false;
		_paths = new LinkedList<SketchPath>();
		for (SketchPath path: existing.getPaths()) {
			_paths.add(new SketchPath(path.getEdges()));
		}
		_replace = existing;
	}

	/**
	 * Called when a new path has been completed by a state above this on the
	 * stack. If it is null, then cancel, otherwise add it to the collection.
	 * When two paths are in the collection, finish (if they share a domain and codomain).
	 * 
	 * @param inPath The last path in, null if it was a cancellation
	 */
	public void passPath(SketchPath inPath) {
		_finished = _ourSketch.getStateManager().getFinished();
		//Check if cancelled
		if (inPath == null) {
			_ourSketch.getSketchModel().cancelInsignificantUpdate();
			_ourSketch.getStateManager().popState();
			return;
		}

		//Check which path is being passed
		if(!_finished){
			_paths.add(inPath);
			if (_paths.size() > 1 && !CommutativeDiagram.isCommutativeDiagram(_paths)) {
				JOptionPane.showMessageDialog(_ourSketch.getParent(), 
						"Invalid or duplicated path selected.\nCommutative Diagram not created.", 
						"Error", JOptionPane.ERROR_MESSAGE);
				_ourSketch.getSketchModel().cancelInsignificantUpdate();
				_ourSketch.getStateManager().popState();
				return;
			}
			EntityNode domain = _paths.getFirst().getDomain(), codomain = _paths.getFirst().getCoDomain();
			_ourSketch.getFrame().setInstructionText("Select a path from '" + 
					domain.getName() + "' to '" + codomain.getName() +
					"' and click 'Next' to continue, or 'Finish' to add constraint.");
			_ourSketch.getStateManager().pushState(new GetPathState(true, true, _ourSketch, domain, codomain));

		}
		else {
			_paths.add(inPath);
			_ourSketch.getSketchModel().cancelInsignificantUpdate();

			_ourSketch.getSketchModel().beginUpdate();
			if(addDiagram()){
				// If we're replacing a CD, remove the old one:
				if (_replace != null) {
					_ourSketch.removeConstraint(_replace);
				}
			}
			else{
				JOptionPane.showMessageDialog(_ourSketch.getParent(), 
						"Invalid path selection.\nCommutative Diagram not created.", 
						"Error", JOptionPane.ERROR_MESSAGE);
			}

			_ourSketch.getSketchModel().endUpdate();
			_ourSketch.getStateManager().popState();
		}
	}

	/**
	 * Add the diagram to the sketch
	 * 
	 * @return true if the diagram was successfully added to the sketch, false otherwise
	 */
	private boolean addDiagram() {
		if (CommutativeDiagram.isCommutativeDiagram(_paths)){
			Constraint newDiagram = _replace != null
				? new CommutativeDiagram(_paths, _replace.getX(), _replace.getY(), true, _ourSketch)
				: new CommutativeDiagram(_paths, _ourSketch);
			_ourSketch.addNewConstraint(newDiagram);
			_ourSketch.setDirty();
			return true;
		}
		return false;
	}

	/**
	 * When this state is pushed on, it sends a message in a popup and then pushes
	 * a path collection state.
	 */
	public void pushedOn() {
		_ourSketch.clearSelection();
		_ourSketch.getStateManager().resetFinished();
		_ourSketch.getSketchModel().beginInsignificantUpdate();

		_ourSketch.getFrame().setInstructionText(
				_paths.size() >= 1
				? "Select a path from '" + _paths.getFirst().getDomain().getName() + "' to '" +
					_paths.getFirst().getCoDomain().getName() +  "' and click 'Next' to continue, or 'Finish' to add constraint."
				: "Select the first path and press 'Next'.");
		setNextButton(true);
		setFinishButton(_paths.size() >= 1);
		setCancelButton(true);
		if (_paths.size() >= 1) {
			// We're adding a path to the CD, so we have a required source and target
			EntityNode domain = _paths.getFirst().getDomain(), codomain = _paths.getFirst().getCoDomain();
			_ourSketch.getStateManager().pushState(new GetPathState(true, true, _ourSketch, domain, codomain));
		}
		else {
			_ourSketch.getStateManager().pushState(new GetPathState(true, false, _ourSketch));
		}
	}

	/**
	 * Nothing to do when it gets popped off, as it always pops itself off when completed. 
	 */
	public void poppedOff() {
	}

	/**
	 * This state is called "New Commutative diagram"
	 * 
	 * @return The string literal described above.
	 */
	public String toString() {
		return "New Commutative Diagram";
	}

}
