package easik.sketch.states;

import java.util.ArrayList;

import javax.swing.JOptionPane;

import easik.sketch.Sketch;
import easik.sketch.constraint.ConstraintException;
import easik.sketch.constraint.EqualizerConstraint;
import easik.sketch.path.SketchPath;
import easik.sketch.edge.*;
import easik.sketch.vertex.EntityNode;


/** 
 * This state makes an equalizer constraint. When pushed, it will then push another
 * state to get a path. Once the first path is received, it will push another request 
 * for a path. This repeated until three paths have been collected and then the 
 * constraint is created (as long as the paths are valid).
 */
public class AddEqualizerConstraintState extends SketchState implements PathAcceptingState {
	/** The paths involved in this constraint */
	private ArrayList<SketchPath> _paths;

	/** Stores whether the user has finished adding paths to this constraint */
	private boolean _finished;

	/** The current path number being selected */
	private int _round;
	
	/**
	 * Constructor for creating a new state to collect paths and make a diagram.
	 *
	 * @param inSketch The sketch being worked with.
	 */
	public AddEqualizerConstraintState(Sketch inSketch) {
		super(inSketch);
		_finished = false;
		_paths = new ArrayList<SketchPath>();
		_round = 0;
	}

	/**
	 * Called when a new path has been completed by a state above this on the
	 * stack. If it is null, then cancel, otherwise add it to the collection.
	 * When we get four paths, then we're finished (if they share a domain and codomain).
	 * 
	 * @param inPath The last path in, null if it was a cancellation
	 */
	public void passPath(SketchPath inPath) {
		// the button state for the next "get path" (compiler complained when not initialized)
		_finished = _ourSketch.getStateManager().getFinished();

		// set on invalid path selection
		String error = null;

		//Check if cancelled
		if (inPath == null) {
			_ourSketch.getSketchModel().cancelInsignificantUpdate();
			_ourSketch.getStateManager().popState();
			return;
		}
		
		_round++;
		if(_round == 1){
			if (inPath.getEdges().size() == 1 && inPath.getFirstEdge() instanceof InjectiveEdge) {
				_paths.add( inPath );
				_ourSketch.getFrame().setInstructionText("Select a path with source node '" + inPath.getCoDomain().getName() +"', and press 'Next'");
				_ourSketch.getStateManager().pushState(new GetPathState(true, false, _ourSketch, inPath.getCoDomain(), null));
			}
			else {
				error = "Invalid equalizer projection: selected edge was not an injective edge.";
			}
		}
		else if(_round == 2){
			if (!(inPath.getDomain() == _paths.get(0).getCoDomain()) ){
				error = "Invalid path selection: Selected path domain did not equal '" + _paths.get(0).getCoDomain().getName() + "'.";
			}
			else{
				_paths.add( inPath);
				EntityNode domain = _paths.get(1).getDomain(), codomain = _paths.get(1).getCoDomain();
				_ourSketch.getFrame().setInstructionText("Select another path from '" +
						domain.getName() + "' to '" + codomain.getName() +
						"', then press 'Next' to add more paths, or 'Finish' to create the equalizer.");
				_ourSketch.getStateManager().pushState(new GetPathState(true, true, _ourSketch, domain, codomain));
			}
		}
		else if (_round >= 3){
			ArrayList<SketchPath> testConstraint = new ArrayList<SketchPath>(_paths);
			testConstraint.add(inPath);
			if (!EqualizerConstraint.isEqualizerConstraint(testConstraint)) {
				error = "Invalid path selection: Path does not go from '" + _paths.get(1).getDomain().getName() +
					"' to '" + _paths.get(1).getCoDomain().getName() + "' or is not unique.";
			}
			else {
				_paths.add(inPath);
				EntityNode domain = _paths.get(1).getDomain(), codomain = _paths.get(1).getCoDomain();
				if (!_finished) {
					_ourSketch.getFrame().setInstructionText("Select another path from '" +
							domain.getName() + "' to '" + codomain.getName() +
							"', then press 'Next' to add more paths, or 'Finish' to create the equalizer.");
					_ourSketch.getStateManager().pushState(new GetPathState(true, true, _ourSketch, domain, codomain));
				}
			}
		}

		if (_finished) {
			_ourSketch.getSketchModel().cancelInsignificantUpdate();
			addDiagram();
			_ourSketch.getStateManager().popState();
			return;
		}

		// if error, tell user and pop this state.
		if(error != null){
			JOptionPane.showMessageDialog(_ourSketch.getParent(), 
					error + " Equalizer constraint not created.",
					"Error",
					JOptionPane.ERROR_MESSAGE);
			_ourSketch.getSketchModel().cancelInsignificantUpdate();
			_ourSketch.getStateManager().popState();
		}
	}

	/**
	 * Add the diagram to the sketch
	 *
	 * @return true if the constraint was successfully added to the sketch, false otherwise
	 */
	private boolean addDiagram() {
		if(EqualizerConstraint.isEqualizerConstraint(_paths)){
			try {
				EqualizerConstraint newDiagram = new EqualizerConstraint(_paths, _ourSketch);
				_ourSketch.addNewConstraint(newDiagram);
				_ourSketch.setDirty();
				return true;
			}
			catch (ConstraintException ce) {}
		}
		return false;
	}


	/**
	 * When this state is pushed on, it sends a message in a popup and then pushes
	 * a path collection state.
	 */
	public void pushedOn() {
		_ourSketch.clearSelection();
		_ourSketch.getStateManager().resetFinished();
		_ourSketch.getSketchModel().beginInsignificantUpdate();
		_ourSketch.getFrame().setInstructionText("Select an injective edge for the equalizer projection and click 'Next'.");

		_ourSketch.getFrame().getNextButton().setEnabled(true);
		_ourSketch.getFrame().getCancelButton().setEnabled(true);
		_ourSketch.getStateManager().pushState(new GetInjectivePathState(true, false, true, _ourSketch));
	}

	/**
	 * Nothing to do when it gets popped off, as it always pops itself off when completed. 
	 */
	public void poppedOff() {
	}

	/**
	 * This state is called "New Equalizer Constraint"
	 * 
	 * @return The string literal described above.
	 */
	public String toString() {
		return "New Equalizer Constraint";
	}

}

