package easik.sketch.states;

import java.util.ArrayList;

import javax.swing.JOptionPane;

import easik.sketch.Sketch;
import easik.sketch.vertex.EntityNode;
import easik.sketch.constraint.ConstraintException;
import easik.sketch.constraint.PullbackConstraint;
import easik.sketch.path.SketchPath;


/** 
 * This state makes a pullback constraint. When pushed, it will then push another
 * state to get a path. Once the first path is received, it will push another request 
 * for a path. This repeated until four paths have been collected and then the 
 * constraint is created (as long as the paths are valid).
 *
 * @author Kevin Green 2006
 * @author Vera Ranieri 2006
 * @since 2006-05-23 Kevin Green
 * @version 2006-08-15 Kevin Green
 */
public class AddPullbackConstraintState extends SketchState implements PathAcceptingState {
	/** The paths involved in this constraint */
	private ArrayList<SketchPath> _paths;

	/** Stores whether the user has finished adding paths to this constraint */
	private boolean _finished;

	/** The current path number being selected */
	private int _round;
	
	/**
	 * Constructor for creating a new state to collect paths and make a diagram.
	 *
	 * @param inSketch The sketch being worked with.
	 */
	public AddPullbackConstraintState(Sketch inSketch) {
		super(inSketch);
		_finished = false;
		_paths = new ArrayList<SketchPath>();
		_round = 0;
	}

	/**
	 * Called when a new path has been completed by a state above this on the
	 * stack. If it is null, then cancel, otherwise add it to the collection.
	 * When we get four paths, then we're finished (if they share a domain and codomain).
	 * 
	 * @param inPath The last path in, null if it was a cancellation
	 */
	public void passPath(SketchPath inPath) {
		_round++;

		// the button state for the next "get path"
		boolean next = _round < 3;
		boolean finished = _round == 3;
		// set on invalid path selection
		String error = null;

		//Check if cancelled
		if (inPath == null) {
			_ourSketch.getSketchModel().cancelInsignificantUpdate();
			_ourSketch.getStateManager().popState();
			return;
		}

		EntityNode codomain = inPath.getCoDomain();		

		if(_round == 1){
			_paths.add( inPath );
			_ourSketch.getFrame().setInstructionText("Select the second path to pullback codomain ('" + codomain.getName() +"'), and press 'Next'.");
			_ourSketch.getStateManager().pushState(new GetPathState(next, finished, _ourSketch, null, codomain));
		}
		else if(_round == 2){
			if (codomain != _paths.get(0).getCoDomain()) {
				error = "Invalid path selection: Selected codomain did not equal '" + _paths.get(0).getCoDomain().getName() + "'.";
			}
			else {
				_paths.add(inPath);

				boolean injective = _paths.get(1).isInjective();
				EntityNode target = _paths.get(0).getDomain();
				_ourSketch.getFrame().setInstructionText(
						"Select " + (injective ? "an injective" : "a") + " pullback projection with target node '" + 
						target.getName() + "', and press 'Next'");
				_ourSketch.getStateManager().pushState(
						injective
							? new GetInjectivePathState(next, finished, false, _ourSketch, null, target)
							: new GetPathState(next, finished, _ourSketch, null, target)
				);
			}
		}
		else if(_round == 3){
			if (codomain != _paths.get(0).getDomain()) {
				error = "Invalid path selection: Selected codomain did not equal '" + _paths.get(0).getDomain() + "'.";
			}
			else {
				_paths.add(inPath);

				boolean injective = _paths.get(0).isInjective();
				EntityNode source = inPath.getDomain();
				EntityNode target = _paths.get(1).getDomain();
				_ourSketch.getFrame().setInstructionText(
						"Select " + (injective ? "an injective" : "a") + " pullback projection from '" +
						source.getName() + "' to '" + target.getName() + "', and press 'Finish'");
				_ourSketch.getStateManager().pushState(
						injective
							? new GetInjectivePathState(next, finished, false, _ourSketch, source, target)
							: new GetPathState(next, finished, _ourSketch, source, target)
				);
			}
		}
		else{// last selection
			_paths.add( inPath );
			if( addDiagram() ) {
				JOptionPane.showMessageDialog(
						_ourSketch.getParent(),
						"Created a Pullback Constraint",
						"EASIK",
						JOptionPane.INFORMATION_MESSAGE);
				_ourSketch.getStateManager().popState();
				return;

			}
			else {
				error = "Invalid path selection: Selected codomain did not equal '" + _paths.get(1).getDomain() + "'.";
			}
		}

		// if error, tell user and pop this state.
		if(error != null){
			JOptionPane.showMessageDialog(_ourSketch.getParent(), 
					error + " Pullback constraint not created.",
					"Error",
					JOptionPane.ERROR_MESSAGE);
			_ourSketch.getSketchModel().cancelInsignificantUpdate();
			_ourSketch.getStateManager().popState();
		}
	}

	/**
	 * Add the diagram to the sketch
	 *
	 * @return true if the constraint was successfully added to the sketch, false otherwise
	 */
	private boolean addDiagram() {
		if(PullbackConstraint.isPullbackConstraint(_paths)){
			_ourSketch.getSketchModel().cancelInsignificantUpdate();
			try {
				PullbackConstraint newDiagram = new PullbackConstraint(_paths, _ourSketch);
				_ourSketch.addNewConstraint(newDiagram);
				_ourSketch.setDirty();
				return true;
			}
			catch (ConstraintException ce) {}
		}
		return false;
	}


	/**
	 * When this state is pushed on, it sends a message in a popup and then pushes
	 * a path collection state.
	 */
	public void pushedOn() {
		_ourSketch.clearSelection();
		_ourSketch.getStateManager().resetFinished();
		_ourSketch.getSketchModel().beginInsignificantUpdate();
		_ourSketch.getFrame().setInstructionText("Select the first path to the pullback codomain and click 'Next'.");

		_ourSketch.getFrame().getNextButton().setEnabled(true);
		_ourSketch.getFrame().getCancelButton().setEnabled(true);
		_ourSketch.getStateManager().pushState(new GetPathState(true, false, _ourSketch));
	}

	/**
	 * Nothing to do when it gets popped off, as it always pops itself off when completed. 
	 */
	public void poppedOff() {
	}

	/**
	 * This state is called "New Pullback Constraint"
	 * 
	 * @return The string literal described above.
	 */
	public String toString() {
		return "New Pullback Constraint";
	}

}

