package easik.sketch.states;

import java.util.ArrayList;

import javax.swing.JOptionPane;

import easik.sketch.Sketch;
import easik.sketch.constraint.Constraint;
import easik.sketch.constraint.SumConstraint;
import easik.sketch.path.SketchPath;


/** 
 * This state makes a sum constraint. When pushed, it will then push another
 * state to get a path. Once the first path is received, it will push another request for
 * a path. After adding the first path the options to either add another path or finish
 * the constraint are available.
 */
public class AddSumConstraintState extends SketchState implements PathAcceptingState {
	/** The paths involved in this constraint */
	private ArrayList<SketchPath> _paths;

	/** Stores whether the user has finished adding paths to this constraint */
	private boolean _finished;

	/**
	 * If set, we delete this constraint immediately before adding the new one.  Used when adding
	 * paths to an existing constraint.
	 */
	private SumConstraint _replace;

	/**
	 * Constructor for creating a new state to collect paths and make a diagram.
	 * 
	 * @param inSketch The sketch being worked with.
	 */
	public AddSumConstraintState(Sketch inSketch) {
		super(inSketch);
		_finished = false;
		_paths = new ArrayList<SketchPath>();
	}

	/**
	 * Constructor for modifying an existing constraint.
	 *
	 * @param inSketch The sketch being worked with
	 * @param existing The existing constraint to have paths added to
	 */
	public AddSumConstraintState(Sketch inSketch, SumConstraint existing) {
		super(inSketch);
		_finished = false;
		_paths = new ArrayList<SketchPath>();
		for (SketchPath path: existing.getPaths()) {
			_paths.add(new SketchPath(path.getEdges()));
		}
		_replace = existing;
	}

	/**
	 * Called when a new path has been completed by a state above this on the
	 * stack. If it is null, then cancel, otherwise check to see if the user
	 * selected the 'next' or 'finish' option.
	 * 
	 * @param inPath The last path in, null if it was a cancellation
	 */
	public void passPath(SketchPath inPath) {
		_finished = _ourSketch.getStateManager().getFinished();
		//Check if cancelled
		if (inPath == null) {
			_ourSketch.getSketchModel().cancelInsignificantUpdate();
			_ourSketch.getStateManager().popState();
			return;
		}
		
		//Check which path is being passed
		if(!_finished){
			_paths.add(inPath);
			// when we have enough paths, tell the user and leave instructions alone until either finish or cancel
			if( _paths.size() == 1){
				_ourSketch.getFrame().setInstructionText("Select a path beginning with an injective edge with target '" +
						_paths.get(0).getCoDomain().getName() + "' and press 'Next' to continue, or 'Finish' to add constraint.");
			}

			_ourSketch.getStateManager().pushState(
					new GetInjectivePathState(true, true, false, _ourSketch, null, _paths.get(0).getCoDomain()));
		}
		
		else {
			_paths.add(inPath);
			_ourSketch.getSketchModel().cancelInsignificantUpdate();

			_ourSketch.getSketchModel().beginUpdate();
			if(addDiagram()){
				// If we're replacing a constraint, remove the old one:
				if (_replace != null) {
					_ourSketch.removeConstraint(_replace);
				}
			}
			else{
				JOptionPane.showMessageDialog(
						_ourSketch.getFrame(),
						"Sum constraint could not be created.\n"
							+ "Please ensure that:\n"
							+ "  1) All edges involved target the same entity\n"
							+ "  2) The first edge in each path is injective\n"
							+ "  3) At least two edges are selected\n"
							+ "  4) All paths are unique",
						"Error",
						JOptionPane.ERROR_MESSAGE);
			}

			_ourSketch.getSketchModel().endUpdate();
			_ourSketch.getStateManager().popState();
		}
		
	}

	/**
	 * Add the diagram to the sketch
	 * @return true if the diagram was successfully added, false otherwise
	 */
	private boolean addDiagram() {
		if(SumConstraint.isSumConstraint(_paths)){
			Constraint newDiagram = _replace != null
				? new SumConstraint(_paths, _replace.getX(), _replace.getY(), true, _ourSketch)
				: new SumConstraint(_paths, _ourSketch);
			_ourSketch.addNewConstraint(newDiagram);
			_ourSketch.setDirty();
			return true;
		}
		return false;
	}


	/**
	 * When this state is pushed on, it sends a message in a popup and then pushes
	 * a path collection state.
	 */
	public void pushedOn() {
		_ourSketch.clearSelection();
		_ourSketch.getStateManager().resetFinished();
		_ourSketch.getSketchModel().beginInsignificantUpdate();

		_ourSketch.getFrame().setInstructionText(
				_paths.size() >= 1
				? "Select a path beginning with an injective edge with target '" + _paths.get(0).getCoDomain().getName() + "' and press 'Next' to continue, or 'Finish' to add constraint."
				: "Select a path beginning with an injective edge and then press 'Next' to add it to the constraint");
		setNextButton(true);
		setFinishButton(_paths.size() >= 1);
		setCancelButton(true);
		_ourSketch.getStateManager().pushState(
				new GetInjectivePathState(true, _paths.size() >= 1, false, _ourSketch,
					null, _paths.size() >= 1 ? _paths.get(0).getCoDomain() : null));
	}

	/**
	 * Nothing to do when it gets popped off, as it always pops itself off when completed. 
	 */
	public void poppedOff() {
	}

	/**
	 * This state is called "New Sum Constraint".
	 * 
	 * @return The string literal described above.
	 */
	public String toString() {
		return "New Sum Constraint";
	}

}

