package easik.sketch.states;

import java.util.ArrayList;

import easik.sketch.Sketch;
import easik.sketch.edge.*;
import easik.sketch.vertex.EntityNode;

/** 
 * The GetPathState is a state which can be used to retrieve a single injective edge
 * or a path beginning with an injective edge from the user.
 */
public class GetInjectivePathState extends GetPathState {

	private boolean _edgeOnly;

	/**
	 * Constructs a new GetInjectivePathState that can begin on any injective edge.
	 * 
	 * @param next Boolean determining whether the user should be allowed to select next on this round.
	 *   True if next could be selected, false otherwise.
	 * @param finish Boolean determining whether the user should be allowed to select finish on this round.  
	 *   True if finish could be selected, false otherwise.
	 * @param edgeOnly Boolean determining whether the user should be selecting just a single, injective
	 *   edge (true), or any path beginning with an injective edge (false).
	 * @param inSketch The sketch
	 */
	public GetInjectivePathState(boolean next, boolean finish, boolean edgeOnly, Sketch inSketch) {
		this(next, finish, edgeOnly, inSketch, null, null);
	}

	/**
	 * Constructs a new GetInjectivePathState that can begin on any injective edge out of the
	 * desired source node, and must end at the specified target node.
	 * 
	 * @param next Boolean determining whether the user should be allowed to select next on this round.
	 *   True if next could be selected, false otherwise.
	 * @param finish Boolean determining whether the user should be allowed to select finish on this round.  
	 *   True if finish could be selected, false otherwise.
	 * @param edgeOnly Boolean determining whether the user should be selecting just a single, injective
	 *   edge (true), or any path beginning with an injective edge (false).
	 * @param sourceNode EntityNode specifying the node at which the path must start (null if it can start
	 *   on any injective edge).
	 * @param targetNode EntityNode specifying the node at which the path must end before the user can
	 *   click next or finish.
	 * @param inSketch The sketch
	 */
	public GetInjectivePathState(boolean next, boolean finish, boolean edgeOnly, Sketch inSketch, EntityNode sourceNode, EntityNode targetNode) {
		super(next, finish, inSketch, sourceNode, targetNode);
		_edgeOnly = edgeOnly;
	}

	/**
	 * Returns the edges that are to be selectable.  Only edges extending from
	 * the current path are selectable; if there is no path, then any edge
	 * qualifies.
	 */
	public Object[] getSelectables() {
		Object[] newSelection = _ourSketch.getSelectionCells();

		if (newSelection.length == 1 && _edgeOnly) {
			return new Object[0]; // Nothing else is selectable, the user is allowed just the one edge
		}

		if (newSelection.length > 0) {
			return super.getSelectables();
		}

		ArrayList<SketchEdge> selectable = new ArrayList<SketchEdge>();

		// We use what GetPathState gives us, but remove anything that isn't injective:
		for (Object o: super.getSelectables()) {
			SketchEdge edge = (SketchEdge) o;
			if (edge.isInjective())
				selectable.add(edge);
		}

		return selectable.toArray();
	}
}
