package easik.sketch.states;



import easik.sketch.Sketch;


/** 
 *	This is the base state. It contains stubs and abstract methods for allowing a 
 *	class to be treated like a state, and to take part in the State Manager
 *
 * @author Rob Fletcher 2005
 * @author Vera Ranieri 2006
 * @author Kevin Green 2006
 * @version 2006-08-02 Kevin Green
 */
public abstract class SketchState {
	/**
	 * Hook to record when selection has been updated
	 */
	public void selectionUpdated() {};		
	/**
	 * Hook to record when state has been pushed on
	 *
	 */
	public abstract void pushedOn();	
	/**
	 * Hook to record when state has been popped off
	 *
	 */
	public abstract void poppedOff();	
	/**
	 * @return The string representation of this state
	 */
	public abstract String toString();
	/**
	 * Records state of the next button
	 */
	private boolean _nextOn = false;
	/**
	 * Records the state of the cancel button
	 */
	private boolean _cancelOn = false;	
	/**
	 * Records the state of the finish button
	 */
	private boolean _finishOn = false;
	/**
	 * The current sketch
	 */
	protected Sketch _ourSketch;

	public SketchState(Sketch inSketch){
		_ourSketch = inSketch;
	}
	/**
	 * Returns selectable cells.  The default is to return all graph cells.
	 */
	public Object[] getSelectables() {
		return _ourSketch.getRoots();
	}

	/**
	 * When this state gets the top spot again, this gets executed. Restores buttons
	 * and labels. If overridden, be sure to at call super.gotfocus() first.
	 */
	public void gotFocus() {
		setCancelButton( _cancelOn );
		setNextButton( _nextOn );
		setFinishButton(_finishOn);
	}
	
	/**
	 * Optional hook for when the cancel button is clicked
	 *
	 */
	public void cancelClicked() {
	}
	
	/**
	 * Optional hook for when the next button is clicked
	 *
	 */
	public void nextClicked() {
	}
	
	/**
	 * Optional hook for when the finish button is clicked.
	 * 
	 * @since 2006-05-30 Vera Ranieri
	 */
	public void finishClicked(){	
	}
	
	/**
	 * Use this function to toggle the usability of the cancel button. It
	 * remembers the state in a separate boolean so it can be restored.
	 * 	 
	 * @param inButton Status of the button
	 */
	protected void setCancelButton( boolean inButton ) {
		_cancelOn = inButton;
		_ourSketch.getFrame().getCancelButton().setEnabled( inButton );
		if(_cancelOn || _nextOn || _finishOn)
			_ourSketch.getFrame().setButtonPaneVisibility(true);
		else
			_ourSketch.getFrame().setButtonPaneVisibility(false);
	}
	
	/**
	 * Use this function to toggle the usability of the ok button. It
	 * remembers the state in a separate boolean so it can be restored.
	 * 	 
	 * @param inButton Status of the button
	 */
	protected void setNextButton( boolean inButton ) {
		_nextOn = inButton;
		_ourSketch.getFrame().getNextButton().setEnabled( inButton );
		if(_cancelOn || _nextOn || _finishOn)
			_ourSketch.getFrame().setButtonPaneVisibility(true);
		else
			_ourSketch.getFrame().setButtonPaneVisibility(false);
	}	
	
	/**
	 * Use this function to toggle the usability of the finish button. It
	 * remembers the state in a separate boolean so it can be restored.
	 * 
	 * @param inButton Status of the button
	 * @since 2006-05-30 Vera Ranieri
	 */
	protected void setFinishButton(boolean inButton){
		_finishOn = inButton;
		_ourSketch.getFrame().getFinishButton().setEnabled( inButton );
		if(_cancelOn || _nextOn || _finishOn)
			_ourSketch.getFrame().setButtonPaneVisibility(true);
		else
			_ourSketch.getFrame().setButtonPaneVisibility(false);
	}
}
