package easik.sketch.vertex;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Collection;
import java.util.Collections;
import java.util.Arrays;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.undo.AbstractUndoableEdit;

import easik.database.types.*;
import easik.sketch.Sketch;
import easik.sketch.constraint.*;
import easik.ui.SketchFrame;
import easik.sketch.attribute.EntityAttribute;
import easik.sketch.keys.UniqueIndexable;
import easik.sketch.keys.UniqueKey;
import easik.sketch.edge.SketchEdge;
import easik.sketch.path.SketchPath;
import easik.sketch.util.graph.SketchGraphModel;
import easik.database.types.Type;


/**
 * An entity node represents a table in a database. It has a name,
 * attributes, and unique keys. This class keeps track of all these
 * elements. 
 */
public class EntityNode extends SketchVertex {
	/** List of all attributes for this entity */
	private ArrayList<EntityAttribute> _entityAttributes = new ArrayList<EntityAttribute>();

	/** List of all unique keys for this entity */
	private ArrayList<UniqueKey> _uniqueKeys = new ArrayList<UniqueKey>();

	/** Node representing this object in the information tree */
	private DefaultMutableTreeNode _treeNode;

	/** Node representing the attributes for the information tree */
	private DefaultMutableTreeNode _attribNode = new DefaultMutableTreeNode("Attributes");

	/** Node representing the key for the information tree */
	private DefaultMutableTreeNode _keyNode = new DefaultMutableTreeNode("Keys");
	
	/** The set of constraints this Node is currently in */
	private LinkedHashSet<Constraint> _constraints = new LinkedHashSet<Constraint>();
	
	/**
	 * Creates a new entity node with the name provided at position (0,0).
	 * 
	 * @param nodeName The name of the new node
	 */
	public EntityNode(String nodeName, Sketch inSketch) {
		this(nodeName, 0, 0, inSketch);
	}
	
	/**
	 * Creates a new enity node with the name provided.
	 * Stores visual representation information.
	 * 
	 * @param nodeName Name of the new node
	 * @param x X Coordinate of the new node
	 * @param y Y Coordinate of the new node
	 */
	public EntityNode(String nodeName, int x, int y, Sketch inSketch) {
		super(nodeName, x, y, inSketch);
		_treeNode = new DefaultMutableTreeNode(this);
		_treeNode.add(_attribNode);
		_treeNode.add(_keyNode);
	}

	/**
	 * Changes the name of the EntityNode.  This also updates the name in the
	 * EntityNode's SketchFrame, to maintain consistency.  Note that if you
	 * attempt to rename an EntityNode to one that already exists in the same
	 * frame, a trailing number will be added to make the name unique.  This
	 * means, however, that the name might not end up being what you set--if
	 * that's a problem, make sure the new name isn't taken yet yourself, via
	 * SketchFrame.getEntity().
	 *
	 * @see easik.ui.SketchFrame.getEntity(String)
	 * @param newName the new name to attempt to set.  The actual name set may
	 * 				  have a trailing number appended/incremented.
	 */
	public void setName(String newName) {
		String oldName = getName();
		// The name might have to change, if there are duplicates.  nodeRenamed tells us the final name to use:
		newName = getSketch().nodeRenamed(this, oldName, newName);
		super.setName(newName);
	}

	/**
	 * Creates a new EntityAttribute and add its to the list of attributes
	 * 
	 * @param inName Name of attribute
	 * @param inAttributeType SQL Type of the attribute
	 */
	public void addEntityAttribute(String inName, Type inAttributeType){
		addEntityAttribute(new EntityAttribute(inName, inAttributeType, this));
	}
	
	/**
	 * Creates a new EntityAttribute and add its to the list of attributes
	 * 
	 * @param inAtt	The EntityAttribute to add to this EntityNode.
	 */
	public void addEntityAttribute(final EntityAttribute inAtt){
		_entityAttributes.add(inAtt);
		getAttributeNode().add(inAtt);
		getSketch().refresh();
		getSketch().getFrame().getInfoTreeUI().refreshTree(this);

		getSketch().getSketchModel().postEdit(new AbstractUndoableEdit() {
			public void undo() {
				super.undo();
				_entityAttributes.remove(inAtt);
				getAttributeNode().remove(inAtt);
				getSketch().refresh();
				getSketch().getFrame().getInfoTreeUI().refreshTree(EntityNode.this);
			}
			public void redo() {
				super.redo();
				_entityAttributes.add(inAtt);
				getAttributeNode().add(inAtt);
				getSketch().refresh();
				getSketch().getFrame().getInfoTreeUI().refreshTree(EntityNode.this);
			}
		});
	}
	
	/**
	 * Returns the list of EntityAttributes
	 * 
	 * @return List of EntityAttributes
	 */
	public List<EntityAttribute> getEntityAttributes(){
		return Collections.unmodifiableList(_entityAttributes);
	}
	
	/**
	 * Removes an attribute from the list
	 * 
	 * @param inAttribute The attribute to be removed
	 */
	public void removeEntityAttribute(final EntityAttribute inAttribute){
		SketchGraphModel model = getSketch().getSketchModel();
		model.beginUpdate();

		final int attPos = _entityAttributes.indexOf(inAttribute);
		final int nodePos = getAttributeNode().getIndex(inAttribute);
		_entityAttributes.remove(inAttribute);
		getAttributeNode().remove(inAttribute);

		getSketch().refresh();
		getSketch().getFrame().getInfoTreeUI().refreshTree(this);

		getSketch().getSketchModel().postEdit(new AbstractUndoableEdit() {
			public void undo() {
				super.undo();
				_entityAttributes.add(attPos, inAttribute);
				getAttributeNode().insert(inAttribute, nodePos);
				getSketch().refresh();
				getSketch().getFrame().getInfoTreeUI().refreshTree(EntityNode.this);
			}
			public void redo() {
				super.redo();
				_entityAttributes.remove(inAttribute);
				getAttributeNode().remove(inAttribute);
				getSketch().refresh();
				getSketch().getFrame().getInfoTreeUI().refreshTree(EntityNode.this);
			}
		});

		// Remove references to the attribute from unique keys
		for (UniqueKey curKey: _uniqueKeys) {
			curKey.removeElement(inAttribute);
		}

		// Fix up any empty/duplicate unique keys resulting from the attribute removal
		UniqueKey.cleanup(this);

		model.endUpdate();
	}

	/**
	 * Adds a unique key to the list
	 * 
	 * @param inAtts The attributes used in the key
	 * @param inName The name of the unique key
	 */
	public void addUniqueKey(ArrayList<UniqueIndexable> inAtts, String inName) {
		addUniqueKey(new UniqueKey(this, inAtts, inName));
	}

	/**
	 * Adds a unique key to the list
	 * 
	 * @param inKey The key to be added
	 */
	public void addUniqueKey(final UniqueKey inKey) {
		_uniqueKeys.add(inKey);
		getKeyNode().add(inKey);
		getSketch().refresh();
		getSketch().getFrame().getInfoTreeUI().refreshTree(this);

		getSketch().getSketchModel().postEdit(new AbstractUndoableEdit() {
			public void undo() {
				super.undo();
				_uniqueKeys.remove(inKey);
				getKeyNode().remove(inKey);
				getSketch().refresh();
				getSketch().getFrame().getInfoTreeUI().refreshTree(EntityNode.this);
			}
			public void redo() {
				super.redo();
				_uniqueKeys.add(inKey);
				getKeyNode().add(inKey);
				getSketch().refresh();
				getSketch().getFrame().getInfoTreeUI().refreshTree(EntityNode.this);
			}
		});
	}

	/**
	 * Removes a unique key from the list
	 * 
	 * @param inKey The unique key to be removed
	 */
	public void removeUniqueKey(final UniqueKey inKey){
		final int keyPos = _uniqueKeys.indexOf(inKey);
		final int nodePos = getKeyNode().getIndex(inKey);
		_uniqueKeys.remove(inKey);
		getKeyNode().remove(inKey);
		getSketch().refresh();
		getSketch().getFrame().getInfoTreeUI().refreshTree(this);

		getSketch().getSketchModel().postEdit(new AbstractUndoableEdit() {
			public void undo() {
				super.undo();
				_uniqueKeys.add(keyPos, inKey);
				getKeyNode().insert(inKey, nodePos);
				getSketch().refresh();
				getSketch().getFrame().getInfoTreeUI().refreshTree(EntityNode.this);
			}
			public void redo() {
				super.redo();
				_uniqueKeys.remove(inKey);
				getKeyNode().remove(inKey);
				getSketch().refresh();
				getSketch().getFrame().getInfoTreeUI().refreshTree(EntityNode.this);
			}
		});
	}

	/**
	 * Returns a List of the unique keys
	 * 
	 * @return a List of the unique keys
	 */
	public List<UniqueKey> getUniqueKeys(){
		return Collections.unmodifiableList(_uniqueKeys);
	}

	/**
	 * Returns a Set of SketchEdges of the outgoing edges of this EntityNode.
	 * Note that edges for constraints are <b>not</b> included.
	 *
	 * @return Set of outgoing edges from this EntityNode
	 */
	public Set<SketchEdge> getOutgoingEdges() {
		List outEdges = _theSketch.getGraphLayoutCache().getOutgoingEdges(this, Collections.emptySet(), false, true);
		LinkedHashSet<SketchEdge> outgoing = new LinkedHashSet<SketchEdge>();
		for (Object out: outEdges) {
			if (out instanceof SketchEdge)
				outgoing.add((SketchEdge) out);
		}

		return outgoing;
	}

	/**
	 * Returns a Set of UniqueIndexable-implementing outgoing edges of this
	 * node.
	 */
	public Set<UniqueIndexable> getIndexableEdges() {
		Set<SketchEdge> edges = getOutgoingEdges();
		LinkedHashSet<UniqueIndexable> attribs = new LinkedHashSet<UniqueIndexable>();
		for (SketchEdge edge: edges) {
			if (edge instanceof UniqueIndexable)
				attribs.add((UniqueIndexable) edge);
		}

		return attribs;
	}

	/**
	 * Returns the tree node used to display this EntityNode
	 */
	public DefaultMutableTreeNode getTreeNode() {
		return _treeNode;
	}
	/**
	 * Returns the tree node used to display unique keys
	 * 
	 * @return The tree node used to display unique keys
	 */
	public DefaultMutableTreeNode getKeyNode(){
		return _keyNode;
	}

	public DefaultMutableTreeNode getAttributeNode() {
		return _attribNode;
	}

	/**
	 * Add a constraint to the set of constraints of which this node is a part
	 */
	public void addConstraint(final Constraint con){
		boolean added = _constraints.add(con);

		if (added) {
			getSketch().getSketchModel().postEdit(new AbstractUndoableEdit() {
				public void undo() { super.undo(); _constraints.remove(con); }
				public void redo() { super.redo(); _constraints.add(con); }
			});
		}
	}

	/**
	 * Removes a constraint from the set of constraints this node believes to be in
	 */
	public boolean removeConstraint(final Constraint con){
		boolean removed = _constraints.remove(con);

		if (removed) {
			getSketch().getSketchModel().postEdit(new AbstractUndoableEdit() {
				public void undo() { super.undo(); _constraints.add(con); }
				public void redo() { super.redo(); _constraints.remove(con); }
			});
		}

		return removed;
	}

	/**
	 * Gets the set of constraints of which this node is a part
	 */
	public Set<Constraint> getConstraints(){
		return Collections.unmodifiableSet(_constraints);
	}

	/**
	 * Returns the edges that might need to be "shadowed".  This is used by various
	 * constraints where we need to do automatic path insertion.  For example, sum constraints,
	 * where we have something like A -&gt; B -&gt; C, where A -&gt; B is the sum, but
	 * because B is auto-created by the insertion into A, we have to specify the B -&gt; C
	 * foreign key when inserting into A.
	 *
	 * @param EntityNode table
	 * @return LinkedList of SketchEdges that are not-null (in other words, non-partial edges)
	 */
	public LinkedList<SketchEdge> getNonPartialEdges() {
		LinkedList<SketchEdge> edges = new LinkedList<SketchEdge>();
		for (SketchEdge e: getOutgoingEdges()) {
			if (!e.isPartial()) {
				edges.add(e);
			}
		}
		return edges;
	}

	/**
	 * This method returns the edges that will be "shadowed" in this entity for
	 * allowing various types of constraints.  The problem arises when we have something like:
	 * A -&gt; B -&gt; C, where A is the summand of B, but B has to be specified.
	 * In this case, the B to C edge will be returned as a "shadow" edge.
	 * We handle this for other constraint types, too.  For a good, working, technical
	 * example, see the shadowEdges.easik sample sketch.
	 *
	 * @return a set of edges that will be shadowed by this entity node.
	 */
	public LinkedHashSet<SketchEdge> getShadowEdges() {
		LinkedHashSet<SketchEdge> shadowEdges = getShadowEdges(new LinkedHashSet<EntityNode>(), new LinkedHashSet<SketchEdge>());
		return shadowEdges;
	}

	// Package-only implementation of the above that breaks recursion by ignoring
	// shadowed nodes that we already know about.
	LinkedHashSet<SketchEdge> getShadowEdges(LinkedHashSet<EntityNode> ignore, LinkedHashSet<SketchEdge> constraintEdges) {
		// These are the other entity node that we (potentially) need to shadow:
		LinkedHashSet<EntityNode> shadow = new LinkedHashSet<EntityNode>();

		CONSTRAINT: for (Constraint c: getSketch().getConstraints()) {
			if (c instanceof SumConstraint) {
				SumConstraint s = (SumConstraint) c;
				for (SketchPath path: s.getPaths()) {
					// If this entity is the domain of a sum constraint path,
					// we need to include every entity along that path (except for this entity, of course):
					if (path.getDomain() == this) {
						shadow.addAll(path.getEntities());
						constraintEdges.addAll(path.getEdges());

						continue CONSTRAINT;
					}
				}
			}
			else if (c instanceof ProductConstraint) {
				ProductConstraint p = (ProductConstraint) c;
				for (SketchPath path: p.getPaths()) {
					// If this entity is the codomain of a product constraint path,
					// we need to include every entity, excluding the codomains), along each product path
					if (path.getCoDomain() == this) {
						for (SketchPath prodPath: p.getPaths()) {
							// But we ignore all of the product path edges, since they will be automatically generated:
							constraintEdges.addAll(prodPath.getEdges());

							LinkedList<EntityNode> pathNodes = new LinkedList<EntityNode>();
							pathNodes.addAll(prodPath.getEntities());
							pathNodes.removeLast();
							shadow.addAll(pathNodes);
						}

						continue CONSTRAINT;
					}
				}
			}
			else if (c instanceof EqualizerConstraint) {
				EqualizerConstraint e = (EqualizerConstraint) c;
				// If this entity is the domain of an equalizer, we need to shadow the projection entity.
				// We actually shadow every node along the projection path, which will handle multi-edge
				// projection paths, though EASIK doesn't actually support those (primarily because deletion
				// is needed; an injective *path* is fine for creation, but cannot be cleanly deleted--an
				// injective *edge* doesn't cause that problem).
				if (e.getSourceEntity() == this) {
					SketchPath projection = e.getProjection();
					shadow.addAll(projection.getEntities());
					
					// Ignore the projection edge itself:
					constraintEdges.addAll(projection.getEdges());

					continue CONSTRAINT;
				}
			}
			else if (c instanceof PullbackConstraint) {
				PullbackConstraint pb = (PullbackConstraint) c;
				// A pullback is a little trickier than the above: any insertion into one
				// of the pullback path domains (i.e. projection codomains) means we have
				// to shadow all the entities along *both* projection paths.
				SketchPath projA = pb.getProjectionA(), projB = pb.getProjectionB();
				if (this == projA.getCoDomain() || this == projB.getCoDomain()) {
					LinkedList<EntityNode> projANodes = new LinkedList(projA.getEntities());
					projANodes.removeLast(); // The codomain of the projection should *not* be included, as it won't be generated
					LinkedList<EntityNode> projBNodes = new LinkedList(projB.getEntities());
					projBNodes.removeLast(); // The codomain of the projection should *not* be included, as it won't be generated

					shadow.addAll(projANodes);
					shadow.addAll(projBNodes);

					constraintEdges.addAll(projA.getEdges());
					constraintEdges.addAll(projB.getEdges());

					continue CONSTRAINT;
				}
			}
		}

		LinkedHashSet<SketchEdge> shadowEdges = new LinkedHashSet<SketchEdge>();

		// All of the ignore entities, plus everything we just found should be
		// ignored by any recursion:
		LinkedHashSet<EntityNode> toIgnore = new LinkedHashSet<EntityNode>();
		toIgnore.add(this);
		toIgnore.addAll(ignore);
		toIgnore.addAll(shadow);

		for (EntityNode node: shadow) {
			// If it's ourself (which obviously doesn't need to be shadowed), or a node
			// that a call up the stack is already shadowing the node, ignore it:
			if (node == this || ignore.contains(node))
				continue;

			// Otherwise, shadow its non-partial edges, and all of its shadow edges:
			shadowEdges.addAll(node.getShadowEdges(toIgnore, constraintEdges));
			shadowEdges.addAll(node.getNonPartialEdges());
			shadowEdges.removeAll(constraintEdges); // Remove edges already involved in the sum
		}
		return shadowEdges;
	}

	/**
	 * Determines whether a set of attributes would enforce new uniqueness on
	 * this entity, or whether uniqueness of those attributes is already
	 * enforced by an existing unique key.  Essentially, this boils down to two
	 * things: there can be no existing unique key that is a duplicate of the
	 * specified elements, and there can be no unique key that is a subset of
	 * the specified elements.  The latter is because if a unique key on (A, B)
	 * already exists, then (A, B, C) will be already guaranteed to be unique
	 * because the (A, B) part has to be unique.
	 *
	 * @param inAtts The list/set/collection/array of attributes/edges to check
	 * @return the unique key that already exists that enforces the uniqueness, or
	 *		null if the specified items would form new uniqueness.
	 */
	public UniqueKey uniqueKeyOn(Collection<UniqueIndexable> inElems) {
		return uniqueKeyOn(new HashSet<UniqueIndexable>(inElems));
	}
	/**
	 * Array or multi-value version of uniqueKeyOn(...)
	 */
	public UniqueKey uniqueKeyOn(UniqueIndexable... inElems) {
		return uniqueKeyOn(Arrays.asList(inElems));
	}
	/**
	 * Version of uniqueKeyOn() that takes a set of UniqueIndexable edges.
	 */
	public UniqueKey uniqueKeyOn(Set<UniqueIndexable> inElemSet) {
		for (UniqueKey q: getUniqueKeys()) {
			if (inElemSet.containsAll(q.getElements()))
				return q;
		}

		return null;
	}
}
