package easik.ui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.Point;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowFocusListener;
import java.awt.event.WindowEvent;
import java.util.Arrays;
import java.util.List;

import javax.swing.*;

import easik.*;
import easik.overview.Overview;
import easik.overview.vertex.ViewNode;
import easik.sketch.Sketch;
import easik.sketch.states.*;
import easik.ui.menu.*;
import easik.ui.menu.popup.*;
import easik.ui.tree.popup.*;
import easik.ui.tree.ViewInfoTreeUI;
import easik.view.*;
import easik.os.EasikMenuAdjuster;

/** 
 * This is the swing starting point for the program. It contains a sketch,
 * through which the user can design a database schema, and once satisfied,
 * connect to a MySQL server and manipulate the data contained therein. 
 * 
 * @author Rob Fletcher 2005
 * @author Kevin Green 2006
 * @author Vera Ranieri 2006
 * @version 2006-08-04 Kevin Green
 */
public class ViewFrame extends EasikFrame {

	/** Popup menu */
	private JPopupMenu _editModePopMenu, _manipModePopMenu;

	/** The point where the popup menu should appear */
	private Point _popupPosition;

	/** The current sketch */
	private View _ourView;

	/** Popup menu items */
	private JMenuItem _ViewQueryPopItem;
	private JMenuItem _AddQueryNodePopItem;
	private JMenuItem _DeletePopItem;
	private JMenuItem _DefineQueryNodePopItem;

	/** The right and main panels of the frame */
	private JPanel _rightPane;
	private JPanel _mainPane;

	/** The scroll pane of the view */
	private JScrollPane _sketchScrollPane;

	/** The information tree */
	private ViewInfoTreeUI _infoTreeUI;

	/** The split pane */
	private JSplitPane _mainSplitPane;

	/** The ViewNode representing this frame */
	private ViewNode _myNode;
	
	/**
	 *	Creates a view frame, sets some sizes, sets up the properties
	 *	builds the menus, and lays out the swing components.
	 *
	 * @param inOverview The overivew in which this frame's sketch will exist
	 * @param inSketch The sketch to which our new view points
	 */
	public ViewFrame(Overview inOverview, Sketch inSketch) {
		super("EASIK - Untitled");

		int defaultWidth = Integer.parseInt(_settings.getProperty("view_display_width", "700"));
		int defaultHeight = Integer.parseInt(_settings.getProperty("view_display_height", "500"));
		int defaultLocationX = Integer.parseInt(_settings.getProperty("view_frame_location_x", "0"));
		int defaultLocationY = Integer.parseInt(_settings.getProperty("view_frame_location_y", "25"));
		this.setLocation(defaultLocationX, defaultLocationY);
		this.setSize(defaultWidth, defaultHeight);

		if (ApplicationFrame.EASIK_Icon != null)
			this.setIconImage(ApplicationFrame.EASIK_Icon);

		//Initialize all variables
		_editModePopMenu = new JPopupMenu();
		_manipModePopMenu = new JPopupMenu();
		_popupPosition = new Point(0, 0);
		_rightPane = new JPanel();
		_mainPane = new JPanel();
		_ourView = new View(this, inSketch, inOverview);
		_infoTreeUI = new ViewInfoTreeUI(this);
		
		_infoTreeUI.refreshTree();

		//Build Menus
		buildMenu();
		buildPopupMenus();
		
		//Setup View Pane
		_sketchScrollPane = new JScrollPane(_ourView);
		_sketchScrollPane.setMinimumSize(new Dimension(300, 300));
		
		//Setup entire right pane
		_rightPane.setLayout(new BorderLayout());
		_treeName = new JLabel(_ourView.getDocInfo().getName());
		_rightPane.add(_treeName, BorderLayout.NORTH);
		_rightPane.add(_infoTreeUI, BorderLayout.CENTER);
		
		//Setup main pane
		_mainPane.setLayout(new BorderLayout());
		_mainPane.add(_sketchScrollPane, BorderLayout.CENTER);
		
		//Setup enitre window
		_mainSplitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, _mainPane, _rightPane);
		int dividerPos = Integer.parseInt(_settings.getProperty("view_divider_position", "" + (defaultWidth - 255)));
		_mainSplitPane.setDividerLocation(dividerPos);
		_mainSplitPane.setDividerSize(10);
		_mainSplitPane.setResizeWeight(1);
		_mainSplitPane.setOneTouchExpandable(true);
		_mainSplitPane.setContinuousLayout(true);
		getContentPane().add(_mainSplitPane, BorderLayout.CENTER);

		// Update the display in the overview whenever we lost focus
		addWindowFocusListener(new WindowFocusListener() {
			public void windowGainedFocus(WindowEvent e) {}
			public void windowLostFocus(WindowEvent e) {
				_ourView.updateThumb();
				_ourView.getOverview().refresh();
			}
		});
	}

	/**
	 * Forces a refresh of our title bar when we set visible
	 */
	public void setVisible(boolean vis){
		this.setEasikTitle();
		super.setVisible(vis);
	}

	/**
	 * Trys to close window as long it is not dirty.
	 */
	public void closeWindow(){
		// We're exiting, so update the width/height of the main window, and save the settings
		_settings.setProperty("view_display_width", "" + getWidth());
		_settings.setProperty("view_display_height", "" + getHeight());
		_settings.setProperty("view_divider_position", "" + _mainSplitPane.getDividerLocation());
		_settings.setProperty("view_frame_location_x", "" + getX());
		_settings.setProperty("view_frame_location_y", "" + getY());

		_settings.store();

		// Refresh displayed thumbnail
		_ourView.updateThumb();
		_ourView.getOverview().refresh();

		this.setVisible(false);
	}
	
	/**
	 * Returns the current sketch
	 * 
	 * @return The sketch
	 */
	public View getView(){
		return _ourView;
	}

	/**
	 * Returns the Overview
	 * 
	 * @return The overview
	 */
	public Overview getOverview() {
		return _ourView.getOverview();
	}

	/**
	 * Returns the InfoTreeUI object
	 * 
	 * @return The InfoTreeUI object
	 */
	public ViewInfoTreeUI getInfoTreeUI(){
		return _infoTreeUI;
	}
	
	/**
	 * Sets the InfoTreeUI
	 * 
	 * @param inInfoTreeUI The new InfoTreeUI
	 */
	public void setInfoTreeUI(ViewInfoTreeUI inInfoTreeUI){
		_rightPane.remove(_infoTreeUI);
		_infoTreeUI = inInfoTreeUI;
		_rightPane.add(_infoTreeUI);
		_mainSplitPane.setDividerLocation(_mainSplitPane.getDividerLocation());
	}
	
	/**
	 * Popuplates the popup menus with new menu items.
	 */
	private void buildPopupMenus() {

		_manipModePopMenu.add(_ViewQueryPopItem = new JMenuItem(new ViewQueryAction(_ourView)));

		_editModePopMenu.add(_AddQueryNodePopItem = new JMenuItem(new NewQueryNodeAction(_popupPosition, this)));
		_editModePopMenu.add(_DeletePopItem = new JMenuItem(new DeleteFromViewAction(this)));
		_editModePopMenu.add(_DefineQueryNodePopItem = new JMenuItem(new DefineQueryNodeAction(this)));

		_ourView.addMouseListener(new MouseAdapter() {
			public void mousePressed(MouseEvent evt) {
				mouseReleased(evt);
			}
			public void mouseReleased(MouseEvent evt) {
				if (evt.isPopupTrigger()) {
					Object rightClickedOn = _ourView.getFirstCellForLocation(evt.getX(),evt.getY());
					// When we have a right click, we check to see if it is part of the selection. If 
					// it isn't, we change the selection to include only that which was
					// under the click, otherwise we'll trigger the popup for the current selection
					if ( rightClickedOn != null){
						Object[] selection = _ourView.getSelectionCells();
						if(!Arrays.asList(selection).contains(rightClickedOn)){
							_ourView.setSelectionCell(_ourView.getFirstCellForLocation(evt.getX(),evt.getY()));
						}
					}
					else{
						_ourView.setSelectionCells(new Object[0]);
					}

					_popupPosition.setLocation(evt.getX(), evt.getY());
					JPopupMenu menu = setPopMenu();
					menu.show(evt.getComponent(), evt.getX(), evt.getY());
				}
			}
		});
	}

	/**
	 * Figures out which popup menu should be displayed, based on the editing
	 * mode of the sketch on which this is a view. Ensures the the appropriate
	 * menu items are enabled based on the current selection. 
	 * @return The popup menu that should be displayed. (i.e. the editing popup
	 * if the sketch of which this view in on is in edit mode and the manipulate
	 * option otherwise.)
	 */
	public JPopupMenu setPopMenu(){
		Object[] currentSelection = _ourView.getSelectionCells();

		if ( _ourView.getSketch().getFrame().getMode() == SketchFrame.Mode.MANIPULATE ){
			boolean enableViewQuery = currentSelection.length == 1; 
			_ViewQueryPopItem.setEnabled(enableViewQuery);
			return _manipModePopMenu;
		}
		//dissable all elements
		for(Component c : _editModePopMenu.getComponents()){
			c.setEnabled(false);
		}

		if ( currentSelection.length == 0 ){
			_AddQueryNodePopItem.setEnabled(true);
		}
		else if ( currentSelection.length == 1 ){
			_DefineQueryNodePopItem.setEnabled(true);
		}

		if( currentSelection.length > 0 ){
			_DeletePopItem.setEnabled(true);
		}

		return _editModePopMenu;
	}

	/**
	 * Builds the menu
	 */
	private void buildMenu() {
		JMenuBar mainMenu;
		JMenu menuFile, menuAction, menuHelp;

		mainMenu = new JMenuBar();

		//Make the File Menu
		menuFile = new JMenu("File");
		mainMenu.add(menuFile);
		menuFile.setMnemonic(KeyEvent.VK_F);
		addMenuItem(menuFile, new JMenuItem(new FileClearViewAction(this)), null);
		menuFile.addSeparator();
		addMenuItem(menuFile, new JMenuItem(new DocumentInfoAction(this)), KeyEvent.VK_I);
		menuFile.addSeparator();
		addMenuItem(menuFile, new JMenuItem(new FileQuitAction(this)), KeyEvent.VK_W);
		
		//Create help menu
		menuHelp = new JMenu("Help");
		mainMenu.add(menuHelp);
		menuHelp.setMnemonic(KeyEvent.VK_H);
		menuHelp.add(new HelpAction());
		menuHelp.add(new AboutAction());

		EasikMenuAdjuster menuAdj = Easik.getInstance().getFrame().getMenuAdjuster();
		if (menuAdj != null)
			menuAdj.adjustMenu(mainMenu);
		
		setJMenuBar(mainMenu);
	}

	/**
	 * Assigns then SketchNode representing this frame in the overview
	 */
	public void assignNode(ViewNode inNode){
		_myNode = inNode;
	}
	
	/**
	 * Returns the SketchNode representing this frame in the overview
	 */
	public ViewNode getNode(){
		return _myNode;
	}

	/**
	 * Sets the title of this frame to "EASIK - <view name>"
	 */
	public void setEasikTitle(){
		this.setTitle("EASIK - " + _ourView.getName());
	}
}
