package easik.ui.menu;

import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.io.File;
import java.io.IOException;

import javax.swing.AbstractAction;
import javax.swing.JOptionPane;

import java.sql.SQLException;

import easik.ui.EasikFrame;
import easik.ui.SketchFrame;
import easik.ui.FileChooser;
import easik.ui.FileChooser;
import easik.ui.FileFilter;
import easik.ui.menu.popup.DatabaseOptions;
import easik.ui.menu.popup.ExportOptions;
import easik.sketch.Sketch;
import easik.database.export.SketchExporter;
import easik.database.DatabaseDriver;


/**
 *	
 *	Action to export the current sketch as SQL (either a text file, or directly
 *	to the database).
 *  
 * @author Rob Fletcher 2005
 * @author Vera Ranieri 2006
 * @author Kevin Green 2006
 * @since 2006-05-11 Vera Ranieri
 * @version 2006-08-21 Kevin Green
 */

public class ExportSQLAction extends AbstractAction {
	
	protected Sketch _theSketch;
	protected boolean textExport;

	/**
	 * Create a new ExportSQLAction.  Adds a new menu item to the File menu.
	 */
	public ExportSQLAction(EasikFrame inFrame, Sketch inSketch, boolean text) {
		super((inFrame instanceof SketchFrame ? "" : "Export to ") + (text ? "SQL text file..." : "SQL server..."));
		_theSketch = inSketch;
 		putValue(AbstractAction.MNEMONIC_KEY, new Integer(text ? KeyEvent.VK_T : KeyEvent.VK_S));
 		putValue(AbstractAction.SHORT_DESCRIPTION, text ? "Generate a file containing the SQL needed to generate the SQL schema" : "Generate an SQL schema directly to an SQL database");
		textExport = text;
	}

	/**
	 * Exports the current sketch as an SQL database.  Displays a message if an error occurred.
	 * @param e The action event
	 */
	public void actionPerformed(ActionEvent e ) {
		try {
			DatabaseDriver dbd;

			String type = _theSketch.getDBType();

			if (type == null) return;

			ExportOptions expOpts = new ExportOptions(type, _theSketch.getFrame());

			if ( ! expOpts.isAccepted() ){
				return;
			}

			if (textExport) {
				DatabaseOptions dbopts = new DatabaseOptions(type, _theSketch.getFrame(), true);

				if (!dbopts.isAccepted()) return;

				dbd = DatabaseDriver.loadDriver(type, dbopts.getParams());
				if (_theSketch.setConnectionParams(dbopts.getParams()))
					_theSketch.setDirty();
			}
			else {
				dbd = _theSketch.getDatabaseDriver(type);
			}

			if (dbd == null) return;

			if (!expOpts.isAccepted()) return;

			SketchExporter exporter = dbd.getSketchExporter(_theSketch, expOpts.getParams());

			if (textExport) {
				File selected = FileChooser.saveFile("Select SQL output file", new FileFilter("SQL scripts", "sql", "mysql", "pgsql"), "sql");
				if (selected == null) return;

				try {
					exporter.exportToFile(selected);
					JOptionPane.showMessageDialog(null, "SQL exported successfully.", "Export successful", JOptionPane.INFORMATION_MESSAGE);
				}
				catch (IOException ioe) {
					JOptionPane.showMessageDialog(null, "An error occurred while writing the SQL export file: " + ioe.getMessage(),
							"Error", JOptionPane.ERROR_MESSAGE);
					return;
				}
			}
			else {
				exporter.exportToDB();
				JOptionPane.showMessageDialog(null, 
						"SQL exported successfully. Entering data manipulation mode.",
						"Export successful.",
						JOptionPane.INFORMATION_MESSAGE);
				// We don't have to worry about enableDataManip() failing... if we were able to export, we know it will work
				_theSketch.getFrame().enableDataManip();
			}
		}
		catch (DatabaseDriver.LoadException le) {
			_theSketch.cleanDatabaseDriver();
			JOptionPane.showMessageDialog(null, "An error occurred while exporting SQL: " + le.getMessage(),
					"Error", JOptionPane.ERROR_MESSAGE);
			return;
		}
		catch (SQLException sqle) {
			_theSketch.cleanDatabaseDriver();
			JOptionPane.showMessageDialog(null, "An error occurred while exporting SQL: " + sqle.getMessage(),
					"Error", JOptionPane.ERROR_MESSAGE);
			return;
		}
	}

	/**
	 * At the time this item is added to a menu, the sketch on which it will
	 * act may not be known.  This method sets the sketch to be exported.
	 * @param sketch The sketch that is set to be exported.
	 */
	public void setSketch(Sketch sketch){
		_theSketch = sketch;
	}
}
