package easik.ui.tree;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import java.util.ArrayList;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import easik.view.vertex.QueryNode;
import easik.ui.ViewFrame;
import easik.ui.menu.popup.NewQueryNodeAction;
import easik.ui.tree.popup.*;


/**
 * Class to display the information tree.  This tree contains information about query nodes 
 */
public class ViewInfoTreeUI extends JScrollPane {
	
	/** The information tree */
	private JTree _infoTree;

	/** The information tree model */
	private DefaultTreeModel _infoTreeModel;

	/** The top node of the tree */
	private DefaultMutableTreeNode _topNode;

	/** The 'query nodes' tree node */
	private DefaultMutableTreeNode _tree_nodes;

	/** The popup menu for the tree */
	private JPopupMenu _popupMenu;

	/** The tree popup menu position */
	private Point _popupPosition;

	/** Tree popup menu items */
	private JMenuItem _addQueryNodeItem;
	private JMenuItem _deleteQueryNodeItem;
	private JMenuItem _defineQueryNodeItem;

	/** The frame in which this tree exists */
	private ViewFrame _theFrame;

	/** 
	 * An ArrayList of Strings which represent an expansion state of the info tree. The tree
	 * can then be reverted back to the state as defined in expanState.
	 */
	private ArrayList<String> expanState;
	
	/**
	 * Constructs new view info tree.
	 * @param inFrame The view frame in which this tree exists.
	 */
	public ViewInfoTreeUI(ViewFrame inFrame){
		//Create the top node.
	    _topNode = new DefaultMutableTreeNode("View");

		_theFrame = inFrame;
	    
	    //Create a tree that allows one selection at a time.
	    _infoTreeModel = new DefaultTreeModel(_topNode);
	    _infoTree = new JTree(_infoTreeModel);
	    _infoTree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
	    _infoTree.setRootVisible(false); //Hides root node
	    _infoTree.setShowsRootHandles(true); //Shows nodes off root
	    
	    DefaultTreeCellRenderer renderer = new DefaultTreeCellRenderer();
	    renderer.setOpenIcon(null);
	    renderer.setClosedIcon(null);
	    renderer.setLeafIcon(null);
	    _infoTree.setCellRenderer(renderer);
	    
	    
		//Add tree into scroll pane and then to this pane
		this.setViewportView(_infoTree);
		this.setMinimumSize(new Dimension(150, 200));
		
		_popupMenu = new JPopupMenu();
		_popupPosition = new Point(0, 0);
		buildPopupMenu();
	    
	    //Create Category Heads
	    _tree_nodes = new DefaultMutableTreeNode("Query Nodes");
	    _topNode.add(_tree_nodes);

		//Initialize ArrayList of expansion state information
		expanState = new ArrayList<String>();
}
	
	/**
	 * Creates the popup menu
	 */
	private void buildPopupMenu() {
		_popupMenu.add(_addQueryNodeItem = new JMenuItem(new NewQueryNodeAction(_popupPosition, _theFrame)));
		_popupMenu.add(_deleteQueryNodeItem = new JMenuItem(new DeleteFromViewFromTreeAction(_theFrame)));
		_popupMenu.add(_defineQueryNodeItem = new JMenuItem(new DefineQueryNodeFromTreeAction(_theFrame)));
		
		_infoTree.addMouseListener(new MouseAdapter() {

			public void mousePressed(MouseEvent evt) {
				if (evt.isPopupTrigger()) {
					_infoTree.setSelectionRow(_infoTree.getClosestRowForLocation(evt.getX(), evt.getY()));
					_popupPosition.setLocation(evt.getX(), evt.getY());
					if(setPopMenuItems()){
						_popupMenu.show(evt.getComponent(), evt.getX(), evt.getY());
					}
				}
			}
			public void mouseReleased(MouseEvent evt) {
				if (evt.isPopupTrigger()) {
					_infoTree.setSelectionRow(_infoTree.getClosestRowForLocation(evt.getX(), evt.getY()));
					_popupPosition.setLocation(evt.getX(), evt.getY());
					if(setPopMenuItems()){
						_popupMenu.show(evt.getComponent(), evt.getX(), evt.getY());
					}
				}
			}
		});
	}

	/**
	 * Stores the current expansion state of the info tree.
	 */
	public void storeExpansion(){
		expanState.clear();
		for(int i=0; i < _infoTree.getRowCount(); i++){
			expanState.add(TreeExpansionUtil.getExpansionState(_infoTree, i));
		}
	}

	/**
	 * Reverts the expansion state of the info tree to the last
	 * stored state.
	 */
	public void revertExpansion(){
		for(int i=0; i<expanState.size(); i++){
			TreeExpansionUtil.restoreExpansionState(_infoTree, i, expanState.get(i));
		}
	}
	
	/**
	 * Sets which of the menu items will be visible
	 * 
	 * @return true if the popup should be displayed, false otherwise
	 */
	public boolean setPopMenuItems(){
		//If there is nothing seleceted then just do nothing
		if(_theFrame.getInfoTreeUI().getInfoTree().isSelectionEmpty())
		{
			return false;
		}
		
		//Get currently selected object //FIXME is _theFrame().getInfoTreeUI() returning this??
		DefaultMutableTreeNode curSelectedNode = (DefaultMutableTreeNode) 
					_theFrame.getInfoTreeUI().getInfoTree().getSelectionPath().getLastPathComponent();
		Object curSelected = curSelectedNode.getUserObject();
		
		//Hide all elements
		for(Component c : _popupMenu.getComponents()){
			c.setVisible(false);
		}
		
		//Check what is currently selected
		if(curSelected instanceof QueryNode){
			_deleteQueryNodeItem.setVisible(true);
			_defineQueryNodeItem.setVisible(true);
		}
		else if(curSelectedNode == _tree_nodes){
			_addQueryNodeItem.setVisible(true);
		}
		return true;
	}
	
	/**
	 * Refreshes the tree so updates visualize properly
	 */
	public void refreshTree(){
		storeExpansion();
		_infoTreeModel.reload(_topNode);
		revertExpansion();
	}
	
	/**
	 * Refreshes the tree so updates visualize properly
	 * 
	 * @param inNode The node to be refreshed
	 */
	public void refreshTree(DefaultMutableTreeNode inNode){
		storeExpansion();
		_infoTreeModel.reload(inNode);
		revertExpansion();
	}
	
	/**
	 * Returns the information tree model
	 * 
	 * @return The information tree model
	 */
	public DefaultTreeModel getInfoTreeModel(){
		return _infoTreeModel;
	}
	
	/**
	 * Returns the information tree
	 * 
	 * @return The information tree
	 */
	public JTree getInfoTree(){
		return _infoTree;
	}
	
	/**
	 * Returns the Entities tree node
	 * 
	 * @return The Entities tree node
	 */
	public DefaultMutableTreeNode getEntities(){
		return _tree_nodes;
	}
	
	/**
	 * Adds an entity and its attributes to the tree
	 * 
	 * @param inEntity The entity to be added to the tree
	 */
	public void addQueryNode(QueryNode inNode){
		//Create entity Node
		DefaultMutableTreeNode myQueryNode = new DefaultMutableTreeNode(inNode);
		_tree_nodes.add(myQueryNode);
		inNode.setNode(myQueryNode);
		
		refreshTree(_tree_nodes);
		_infoTree.scrollPathToVisible(new TreePath(myQueryNode.getPath()));
	}
	
	/**
	 * Removes an entity from the tree
	 * 
	 * @param toRemove Entity to be removed from the tree
	 */
	public void removeQueryNode(QueryNode toRemove){
		_tree_nodes.remove(toRemove.getNode());
		refreshTree(_tree_nodes);
	}
}
