package easik.view;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.Rectangle2D;
import java.awt.Color;
import java.awt.image.*;
import java.awt.event.*;
import java.awt.Image;
import java.util.*;

import java.sql.*;

import javax.swing.*;

import org.jgraph.JGraph;
import org.jgraph.graph.*;
import org.jgraph.event.*;

import easik.*;
import easik.overview.Overview;
import easik.sketch.Sketch;
import easik.sketch.vertex.*;
import easik.ui.*;
import easik.ui.tree.ViewInfoTreeUI;
import easik.view.vertex.*;
import easik.ui.menu.popup.DefineQueryNodeAction;

/** 
 */
public class View extends JGraph {

	/** Hash Map of all entity nodes, indexed by their name */
	private LinkedHashMap<String, QueryNode> _queryNodes;

	/** The current ApplicationFrame */
	private ViewFrame _viewFrame;

	/** The current DocumentInfo */
	private DocumentInfo _docInfo;

	/** The overview in which this view exists */
	private Overview _theOverview;

	/** The sketch this view is on */
	private Sketch _ourSketch;

	/**
	 * The default constructor sets all the visual settings for the JGraph, as well as
	 * initialising the sketch to be empty. It also adds appropriate listeners for all 
	 * of the actions we are concerned with.
	 * 
	 * @param inFrame The view frame of the sketch 
	 * @param inSketch The sketch this view is on
	 * @param inOverview The overview in which this 
	 */
	public View(ViewFrame inFrame, Sketch inSketch, Overview inOverview) {
		super();

		setBackground(Easik.getInstance().getSettings().getColor("view_canvas_background"));
		_viewFrame = inFrame;
		_theOverview = inOverview;
		_ourSketch = inSketch;

		setAntiAliased(true);
		setDisconnectable(false);
		setConnectable(false);
		setEditable(false);
		setSizeable(false);
		initialiseView();
		getGraphLayoutCache().setAutoSizeOnValueChange(true);
		addGraphSelectionListener(new GraphSelectionListener() {
			public void valueChanged(GraphSelectionEvent gsEvent) {
				View.this.getGraphLayoutCache().reload();
			}
		});

		// Set up mouse listener to watch for double clicks
		//  - Double clicks edit a query node
		this.addMouseListener(new MouseAdapter(){
			public void mouseClicked(MouseEvent e){
				if (_ourSketch.getFrame().getMode() == SketchFrame.Mode.EDIT && e.getClickCount() == 2) {
					Object[] currSelection = getSelectionCells();
					if (currSelection.length == 1 && currSelection[0] instanceof QueryNode) {
						DefineQueryNodeAction.updateNode((QueryNode)currSelection[0]);
					}
				}
			}
		});
		
		updateUI();
	}

	/**
	 * Since this is a Swing component, this method is overloading a method of
	 * JGraph to adjust the look and feel. The feel we are changing is ignoring all but
	 * left clicks, allowing for right-click functionality not affecting the selections.
	 */
	public void updateUI() {
		this.setUI(new GraphUI());
		this.invalidate();
	}

	/**
	 *	Returns the parental application frame to whoever asks for it.
	 * @return The current application frame
	 */
	public ViewFrame getFrame() {
		return _viewFrame;
	}
	
	/**
	 * Gets the document information
	 * @return The document information 
	 */
	public DocumentInfo getDocInfo(){
		return _docInfo;
	}

	/**
	 * When we initialise the sketch, we flush out all the data concerning the sketch
	 * itself. Even the modelAdapter is reinitialized.
	 *
	 *	This methods serves as a "new sketch" function. 
	 */
	public void initialiseView() {
		GraphModel model = new ViewGraphModel(this);
		GraphLayoutCache glc = new GraphLayoutCache(model, new DefaultCellViewFactory());
		setModel(model);
		setGraphLayoutCache(glc);

		_queryNodes = new LinkedHashMap<String, QueryNode>();
		if(_viewFrame.getInfoTreeUI() != null)
		{
			_viewFrame.setInfoTreeUI(new ViewInfoTreeUI(_viewFrame)); //Wipe Tree
			_viewFrame.getInfoTreeUI().refreshTree();
		}
		_docInfo = new DocumentInfo(_viewFrame);
		_viewFrame.setTitle("EASIK - Untitled");
	}

	/**
	 * When we initialise a new sketch, we need to clear the selection
	 * buffer just in case something is still selected. Or else it will remain
	 * selected because there will be no events removing it.	 
	 */
	public void newView() {
		initialiseView();
		clearSelection();
	}

	/**
	 * Used to initialise a new sketch based on provided data (usually from the Sketch loading
	 * methods).
	 * 
	 * @param queryNodes A map of all of the query nodes in the view
	 * @param head The header created from the loaded XML file.
	 */
	public void initializeFromData(
		Map<String, QueryNode> queryNodes,
		DocumentInfo head)
	{
		initialiseView();
		
		_queryNodes = new LinkedHashMap<String,QueryNode>(queryNodes);
		_docInfo = head;
		
		_viewFrame.setTreeName(_docInfo.getName());

		//Add the entities
		for (QueryNode node: _queryNodes.values())
			if (node != null)
				addQueryNode(node);

		refresh();
	}

	/**
	 * Refreshes the specified items in the view GUI.  If called with no
	 * arguments, all items are refreshed.
	 */
	public void refresh(GraphCell... cells) {
		setBackground(Easik.getInstance().getSettings().getColor("view_canvas_background"));
		Object[] toRefresh = cells.length > 0 ? cells : getRoots();

		setBackground(Easik.getInstance().getSettings().getColor("view_canvas_background"));

		((DefaultGraphModel) getModel()).cellsChanged(toRefresh);

		if (cells.length == 0)
			super.refresh();
	}

	/**
	 * Overrides superclass's refresh() method to go through the above
	 */
	public void refresh() {
		refresh(new GraphCell[0]);
	}

	/**
	 * Accessor for a single query node in the view.  Takes the node's name.
	 *
	 * @param entName the entity name
	 * @return QueryNode associated with that name, or null if the entity does not exist
	 */
	public QueryNode getQueryNode(String nodeName) {
		return _queryNodes.get(nodeName);
	}

	/**
	 * Accessor for the entities in the sketch
	 *
	 * @return Collection of the entities
	 */
	public Collection<QueryNode> getQueryNodes() {
		return Collections.unmodifiableCollection(_queryNodes.values());
	}

	/**
	 * Called internally by QueryNode when a node is renamed, to keep the
	 * entity node map consistent.  Should not be called directly; instead
	 * call sketch.getEntity("currentname").setName("newname").
	 *
	 * @see easik.view.vertex.QueryNode.setName(String)
	 * @param node the QueryNode being renamed
	 * @param oldName the old name of the node
	 * @param newName the candidate new name
	 * @return a string containing the final new node name, for QueryNode to use.
	 */
	public String nodeRenamed (QueryNode node, String oldName, String newName) {
		// If the name already exists, we have to rename it
		while (_queryNodes.containsKey(newName))
			newName = EasikTools.incrementName(newName);

		_queryNodes.put(newName, _queryNodes.remove(oldName));
		return newName;
	}

	/**
	 * Removes an entity, and also cascades to remove all the arrows 
	 * involved with it.
	 * 
	 * @param toRemove The entity about to be removed
	 */
	public void removeQueryNode(QueryNode toRemove) {

		_queryNodes.remove(toRemove.toString());
		getGraphLayoutCache().remove(new Object[] {toRemove});
		_viewFrame.getInfoTreeUI().removeQueryNode(toRemove);
	}

	/**
	 * Add a new, empty entity at point X, Y 
	 *
	 * @param name The name of the new entity being added
	 * @param x X Coordinate of new entity
	 * @param y Y Coordinate of new entity
	 */
	public void addNewQueryNode(String name, double x, double y) {
		QueryNode newQueryNode;

		newQueryNode = new QueryNode(name, (int) x, (int) y, this, "");
		addQueryNode(newQueryNode);

		//_theOverview.refresh();
	}

	/**
	 * Add an entity to the graph, deals with all of the dependencies.
	 * 
	 * @param theEntity
	 */
	public void addQueryNode(QueryNode theNode) {
		// The next call will fire a rendering. At this point, the model adapter does not know
		// where it should place the node, and picks a default value. This will cause an update
		// in our Node's x and y coordinates, making it forget where it was initialized.
		// We store it's initializeded position so as not to lose them in the first rendering.
		int initX = theNode.getX();
		int initY = theNode.getY();

		// Set the on-screen position of our entity to the attributes of the entity
		AttributeMap nAttribs = theNode.getAttributes();
		GraphConstants.setAutoSize(nAttribs, true);
		GraphConstants.setBounds(nAttribs, new Rectangle2D.Double(initX, initY, 0, 0));

		// Add our entity to the graph				
		getGraphLayoutCache().insert(theNode);

		// Add our entity to our table of entities
		_queryNodes.put(theNode.toString(), theNode);

		//Add Entity to tree
		_viewFrame.getInfoTreeUI().addQueryNode(theNode);

		// Reload the graph to respect the new changes
		refresh();

	}

	/**
	* Returns the next available name, starting from 'NewName0', so we don't get duplicates.
	* 
	* @return The next new name.
	*/
	public String getNewName() {
		return getNewName("NewQueryNode0");
	}

	/**
	 * Returns the next available name, starting from the passed-in name.
	 *
	 * @param name the initial name to try; if it exists, we add/increment a number until it's unique
	 * @return the first available unused name
	 */
	public String getNewName(String name) {
		while (isNameUsed(name))
			name = EasikTools.incrementName(name);

		return name;
	}

	/**
	* Checks to see if a name is in use, so that we will not have several instances
	* at once.
	* 
	* @param inName The desired new name to check against
	* @return Is it used or not.
	*/
	public boolean isNameUsed(String inName) {
		
		Set<String> keys = _queryNodes.keySet();
		for (String name: keys){
			if(name.equalsIgnoreCase(inName))
				return true;
		}
		for (EntityNode en: _ourSketch.getEntities()){
			if(en.getName().equalsIgnoreCase(inName))
				return true;
		}
		
		return false;
	}

	/**
	 * Determines if this sketch contains any elements.
	 *
	 * @return True if this sketch contains any elements, and false otherwise
	 */
	public boolean isEmpty(){
		return getRoots().length == 0;
	}

	/**
	 * Getter method for the overview in which this sketch exists.
	 *
	 * @return The overview in which this sketch exists
	 */
	public Overview getOverview(){
		return _theOverview;
	}

	/**
	 * Takes a snapshot of the current sketch, sends the image to scaleImage(..), and sets in
	 * the node representing this sketch.
	 */
	public void updateThumb(){
		if( ! _viewFrame.isDisplayable()){
			// set preferred size to current size, and then pack to make sketch displayable.
			// This is required for JGraph.getImage(..) to work properly
			_viewFrame.setPreferredSize(_viewFrame.getSize());
			_viewFrame.pack();
		}

		//We always want our thumbnail background to be white
		Color oldColor = getBackground();
		setBackground(java.awt.Color.WHITE);
		Image temp = (Image)getImage(null,0);
		setBackground(oldColor);

		if(temp != null){
			temp = scaleThumb((BufferedImage)temp);
		}

		_viewFrame.getNode().setThumbnail(temp);
	}
	
	/**
	 * Scales image by factor specified for x and y directions.
	 *
	 * @param inThumb The image to be scaled
	 *
	 * @return The scaled image
	 */
	private static Image scaleThumb(BufferedImage inThumb){
		double scaleFactor = Double.parseDouble(Easik.getInstance().getSettings().getProperty("thumb_scale_factor", "0.5"));
		return inThumb.getScaledInstance((int)(inThumb.getWidth()*scaleFactor), (int)(inThumb.getHeight()*scaleFactor), Image.SCALE_SMOOTH);
	}

	/**
	 * Gets the name of this sketch. Simply a shortcut for _docInfo.getName().
	 */
	public String getName(){
		return _docInfo.getName();
	}

	/**
	 * Gets the sketch to which this view is associated.
	 * @return The sketch to which this view is associated.
	 */
	public Sketch getSketch(){
		return _ourSketch;
	}

	/**
	 * Used to mark a view as dirty.  When called, we mark the Overview as dirty,
	 * and update the last modification time of the view to the current time.
	 */
	public void setDirty() {
		getOverview().setDirty(true);
		getOverview().refresh(_viewFrame.getNode());
		getDocInfo().updateModificationDate();
	}
}
